/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: discdomain.cpp
**  
** Description:
** The source file for a discretized domain. 
**  
** Last modified: 
** 09 January 2006
** 
** Comments:
** 1) Current implementation uses the Mersenne Twister for random number generation
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <sstream>

/* Application Includes */
#include "discdomain.h"
#include "random.h"

using namespace std;
using namespace random;

namespace domain {

/** Default constructor
**/
DiscDomain::DiscDomain(double lowerBound, double upperBound) : mPartitions(1) {
	// Checks if the upper bound is more than the lower bound.
	// If it is not, then reset to the default, i.e. lowerBound = 0.0, upperBound = 1.0
	if (upperBound <= lowerBound) {
		#ifdef DISCDOMAIN_DEBUG
			cout << "[Error] DiscDomain::DiscDomain() - Upper bound " << upperBound << " is greater than lower bound "
				 << lowerBound << ". Reset to default (0.0, 1.0]" << endl;
		#endif
		mLowerBound = 0.0;
		mUpperBound = 1.0;
	} else {
		mLowerBound = lowerBound;
		mUpperBound = upperBound;
	}
	
	// Allocate space for the upper boundary list
	mUpperBoundaryList = new double[mPartitions];
	mUpperBoundaryList[0] = mUpperBound;
}

/** Copy constructor
**/
DiscDomain::DiscDomain(const DiscDomain& src) {
	copyFrom(src);
}

/** Assignment constructor
**/
DiscDomain& DiscDomain::operator=(const DiscDomain& src) {

	// Checks for self assignment
	if (this == &src) {
		return (*this);	
	}

	// Delete the upperboundary list
	delete[] mUpperBoundaryList;
	
	// Copy contents
	copyFrom(src);
	
	return (*this);
}

/** Default destructor
**/
DiscDomain::~DiscDomain() {

	#ifdef DISCDOMAIN_DEBUG
		cout << "[Info] DiscDomain::~DiscDomain() - Destroying discretize domain [" << mLowerBound << ", " << mUpperBound << "]" << endl;
	#endif

	// Destroy the boundarylist
	delete[] mUpperBoundaryList;
}

/** Returns the index of the value by checking with the mUpperBoundaryList
* Note that only the first partition is close. The rest are half open at the lower bound
**/
int DiscDomain::indexOf(double value) {
	
	// Checks if the value is within the boundary. If it is not, return -1
	if (value < mLowerBound || value > mUpperBound) {
		#ifdef DISCDOMAIN_DEBUG
			cout << "[Error] DiscDomain::indexOf() - value " << value << " out of bounds" << endl;
		#endif
		return -1;
	}
	
	int index;
	for (index = 0; index < mPartitions; index++) {
		if (value <= mUpperBoundaryList[index]) {
			break;
		}
	}
	
	return index;
}

/** Returns the number of partitions in this domain
**/
int DiscDomain::partitionSize() {
	return mPartitions;
}

/** Repartitions the domain into bins
**/
bool DiscDomain::repartition(int partitions) {
	
	// Check if the number of partitions is more than or equal to 1
	if (partitions < 1) {
		#ifdef DISCDOMAIN_DEBUG
			cout << "[Error] DiscDomain::repartition() - Invalid number of partitions (" << partitions << ")" << endl;
		#endif
		return false;
	}
	
	// If partitions is equal to the current number of partitions, then nothing needs to be done
	if (partitions == mPartitions) {
		#ifdef DISCDOMAIN_DEBUG
			cout << "[Info] DiscDomain::repartition() - Number of partitions unchanged" << endl;
		#endif
		return true;
	}
	
	// Delete the current upperbound list
	delete[] mUpperBoundaryList;
	
	mPartitions = partitions;
	double width = (mUpperBound - mLowerBound) / (mPartitions);
	mUpperBoundaryList = new double[mPartitions];
	for (int i = 0; i < mPartitions - 1; i++) {
		mUpperBoundaryList[i] = mLowerBound + ((i + 1) * width);
	}
	mUpperBoundaryList[mPartitions - 1] = mUpperBound;
	
	return false;
}

/** Finds the median value of any boundary
**/
double DiscDomain::midValue(int partition) {
	
	// Checks if the given partition index is valid
	if (partition < 0 || partition >= mPartitions) {
		#ifdef DISCDOMAIN_DEBUG
			cout << "[Error] DiscDomain::midValue() - Partition index (" << partition << ") out of bounds" << endl;
		#endif
		return -1.0;
	}
	
	double lower;
	double upper;
	double mid;
	
	if (partition == 0) {
		lower = mLowerBound;
	} else {
		lower = mUpperBoundaryList[partition - 1];
	}
	upper = mUpperBoundaryList[partition];
	mid = (upper - lower) / 2.0 + lower;
	
	return mid;
}

/** Generates a random number between the upper and lower bound
* Uses the Mersenne Twister random number generator.
**/
double DiscDomain::random() {
	double value = MersenneTwister::randRealClose() * (mUpperBound - mLowerBound) + mLowerBound;
	return value;
}

/** Generates a random number between the [0.0, 1.0]
* Uses the Mersenne Twister random number generator.
**/
double DiscDomain::randomUnit() {
	return MersenneTwister::randRealClose();
}

/** Generates a random number within the partition that is given by the index
* Note that only the first partition is close. The rest are half open at the lower bound.
* Hence if the index is zero, use the close random between upper and lower bound. Otherwise
* generate one that is more than the lower bound and less than or equal to upper bound
* If the index is out of scope, returns a 0. 
**/
double DiscDomain::random(int index) {
	
	// Checks that the index is valid
	if (index < 0 || index >= mPartitions) {
		#ifdef DISCDOMAIN_DEBUG
			cout << "[Error] DiscDomain::random(2) - Partition index out of bounds (" << index << ")" << endl;
		#endif
		return 0.0;
	}
	
	double lowerBound;
	double upperBound;
	double value;
	
	if (index == 0) {
		lowerBound = mLowerBound;
	} else {
		lowerBound = mUpperBoundaryList[index - 1];
	}
	upperBound = mUpperBoundaryList[index];
	
	if (index == 0) {
		value = MersenneTwister::randRealClose() * (upperBound - lowerBound) + lowerBound;
	} else {
		// The random number is subtracted from 1 to convert from [0, 1) into (0, 1]
		value = (1.0 - MersenneTwister::randRealHalf())  * (upperBound - lowerBound) + lowerBound;
	}
	
	return value;
}

/** Returns a string representation of the discretized domain
**/
string DiscDomain::toString() const {
	ostringstream str;
	
	str << "---DiscDomain---" << endl;
	str << "mLowerBound: " << mLowerBound << endl;
	str << "mUpperBound: " << mUpperBound << endl;
	str << "mPartitions: " << mPartitions << endl;
	str << "Partitions:" << endl;
	str << "(" << mLowerBound << " - " << mUpperBoundaryList[0] << ") "; 
	for (int i = 1; i < mPartitions; i++) {
		str << "(" << mUpperBoundaryList[i - 1] << " - " << mUpperBoundaryList[i] << ") "; 
	}
	str << endl;
	
	return str.str();
}

/** Helper function for copying contents
* It copies everything but assumes that allocated memory (if any)
* has already been deleted.
**/
void DiscDomain::copyFrom(const DiscDomain& src) {
	
	mLowerBound = src.mLowerBound;
	mUpperBound = src.mUpperBound;
	mPartitions = src.mPartitions;
	
	mUpperBoundaryList = new double[mPartitions];
	for (int i = 0; i < mPartitions; i++) {
		mUpperBoundaryList[i] = src.mUpperBoundaryList[i];
	}
}
	
}
