/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: discretedistribution.cpp
**  
** Description:
** The source file for a discrete probability distribution table
**  
** Last modified: 
** 16 March 2007
** 
** Comments:
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <sstream>
#include <fstream>
#include <map>
#include <vector>
#include <math.h>

/* Application Includes */
#include "discretedistribution.h"

using namespace std;

namespace distribution {

/** Checks if the variable name is valid
* Checks if a given variable name is valid, according to standard variable rules.
* 1) It must not be empty
* 2) It must start with an alphabet or underscore "_"
* 3) It must not contain any invalid characters, such as ", ", "?", " " etc.
**/
bool isValidName(string varName) {

	// Checks for empty string
	if (varName == "") return false;
	
	char c;
	
	// Checks the first charactor
	c = varName[0];
	if (c != '_' && !isalpha(c)) return false;
	
	// Checks the rest of the characters. The rest can only
	// be alphanumeric or '_'
	int length = strlen(varName.c_str());
	for (int i = 1; i < length; i++) {
		c = varName[i];
		if (c != '_' && !isalnum(c)) return false;
	}
	
	return true;
}

/** Default constructor
**/
DiscreteDistribution::DiscreteDistribution() : mInitialized(false), mData(0), mDimension(0), mDataSize(0) {
	// Do nothing
	
	#ifdef DISCRETEDISTRIBUTION_DEBUG
		cout << "[Info] DiscreteDistribution::DiscreteDistribution() - Creating DiscreteDistribution..." << endl;
	#endif
	
}

/** Copy constructor
* Copies all the contents (By creating a new instance of them) from src to the new instance.
**/
DiscreteDistribution::DiscreteDistribution(const DiscreteDistribution& src) {
	copyFrom(src);
}

/** Assignment constructor
**/
DiscreteDistribution& DiscreteDistribution::operator=(const DiscreteDistribution& src) {

	// Checks for self assignment
	if (this == &src) {
		return (*this);	
	}
	
	// Clears all the entries in the data array, variable list and size list
	if (mData != 0) {
		delete[] mData;
	}
	mVariableList.clear();
	mSizeList.clear();
	mOffsetList.clear();
	
	// Copy contents
	copyFrom(src);
	
	return (*this);
}

/** Destructor
**/
DiscreteDistribution::~DiscreteDistribution() {

	#ifdef DISCRETEDISTRIBUTION_DEBUG
		cout << "[Info] DiscreteDistribution::~DiscreteDistribution() - Destroying DiscreteDistribution : ";
		for (int i = 0; i < mVariableList.size(); i++) {
			cout << mVariableList[i] << " ";
		}
		cout << endl;
	#endif

	// Clears the entries in the data array, variable list and size list
	if (mData != 0) {
		delete[] mData;
	}
	mVariableList.clear();
	mSizeList.clear();
	mOffsetList.clear();
}

/** Adds a variable (dimension) to the distribution table
* This method is only valid if the table has not been initialized.
**/
bool DiscreteDistribution::addVariable(string variable, int stateSize) {
	
	// Checks if the distribution is already initialized. If it is, then no more variables can
	// be added and an empty string is returned.
	if (mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::addVariable() - Distribution is already initialized" << endl;
		#endif
		return false;
	}
	
	// Checks if the number of states is correct and that the variable name is valid
	if (!isValidName(variable) || stateSize < 1) {
		return false;
	}
	
	// Checks if the variable already exists
	if (hasVariable(variable)) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::addVariable() - Variable " << variable << " exists" << endl;
		#endif
		return false;
	}
	
	// Adds variable in the variable list
	mVariableList.push_back(variable);
	mSizeList.push_back(stateSize);
	mDimension++;
	return true;
}

/** Removes a variable from the distribution table
* This method is only valid if the table has not been initialized.
**/
bool DiscreteDistribution::removeVariable(string variable) {

	// Checks if the table is already initialized
	if (mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::removeVariable() - Distribution is already initialized" << endl;
		#endif
		return false;
	}
	
	int index;
	for (index = 0; index < mVariableList.size(); index++) {
		if (mVariableList.at(index) == variable) {
			mVariableList.erase(mVariableList.begin() + index);
			mSizeList.erase(mSizeList.begin() + index);
			mDimension--;
			return true;
		}
	}
	
	// If this point is reached, no such variable is found and a false is returned
	#ifdef DISCRETEDISTRIBUTION_DEBUG 
		cout << "[Error] DiscreteDistribution::removeVariable() - Variable " << variable <<  " does not exists" << endl;
	#endif
	return false;
	
}

/** Checks if a variable is a member of the distribution
**/
bool DiscreteDistribution::hasVariable(string variable) const {
	
	vector<string>::const_iterator iter = find(mVariableList.begin(), mVariableList.end(), variable);
	if (iter != mVariableList.end()) {
		return true;
	} else {
		return false;
	}

}

/** Checks the index of a variable in this distribution
* If the variable is not found, a -1 is returned
**/
int DiscreteDistribution::indexOf(string variable) const {

	int index;
	for (index = 0; index < mVariableList.size(); index++) {
		if (mVariableList.at(index) == variable) {
			return index;
		}
	}
	
	// No variable found. Return -1
	return -1;
}

/** Initializes the distribution
* Allocates space for the distribution table.
**/
bool DiscreteDistribution::initializeTable() {
	
	// Checks if there is at least one dimension
	if (mDimension < 1) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::initializeTable() - This distribution has no dimensions" << endl;
		#endif
		return false;
	}
	
	// Compute the size needed by multiplying the contents of the size list
	// also populates the offset list
	mDataSize = 1;
	int offset = 1;
	int rindex = mSizeList.size() - 1;
	for (int index = 0; index < mSizeList.size(); index++) {
		mDataSize = mDataSize * mSizeList[index];
		mOffsetList.insert(mOffsetList.begin(), offset);
		offset = offset * mSizeList[rindex--];
	}


	#ifdef DISCRETEDISTRIBUTION_DEBUG 
		cout << "[Info] DiscreteDistribution::initializeTable() - Size of data array: " << mDataSize << endl;
	#endif
	
	// Dynamically creates size and sets all the entries to 0
	mData = new double[mDataSize];
	for (int index = 0; index < mDataSize; index++) {
		mData[index] = 0.0;
	}
	mInitialized = true;
	
	return true;
}

/** Checks if the distribution has been initialized
**/
bool DiscreteDistribution::isInitialized() const {
	return mInitialized;
}

/** Returns the state size of the given variable
* If the variable is not found, returns -1
**/
int DiscreteDistribution::getSize(string variable) const {
	int index = indexOf(variable);
	if (index != -1) {
		return mSizeList.at(index);
	} else {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::getSize() - Variable " << variable << " does not exist" << endl;
		#endif
		return -1;
	}
}

/** Returns the dimension of this discrete distribution
**/
int DiscreteDistribution::getDimension() const {
	return mDimension;
}

/** Returns the size of the data array, i.e. total number of entries in the data array
**/
int DiscreteDistribution::getDataSize() const {
	return mDataSize;
}

/** Given an unordered vector of <variable, state> pair, return the probability
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::getProbability(const vector<pair> pairList) const {
	
	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::getProbability(1) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	// Finds the index in the data array
	int dataIndex = dataIndexOf(pairList);
	
	if (dataIndex == -1) {
		return -1.0;
	} else {
		return mData[dataIndex];
	}
	
}

/** Given an unordered vector of <variable, state> pair, set the probability to the given value
* On error, returns a -1.0
**/
double DiscreteDistribution::setProbability(const vector<pair> pairList, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::setProbability(1) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	// Checks if the value is valid, i.e. above or equals to 0
	if (value < 0.0) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::setProbability(1) - Invalid value : " << value << endl;
		#endif
		return -1.0;
	}
	
	// Finds the index in the data array
	int dataIndex = dataIndexOf(pairList);
	
	if (dataIndex == -1) {
		return -1.0;
	} else {
		mData[dataIndex] = value;
		return value;
	}

}

/** Given an unordered vector of <variable, state> pair, increment the probability by the given amount
* The check on the value will only come after. We allow negative value but the final entry on the data array MUST
* always be positive. 
**/
double DiscreteDistribution::incProbability(const vector<pair> pairList, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::incProbability(1) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	// Finds the index in the data array
	int dataIndex = dataIndexOf(pairList);
	
	if (dataIndex == -1) {
		return -1.0;
	} 
	
	double newValue = mData[dataIndex] + value;
	if (newValue < 0.0) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::incProbability(1) - Invalid value : " << value << endl;
		#endif
		return -1.0;
	} else {
		mData[dataIndex] = newValue;
		return newValue;
	}
}

/** Given an unordered vector of <variable, state> pair, multiply the probability by the given amount
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::multProbability(const vector<pair> pairList, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::multProbability(1) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	// Checks if the value is valid, i.e. above or equals to 0
	if (value < 0.0) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::multProbability(1) - Invalid value : " << value << endl;
		#endif
		return -1.0;
	}
	
	// Finds the index in the data array
	int dataIndex = dataIndexOf(pairList);
	
	if (dataIndex == -1) {
		return -1.0;
	} else {
		mData[dataIndex] = mData[dataIndex] * value;
		return mData[dataIndex];
	}

}

/** Given a string expression, get the probability
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::getProbability(string expr) const {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::getProbability(2) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	vector<int> stateList = convExprToStateList(expr);
	return getProbability(stateList);
}

/** Given a string expression, set the probability to the given value
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::setProbability(string expr, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::setProbability(2) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	vector<int> stateList = convExprToStateList(expr);
	return setProbability(stateList, value);
}

/** Given a string expression, increment the probability by the given amount
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::incProbability(string expr, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::incProbability(2) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	vector<int> stateList = convExprToStateList(expr);
	return incProbability(stateList, value);
}

/** Given a string expression, multiply the probability by the given value
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::multProbability(string expr, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::multProbability(2) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	vector<int> stateList = convExprToStateList(expr);
	return multProbability(stateList, value);

}

/** Given an ordered vector of states, return the probability
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::getProbability(const vector<int> stateList) const {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::getProbability(3) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	
	
	// Finds the index in the data array
	int dataIndex = dataIndexOf(stateList);
	
	if (dataIndex == -1) {
		return -1.0;
	} else {
		return mData[dataIndex];
	}

}

/** Given an ordered vector of states, set the probability to the given value
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::setProbability(const vector<int> stateList, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::setProbability(3) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	// Checks if the value is valid, i.e. above or equals to 0
	if (value < 0.0) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::setProbability(3) - Invalid value : " << value << endl;
		#endif
		return -1.0;
	}
	
	// Finds the index in the data array
	int dataIndex = dataIndexOf(stateList);
	
	if (dataIndex == -1) {
		return -1.0;
	} else {
		mData[dataIndex] = value;
		return value;
	}

}

/** Given an ordered vector of states, increment the probability by the given amount
* The check on the value will only come after. We allow negative value but the final entry on the data array MUST
* always be positive. 
**/
double DiscreteDistribution::incProbability(const vector<int> stateList, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::incProbability(3) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	// Finds the index in the data array
	int dataIndex = dataIndexOf(stateList);
	
	if (dataIndex == -1) {
		return -1.0;
	} 
	
	double newValue = mData[dataIndex] + value;
	if (newValue < 0.0) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::incProbability(3) - Invalid value : " << value << endl;
		#endif
		return -1.0;
	} else {
		mData[dataIndex] = newValue;
		return newValue;
	}
	
}

/** Given an ordered vector of states, multiply the probability by the given value
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::multProbability(const vector<int> stateList, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::multProbability(3) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	// Checks if the value is valid, i.e. above or equals to 0
	if (value < 0.0) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::multProbability(3) - Invalid value : " << value << endl;
		#endif
		return -1.0;
	}
	
	// Finds the index in the data array
	int dataIndex = dataIndexOf(stateList);
	
	if (dataIndex == -1) {
		return -1.0;
	} else {
		mData[dataIndex] = mData[dataIndex] * value;
		return mData[dataIndex];
	}
	
}

/** Given the index of the distribution, returns the probability at that index
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::getProbability(int index) const {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::getProbability(4) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	
	
	// Returns the position of that index if the index is within bounds. Otherwise returns a -1
	if (index < 0 || index >= mDataSize) {
		return -1.0;
	} else {
		return mData[index];
	}

}

/** Given the index of the distribution, sets the probability at that index
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::setProbability(int index, double value) {
	
	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::setProbability(4) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	// Checks if the value is valid, i.e. above or equals to 0
	if (value < 0.0) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::setProbability(4) - Invalid value : " << value << endl;
		#endif
		return -1.0;
	}
	
	// Updates the distribution if the index is within bounds
	if (index < 0 || index >= mDataSize) {
		return -1.0;
	} else {
		mData[index] = value;
		return value;
	}
}

/** Given the index of the distribution, increment the probability by the given value
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::incProbability(int index, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::incProbability(4) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	// Checks if the index is out of bounds
	if (index < 0 || index >= mDataSize) {
		return -1.0;
	} 
	
	double newValue = mData[index] + value;
	if (newValue < 0.0) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::incProbability(4) - Invalid value : " << value << endl;
		#endif
		return -1.0;
	} else {
		mData[index] = newValue;
		return newValue;
	}
	
}

/** Given the index of the distribution, multiply the probability by the given value.
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::multProbability(int index, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::multProbability(4) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	// Checks if the value is valid, i.e. above or equals to 0
	if (value < 0.0) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::multProbability(4) - Invalid value : " << value << endl;
		#endif
		return -1.0;
	}
	
	// Checks if the index is out of bounds
	if (index < 0 || index >= mDataSize) {
		return -1.0;
	} else {
		mData[index] = mData[index] * value;
		return mData[index];
	}
	
}

/** Given a string expression, replace the position in the table if the given value is larger
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::replaceIfGreaterProbability(string expr, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::replaceIfGreaterProbability(1) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	vector<int> stateList = convExprToStateList(expr);
	return replaceIfGreaterProbability(stateList, value);

}

/** Given an ordered vector of states, replace the probability if the given value is greater
* A value check need not be done as the original distribution is already assumed to be correct.
* So if the given value is invalid, it means nothing will be done to the table.
*/
double DiscreteDistribution::replaceIfGreaterProbability(const vector<int> stateList, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::replaceIfGreaterProbability(2) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	// Finds the index in the data array
	int dataIndex = dataIndexOf(stateList);
	
	if (dataIndex == -1) {
		return -1.0;
	} 
	
	// Checks the value at the data index with the given value, replacing it if the value is greater
	if (value > mData[dataIndex]) {
		mData[dataIndex] = value;
	}
	
	return mData[dataIndex];

}

/** Given the index of the distribution, replace the probability if the given value is greater
* A value check need not be done as the original distribution is already assumed to be correct.
* So if the given value is invalid, it means nothing will be done to the table.
**/
double DiscreteDistribution::replaceIfGreaterProbability(int index, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::replaceIfGreaterProbability(3) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	// Checks if the index is out of bounds
	if (index < 0 || index >= mDataSize) {
		return -1.0;
	} 
	
	// Checks the value at the data index with the given value, replacing it if the value is greater
	if (value > mData[index]) {
		mData[index] = value;
	}
	
	return mData[index];

}

/** Given a string expression, replace the position in the table if the given value is smaller
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::replaceIfSmallerProbability(string expr, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::replaceIfSmallerProbability(1) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	vector<int> stateList = convExprToStateList(expr);
	return replaceIfSmallerProbability(stateList, value);

}

/** Given an ordered vector of states, replace the probability if the given value is greater
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::replaceIfSmallerProbability(const vector<int> stateList, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::replaceIfSmallerProbability(2) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	// Finds the index in the data array
	int dataIndex = dataIndexOf(stateList);
	
	if (dataIndex == -1) {
		return -1.0;
	} 
	
	// Checks the value at the data index with the given value, replacing it if the value is greater
	if ((value < mData[dataIndex] || mData[dataIndex] < 0.0) && value >= 0.0) {
		mData[dataIndex] = value;
	}
	
	return mData[dataIndex];

}

/** Given the index of the distribution, replace the probability if the given value is smaller
* On error, a -1.0 will be returned.
**/
double DiscreteDistribution::replaceIfSmallerProbability(int index, double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::replaceIfSmallerProbability(3) - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}	

	// Checks if the index is out of bounds
	if (index < 0 || index >= mDataSize) {
		return -1.0;
	} 
	
	// Checks the value at the data index with the given value, replacing it if the value is greater
	if (value < mData[index] && value >= 0.0) {
		mData[index] = value;
	}
	
	return mData[index];
	
}
	
/** Marginalize the distribution table along the given variables
* For example, if the distribution is joint between x1, x2, x3, then a marginalization along x3
* will return a new distribution between x1 and x2 only.
**/
DiscreteDistribution* DiscreteDistribution::marginalize(const vector<string> variableList) const {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::marginalize(1) - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}	

	///////////////////////////////////////////////////////////////////////////////
	// Process the variable list. It will check for the following things
	// 1) The variables is less than the dimension (one cannot marginalize over ALL the variables
	// 2) The variables are valid and they are not repeated.
	// Then it will fill up a vector of state size for each variable
	///////////////////////////////////////////////////////////////////////////////
	
	// 1) Check for size
	if (variableList.size() >= mDimension) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::marginalize(1) - variable list more than or equals dimension size" << endl;
		#endif
		return 0;
	}

	int index;
	bool* bArray = new bool[mDimension];		// boolean array for checking of repeated variables.
	for (int i = 0; i < mDimension; i++) {
		bArray[i] = false;
	}

	// 2) Check for valid variables and repeated variables
	for (int i = 0;  i < variableList.size(); i++) {
		index = indexOf(variableList.at(i));
		if (index == -1) {
			#ifdef DISCRETEDISTRIBUTION_DEBUG 
				cout << "[Error] DiscreteDistribution::marginalize(1) - Variable " << variableList.at(i) << " does not exist" << endl;
			#endif
			delete[] bArray;
			return 0;
		}
		// Checks if the corresponding location in the boolean array is true. If it is, then the variable
		// is repeated and an error occurs
		if (bArray[index]) {
			#ifdef DISCRETEDISTRIBUTION_DEBUG 
				cout << "[Error] DiscreteDistribution::marginalize(1) - Variable " << variableList.at(i) << " is repeated" << endl;
			#endif
			delete[] bArray;
			return 0;
		} else {
			bArray[index] = true;
		}
	}
	
	// Creates a new discrete distribution and a vector with the state size of the variables to be marginalized over
	// If the boolean array is false, it means that the variable is not being marginalized over and hence it 
	// is added to the new distribution. Note that the order of the remaining variables is the same for the
	// old distribution and the new distribution. This is important later for performing the actual
	// marginalization
	DiscreteDistribution* newdist = new DiscreteDistribution();
	for (int i = 0; i < mDimension; i++) {
		if (!bArray[i]) {
			newdist->addVariable(mVariableList.at(i), mSizeList.at(i));
		}
	}
	
	// Initialize the new distribution to allocate space
	newdist->initializeTable();

	// Compute the states based on the index, then pushes the ones that are true
	// into a state list, which will be used for incrementing the new probabilities
	vector<int> newStateList;
	int quotient;
	int remainder;
	for (int i = 0; i < mDataSize; i++) {
		remainder = i;
		for (int j = 0; j < mDimension; j++) {
			quotient = remainder / mOffsetList.at(j);
			remainder = remainder % mOffsetList.at(j);
			// If the current variable is not being marginalized over, put it's state in the statelist
			if (!bArray[j]) {
				newStateList.push_back(quotient);
			}
		}
		// Increment the entry in the new distribution table and clear the state list
		if (newdist->incProbability(newStateList, mData[i]) < 0.0) {
			#ifdef DISCRETEDISTRIBUTION_DEBUG 
				cout << "[Error] DiscreteDistribution::marginalize(1) - Error performing marginalization" << endl;
			#endif
			delete[] bArray;
			delete newdist;
			return 0;
		}
		newStateList.clear();
	}
	
	// No error, delete boolean array and return the distribution table
	delete[] bArray;
	return newdist;
}

/** Marginalize the distribution table along variables given in the string expression
* This method parses the given expression into a vector of variable names. Then it will
* call the first marginalize method which takes a vector of variables as its parameter
**/
DiscreteDistribution* DiscreteDistribution::marginalize(string expr) const {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::marginalize(2) - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}	
	
	// Parses the expression into a vector of variables
	vector<string> variableList;
	string variable;
	istringstream in(expr);
	while (!in.eof()) {
		in >> variable;
		variableList.push_back(variable);
	}
	
	// Calls the first marginalize function
	return marginalize(variableList);
	
}

/** Maximize the distribution table along the given variables
* For example, if the distribution is joint between x1, x2, x3, then a maximization along x3
* will return a new distribution between x1 and x2 only.
**/
DiscreteDistribution* DiscreteDistribution::maximize(const vector<string> variableList) const {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::maximize(1) - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}	

	///////////////////////////////////////////////////////////////////////////////
	// Process the variable list. It will check for the following things
	// 1) The variables is less than the dimension (one cannot maximize over ALL the variables
	// 2) The variables are valid and they are not repeated.
	// Then it will fill up a vector of state size for each variable
	///////////////////////////////////////////////////////////////////////////////

	// 1) Check for size
	if (variableList.size() >= mDimension) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::maximize(1) - variable list more than or equals dimension size" << endl;
		#endif
		return 0;
	}

	int index;
	bool* bArray = new bool[mDimension];		// boolean array for checking of repeated variables.
	for (int i = 0; i < mDimension; i++) {
		bArray[i] = false;
	}

	// 2) Check for valid variables and repeated variables
	for (int i = 0;  i < variableList.size(); i++) {
		index = indexOf(variableList.at(i));
		if (index == -1) {
			#ifdef DISCRETEDISTRIBUTION_DEBUG 
				cout << "[Error] DiscreteDistribution::maximize(1) - Variable " << variableList.at(i) << " does not exist" << endl;
			#endif
			delete[] bArray;
			return 0;
		}
		// Checks if the corresponding location in the boolean array is true. If it is, then the variable
		// is repeated and an error occurs
		if (bArray[index]) {
			#ifdef DISCRETEDISTRIBUTION_DEBUG 
				cout << "[Error] DiscreteDistribution::maximize(1) - Variable " << variableList.at(i) << " is repeated" << endl;
			#endif
			delete[] bArray;
			return 0;
		} else {
			bArray[index] = true;
		}
	}

	// Creates a new discrete distribution and a vector with the state size of the variables to be maximize over
	// If the boolean array is false, it means that the variable is not being maximized over and hence it 
	// is added to the new distribution. Note that the order of the remaining variables is the same for the
	// old distribution and the new distribution. This is important later for performing the actual
	// maximization
	DiscreteDistribution* newdist = new DiscreteDistribution();
	for (int i = 0; i < mDimension; i++) {
		if (!bArray[i]) {
			newdist->addVariable(mVariableList.at(i), mSizeList.at(i));
		}
	}

	// Initialize the new distribution to allocate space
	newdist->initializeTable();
	
	// Compute the states based on the index, then pushes the ones that are true
	// into a state list, which will be used for incrementing the new probabilities
	vector<int> newStateList;
	int quotient;
	int remainder;
	for (int i = 0; i < mDataSize; i++) {
		remainder = i;
		for (int j = 0; j < mDimension; j++) {
			quotient = remainder / mOffsetList.at(j);
			remainder = remainder % mOffsetList.at(j);
			// If the current variable is not being maximized over, put it's state in the statelist
			if (!bArray[j]) {
				newStateList.push_back(quotient);
			}
		}
		// Increment the entry in the new distribution table and clear the state list
		if (newdist->replaceIfGreaterProbability(newStateList, mData[i]) < 0.0) {
			#ifdef DISCRETEDISTRIBUTION_DEBUG 
				cout << "[Error] DiscreteDistribution::maximize(1) - Error performing maximization" << endl;
			#endif
			delete[] bArray;
			delete newdist;
			return 0;
		}
		newStateList.clear();
	}
	
	// No error, delete boolean array and return the distribution table
	delete[] bArray;
	return newdist;

}

/** Maximize the distribution table along the variables given in the string expression
* This method parses the given expression into a vector of variable names. Then it will 
* call the first maximize method which takes a vector of variables as its parameters
**/
DiscreteDistribution* DiscreteDistribution::maximize(string expr) const {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::maximize(2) - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}	
	
	// Parses the expression into a vector of variables
	vector<string> variableList;
	string variable;
	istringstream in(expr);
	while (!in.eof()) {
		in >> variable;
		variableList.push_back(variable);
	}
	
	// Calls the first marginalize function
	return maximize(variableList);

}

/** Computes the difference given another distribution. 
* The difference is the average difference between all the entries and that
* the tables must be compatible. It returns a -1.0 on error
* Note that for this method the order of the variable matters
**/
double DiscreteDistribution::difference(const DiscreteDistribution* distribution) {

	// Checks for existence of given table and whether it is initialized
	if (distribution == 0 || !distribution->mInitialized || !mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::difference() - Distribution is not valid" << endl;
		#endif
		return -1.0;
	}

	// Checks for self difference
	if (distribution == this) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Info] DiscreteDistribution::difference() - Computing difference with same instance" << endl;
		#endif
		return 0.0;
	}	
	
	// Checks for compatibility (even the variable order matters)
	bool isCompatible = true;
	for (int i = 0; i < mDimension; i++) {
		if (mVariableList.at(i) != distribution->mVariableList.at(i) ||
		    mSizeList.at(i) != distribution->mSizeList.at(i)) {
			isCompatible = false;
			break;
		}
	}
	if (!isCompatible) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Info] DiscreteDistribution::difference() - Distribution is not compatible" << endl;
		#endif
		return -1.0;
	}
	
	// Computes the difference
	double difference = 0.0;
	for (int i = 0; i < mDataSize; i++) {
		if (mData[i] > distribution->mData[i]) {
			difference += (mData[i] - distribution->mData[i]);
		} else {
			difference += (distribution->mData[i] - mData[i]);
		}
	}
	
	// difference = difference / mDataSize;
	
	return difference;
}

/** Does a multiplication of the distribution along their common variables
* If the normalize argument is true, the distribution will be normalized after multiplication.
* Note: The multiplication of the distributions might not be exactly correct. Will check out more and update
* it at a later date. Now it should suffice for our purposes.
**/
bool DiscreteDistribution::multiply(const DiscreteDistribution* table, bool normalize) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::multiply() - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}	

	// Checks that the given table is valid and that it is initialized as well
	if (table == 0 || !table->mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::multiply() - Invalid distribution" << endl;
		#endif
		return false;
	}
	
	// Checks for compatibility, i.e. 
	// table->variableList \subset variableList, and for the same variable, the state size must match
	// At the same time, it does a mapping to the variable map. The variable map is an array which is the size of the
	// dimension list. So for example if the variable list for this distribution is [x1, x2, x3, x4] and the
	// variable list for the given distribution is [x4, x2], then the variable map will contain
	// [ -1, 1, -1, 0]. This will indicate later on the position of the given stateList we are going to occupy when looking for the
	// appropiate value from the given distribution to multiply
	int index;
	int* varMap = new int[mDimension];
	vector<int> givenStateList;
	for (int i = 0; i < mDimension; i++) {
		varMap[i] = -1;
	}
	for (int i = 0; i < table->mDimension; i++) {
		index = indexOf(table->mVariableList.at(i));
		if (index == -1 || mSizeList.at(index) != table->mSizeList.at(i)) {
			#ifdef DISCRETEDISTRIBUTION_DEBUG 
				cout << "[Error] DiscreteDistribution::multiply() - Incompatible distributions, " << 
					    mVariableList.at(i) << " does not match" << endl;
			#endif
			delete[] varMap;
			return false;
		}
		varMap[index] = i;
		givenStateList.push_back(0);
	}
	
	// Initialize the current statelist for computation of index
	vector<int> currStateList;
	for (int i = 0; i < mDimension; i++) {
		currStateList.push_back(0);
	}
	
	// Performs the multiplication. 
	int quotient;
	int remainder;
	for (int i = 0; i < mDataSize; i++) {
		remainder = i;
		for (int j = 0; j < mDimension; j++) {
			quotient = remainder / mOffsetList.at(j);
			remainder = remainder % mOffsetList.at(j);
			// If the current variable is not being marginalized over, put it's state in the statelist
			if (varMap[j] != -1) {
				givenStateList[varMap[j]] = quotient;
			}
			currStateList[j] = quotient;
		}
		// Multiple the entry in the currStateList with the entry found in the givenStateList
		if (multProbability(currStateList, table->getProbability(givenStateList)) < 0.0) {
			#ifdef DISCRETEDISTRIBUTION_DEBUG 
				cout << "[Fatal Error] DiscreteDistribution::multiply() - Error performing multiplication" << endl;
			#endif
			delete[] varMap;
			return false;
		}
	}

	delete[] varMap;

	// Checks if need to normalize
	if (normalize) {
		return this->normalize();
	} else {
		return true;
	}
	
}

/** Normalize the distribution table such that the sum of all the data is 1.0
* Note that one feature is that the last entry might be larger a bit 
* in order to make the sum equal to 1.0
**/
bool DiscreteDistribution::normalize() {
	
	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::normalize() - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}	
	
	double total = 0.0;
	double currTotal = 0.0;
	for (int i = 0; i < mDataSize; i++) {
		total += mData[i];
	}
	if (total > 0.0) {
		for (int i = 0; i < mDataSize - 1; i++) {
			mData[i] = mData[i] / total;
			currTotal += mData[i];
		}
		mData[mDataSize - 1] = 1.0 - currTotal;
	} else {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Info] DiscreteDistribution::normalize() - Distribution is 0" << endl;
		#endif
		// Do nothing as the distribution is 0
		return false;
	}
	
	return true;
}

/** Discards all probability information and make the distribution uniform
* Note that one feature is that the last entry might be larger than the rest
* in order to make all their sums equals to 1.0
**/
bool DiscreteDistribution::makeUniform() {
	
	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::makeUniform() - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}	

	double currTotal = 0.0;
	double entry = 1.0 / mDataSize;
	for (int i = 0; i < mDataSize - 1; i++) {
		mData[i] = entry;
		currTotal += entry;
	}
	mData[mDataSize - 1] = 1.0 - currTotal;
	
	return true;
}

/** Randomly sets values for the distribution
* Note: Current this method uses the random number generator by ansi-C. This will be
* changed in future versions.
**/
bool DiscreteDistribution::makeRandom() {
	
	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::makeRandom() - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}
	
	// Initiate random number generator
	srand(time(NULL));
	for (int i = 0; i < mDataSize; i++) {
		mData[i] = (double)rand();
	}
	normalize();
	
	return true;
}

/** Sets all the values of the distribution to one
**/
bool DiscreteDistribution::makeOne() {
	
	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::makeOne() - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}
	
	for (int i = 0; i < mDataSize; i++) {
		mData[i] = 1.0;
	}
	
	return true;
}

/** Replaces the contents of the old distribution by the new distribution
* It has to check for compatibility first, which includes the ordering of the variables
**/
bool DiscreteDistribution::replace(const DiscreteDistribution& src) {
	
	// Checks for self replace
	if (this == &src) {
		return true;
	}
	
	// Checks for initialized distributions
	if (!mInitialized || !src.mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::replace() - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}
	
	// Checks for compatibility
	if (mDimension != src.mDimension) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::replace() - Incompatible distribution" << endl;
		#endif
		return false;
	}
	for (int index = 0; index < mDimension; index++) {
		if (mVariableList.at(index) != src.mVariableList.at(index) ||
		    mSizeList.at(index) != src.mSizeList.at(index)) {
			#ifdef DISCRETEDISTRIBUTION_DEBUG 
				cout << "[Error] DiscreteDistribution::replace() - Incompatible distribution" << endl;
			#endif
		return false;
		}
	}
	
	// Replace the old data array with the new one
	for (int i = 0; i < mDataSize; i++) {
		mData[i] = src.mData[i];
	}
	
	return true;
	
}

/** Replaces the contents of the old distribution by the new distribution
* The only check is the initialize and the datasize. Hence this method should only
* be used when one is sure that the src is compatible with the current distribution.
**/
bool DiscreteDistribution::replaceNoCheck(const DiscreteDistribution& src) {

	// Checks for initialized distributions
	if (!mInitialized || !src.mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::replaceNoCheck() - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}

	// Checks for same data size
	if (mDataSize != src.mDataSize) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::replaceNoCheck() - Incompatible distribution" << endl;
		#endif
		return false;
	}
	
	// Replace the old data array with the new one
	for (int i = 0; i < mDataSize; i++) {
		mData[i] = src.mData[i];
	}
	
	return true;
}

/** Returns the maximum value in the table
* Returns a -1.0 if the table has not been initialized
**/
double DiscreteDistribution::maximum() const {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::maximum() - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}

	// Loops through the data array and picks up the maximum entry
	double max = mData[0];
	for (int i = 1; i < mDataSize; i++) {
		if (max < mData[i]) {
			max = mData[i];
		}
	}
	
	return max;

}

/** Returns the minimum value in the table
* Returns a -1.0 if the table has not been initialized
**/
double DiscreteDistribution::minimum() const {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::minimum() - Distribution is not initialized yet" << endl;
		#endif
		return -1.0;
	}

	// Loops through the data array and picks up the maximum entry
	double min = mData[0];
	for (int i = 1; i < mDataSize; i++) {
		if (min > mData[i]) {
			min = mData[i];
		}
	}
	
	return min;

}

/** Multiplies the table throughout by a given value
**/
bool DiscreteDistribution::multiply(double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::multiply() - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}

	if (value < 0.0) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::multiply() - Value is negative (" value << ")" << endl;
		#endif
		return false;
	}
	
	// Loops through the data array and multiplies them with the value. The value must be positive
	for (int i = 0; i < mDataSize; i++) {
		mData[i] = mData[i] * value;
	}
	
	return true;

}

/** Sets all the entries of the distribution table to the given value
* No checks are made to the value
**/
bool DiscreteDistribution::setValue(double value) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::setValue() - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}	

	// Loops through the data array and multiplies them with the value. The value can be any number
	for (int i = 0; i < mDataSize; i++) {
		mData[i] = value;
	}
	
	return true;
}

/** Replaces any occurances of the from value to the to value
**/
bool DiscreteDistribution::replaceValue(double fromValue, double toValue) {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::replaceValue() - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}	

	// Loops through the data array and multiplies them with the value. The value can be any number
	for (int i = 0; i < mDataSize; i++) {
		if (mData[i] == fromValue) 
			mData[i] = toValue;
	}
	
	return true;

}

/** Multiplies the entire data array by the exponential function
**/
bool DiscreteDistribution::exponent() {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::exponent() - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}

	// Loops through the data array and apply them as powers to the exponential function
	for (int i = 0; i < mDataSize; i++) {
		mData[i] = exp(mData[i]);
	}
	
	return true;

}

/** Inverse all the entries in the distribution table
**/
bool DiscreteDistribution::inverse() {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::inverse() - Distribution is not initialized yet" << endl;
		#endif
		return false;
	}

	// Loops through the data array and invert them
	for (int i = 0; i < mDataSize; i++) {
		mData[i] = 1.0 / mData[i];
	}
	
	return true;
	
}
	
/** Gets the index of the maximum value
* On error, a -1 is returned.
**/
int DiscreteDistribution::maximumIndex() const {

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::maximumIndex() - Distribution is not initialized yet" << endl;
		#endif
		return -1;
	}	

	// Loops through the entire list and notes the index of the maximal value
	int maxIndex = 0;
	double max = mData[maxIndex];
	for (int i = 1; i < mDataSize; i++) {
		if (max < mData[i]) {
			max = mData[i];
			maxIndex = i;
		}
	}
	
	return maxIndex;
}

/** Returns the list of variables
**/
vector<string> DiscreteDistribution::getVariableList() const {
	// Returns a copy of the variable list
	return mVariableList;
}
	
/** Given a vector of unordered <variable, state> pair, find the index in the data array
**/
int DiscreteDistribution::dataIndexOf(const vector<pair> pairList) const {

	// Checks the dimension of the pairList
	if (pairList.size() != mDimension) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::dataIndexOf(1) - Array dimension doesn't match [" << 
				    mDimension << ", " << pairList.size() << "]" <<endl;
		#endif
		return -1;
	}
	
	// Checks the index of the size and does the computation of the index location in the data array
	// Assuming three sizes - size1, size2, size3, and three indices - index1, index2, index3
	// The formula is computed from the following equation
	// index = index1 * (size2 * size3) + index2 * (size3) + index3 * 1
	// The products of sizes are provided by the offsetList
	int dataIndex = 0;
	int varIndex;
	string variable;
	int state;
	for (int i = 0; i < pairList.size(); i++) {
		variable = pairList.at(i).variable;
		state = pairList.at(i).state;
		
		// Checks the index of the variable, i.e. it exists and
		// whether the size is correct
		varIndex = indexOf(variable);
		if (varIndex == -1 || state < 0 || state >= mSizeList.at(varIndex)) {
			#ifdef DISCRETEDISTRIBUTION_DEBUG 
				cout << "[Error] DiscreteDistribution::dataIndexOf(1) - Pair <" << 
					    variable << ", " << state << "> does not exist" << endl;
			#endif
			return -1;
		}
		
		// Adds on to the index
		dataIndex = dataIndex + (state * (mOffsetList.at(varIndex)));
	}

	return dataIndex;
}

/** Given a vector of ordered state list, find the index in the data array
**/
int DiscreteDistribution::dataIndexOf(const vector<int> stateList) const {
	
	// Checks the dimension of the stateList
	if (stateList.size() != mDimension) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::dataIndexOf(2) - Array dimension doesn't match [" << 
				    mDimension << ", " << stateList.size() << "]" <<endl;
		#endif
		return -1;
	}
	
	// Compute the dataIndex, (and checking if the state is within limits)
	int dataIndex = 0;
	int state;
	for (int index = 0; index < stateList.size(); index++) {
		state = stateList.at(index);
		if (state < 0 || state >= mSizeList.at(index)) {
			#ifdef DISCRETEDISTRIBUTION_DEBUG 
				cout << "[Error] DiscreteDistribution::dataIndexOf(2) - State <" << 
					    mVariableList.at(index) << ", " << state << "> does not exist" << endl;
			#endif		
			return -1;
		}
		dataIndex = dataIndex + (state * (mOffsetList.at(index)));
	}
	
	return dataIndex;
}

/** Converts a string expression to a stateList.
* The string should be in the following format "x1 3 x2 3 x2 1" etc, i.e. variable followed by state
**/
vector<int> DiscreteDistribution::convExprToStateList(string expr) const {
	vector<int> stateList;
	
	istringstream in(expr);
	int varIndex;
	string variable;
	int state;

	// Initialize the size of the state list and initialized it with -1
	for (int i = 0; i < mDimension; i++) {
		stateList.push_back(-1);
	}
	
	while (!in.eof()) {

		in >> variable;
		in >> state;
		
		// Checks if there is any error in conversion
		if (in.fail()) {
			#ifdef DISCRETEDISTRIBUTION_DEBUG 
				cout << "[Error] DiscreteDistribution::convExprToStateList() - Expression error : " << expr << endl; 
			#endif
			stateList.clear();
			return stateList;
		}

		// Checks if the variable and states are valid
		varIndex = indexOf(variable);
		if (varIndex == -1 || state < 0 || state >= mSizeList.at(varIndex)) {
			#ifdef DISCRETEDISTRIBUTION_DEBUG 
				cout << "[Error] DiscreteDistribution::convExprToStateList() - Pair <" << 
					    variable << ", " << state << "> does not exist" << endl;
			#endif
			stateList.clear();
			return stateList;
		}
		
		// Checks if the variable has not yet been set
		if (stateList.at(varIndex) != -1)  {
			#ifdef DISCRETEDISTRIBUTION_DEBUG 
				cout << "[Error] DiscreteDistribution::convExprToStateList() - Repeated variable : " << variable << endl; 
			#endif
			stateList.clear();
			return stateList;
		}
		
		// Sets the state
		stateList[varIndex] = state;
	}
	
	// Checks if all the states have been set
	for (int i = 0; i < stateList.size(); i++) {
		if (stateList.at(i) == -1) {
			#ifdef DISCRETEDISTRIBUTION_DEBUG 
				cout << "[Error] DiscreteDistribution::convExprToStateList() - Some variables not set" << endl;
			#endif
			stateList.clear();
			return stateList;
		}
	}
	
	return stateList;
}

/** Converts a data index into a stateList.
** On error, returns an empty state list
**/
vector<int> DiscreteDistribution::convIndexToStateList(int index) const {
	vector<int> stateList;

	// Checks if the distribution table is initialized
	if (!mInitialized) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::convIndexToStateList() - Distribution is not initialized yet" << endl;
		#endif
		return stateList;
	}	
	
	// Checks that the index is not out of bounds
	if (index >= mDataSize) {
		#ifdef DISCRETEDISTRIBUTION_DEBUG 
			cout << "[Error] DiscreteDistribution::convIndexToStateList() - Index " << index << " out of bounds" << endl; 
		#endif
		return stateList;
	}

	// Progressively divides the index by the offset list and assigns the remainder to the currIndex
	int quotient;
	int remainder = index;
	for (int i = 0; i < mDimension; i++) {
		quotient = remainder / mOffsetList.at(i);
		remainder = remainder % mOffsetList.at(i);
		stateList.push_back(quotient);
	}
	
	return stateList;
}
	
/** Writes the contents of the table to a file
**/
bool DiscreteDistribution::writeFile(string fileName) const {
	fstream outputStream;
	outputStream.open(fileName.c_str(), ios::out);
	if (outputStream.is_open()) {
		outputStream << toStringDetail();
		outputStream.close();
		return true;
	} else {
		return false;
	}
}

/** Returns a string representation of the table
**/
string DiscreteDistribution::toString() const {
	ostringstream str;
	
	str << "---DiscreteDistribution---" << endl;
	str << "mInitialized: " << mInitialized << endl;
	str << "mDimension: " << mDimension << endl;
	str << "mVariableList: ";
	for (int index = 0; index < mVariableList.size(); index++) {
		str << "(" << mVariableList.at(index) << ", " << mSizeList.at(index) << ") "; 
	}
	str << endl;
	str << "mOffsetList: ";
	for (int index = 0; index < mOffsetList.size(); index++) {
		str << mOffsetList.at(index) << " ";
	}
	str << endl;
	str << "mDataSize: " << mDataSize << endl;
	
	return str.str();
}

/** Returns a detailed description of the destribution, including its table contents
**/
string DiscreteDistribution::toStringDetail() const {
	ostringstream str;
	
	str << "---DiscreteDistribution---" << endl;
	str << "mInitialized: " << mInitialized << endl;
	str << "mDimension: " << mDimension << endl;
	str << "mVariableList: ";
	for (int index = 0; index < mVariableList.size(); index++) {
		str << "(" << mVariableList.at(index) << ", " << mSizeList.at(index) << ") "; 
	}
	str << endl;
	str << "mOffsetList: ";
	for (int index = 0; index < mOffsetList.size(); index++) {
		str << mOffsetList.at(index) << " ";
	}
	str << endl;
	str << "mDataSize: " << mDataSize << endl;
	str << "Data: " << endl;
	for (int index = 0; index < mDataSize; index++) {
		str << mData[index] << " ";
	}
	str << endl;
	
	return str.str();

}

/** Returns a concise representation of its table contents
**/
string DiscreteDistribution::toStringSimple() const {
	ostringstream str;
	
	str << "---DiscreteDistribution--- " ;
	for (int index = 0; index < mVariableList.size(); index++) {
		str << "(" << mVariableList.at(index) << ", " << mSizeList.at(index) << ") "; 
	}
	str << endl;
	str << "Data: ";
	for (int index = 0; index < mDataSize; index++) {
		str << mData[index] << " ";
	}
	str << endl;
	
	return str.str();

}

/** Returns a string representation of only its data
**/
string DiscreteDistribution::toStringData() const {
	ostringstream str;
		for (int index = 0; index < mDataSize; index++) {
		str << mData[index] << " ";
	}
	str << endl;
	return str.str();
}

/** Helper function for copying contents
* It copies everything but assumes that allocated memory (if any)
* has already been deleted.
**/
void DiscreteDistribution::copyFrom(const DiscreteDistribution& src) {
	mInitialized = src.mInitialized;
	mDimension = src.mDimension;
	mDataSize = src.mDataSize;
	
	// Copies the variable, size and offset lists
	for (int index = 0; index < src.mVariableList.size(); index++) {
		mVariableList.push_back(src.mVariableList.at(index));
		mSizeList.push_back(src.mSizeList.at(index));
		mOffsetList.push_back(src.mOffsetList.at(index));
	}
	
	// Copies the data
	if (mDataSize > 0) {
		mData = new double[mDataSize];
		for (int index = 0; index < mDataSize; index++) {
			mData[index] = src.mData[index];
		}
	} else {
		mData = 0;
	}
}

}

