/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: discretedistribution.h
**  
** Description:
** The header file storing all the specification for a discrete distribution table. 
**  
** Last modified: 
** 16 March 2007
** 
** Comments:
** 1) For the implementation of this distribution table, the entries are implemented as an array
**     The indices for the various entries will have to be computed.
** 2) Note: The multiplication of the distributions might not be flexible enough, in the sense that
**      one of them must be a subset of the other, when, I think, it is not necessary.
** 3) I've also added another 2 functions in - replace if greater and replace if smaller
** 4) Added more functionalities to access the entries by index, i.e. for greater optimization later 
**     on
** 5) Added function to find the index for the best entry
****************************************************************************/

#ifndef DISCRETEDISTRIBUTION_H_
#define DISCRETEDISTRIBUTION_H_

/* For debugging. Comment out to suppress debugging lines */
// #define DISCRETEDISTRIBUTION_DEBUG

/* Standard Includes */
#include <iostream>
#include <vector>

using namespace std;

namespace distribution {

struct pair {
	string variable;
	int state;
};

bool isValidName(string varName);

class DiscreteDistribution {

public:

	/* Constructors and destructors */
	DiscreteDistribution();
	DiscreteDistribution(const DiscreteDistribution& src);	
	DiscreteDistribution& operator=(const DiscreteDistribution& src);
	virtual ~DiscreteDistribution(); 
	
	/* Methods to add variables and remove variables */
	bool addVariable(string variable, int stateSize);
	bool removeVariable(string variable);
	bool hasVariable(string variable) const;
	int indexOf(string variable) const;
	bool initializeTable();
	bool isInitialized() const;
	int getSize(string variable) const;
	int getDimension() const;
	int getDataSize() const;
	
	/* Methods to get and set the probability values */
	double getProbability(const vector<pair> pairList) const;
	double setProbability(const vector<pair> pairList, double value);
	double incProbability(const vector<pair> pairList, double value);
	double multProbability(const vector<pair> pairList, double value);
	double getProbability(string expr) const;
	double setProbability(string expr, double value); 
	double incProbability(string expr, double value);
	double multProbability(string expr, double value);
	double getProbability(const vector<int> stateList) const;
	double setProbability(const vector<int> stateList, double value);
	double incProbability(const vector<int> stateList, double value);
	double multProbability(const vector<int> stateList, double value);
	
	double getProbability(int index) const;
	double setProbability(int index, double value);
	double incProbability(int index, double value);
	double multProbability(int index, double value);
		
	double replaceIfGreaterProbability(string expr, double value);
	double replaceIfGreaterProbability(const vector<int> stateList, double value);
	double replaceIfGreaterProbability(int index, double value);
	double replaceIfSmallerProbability(string expr, double value);
	double replaceIfSmallerProbability(const vector<int> stateList, double value);
	double replaceIfSmallerProbability(int index, double value);
	
	/* Operations over the distribution tables */
	DiscreteDistribution* marginalize(const vector<string> variableList) const;
	DiscreteDistribution* marginalize(string expr) const;	// Gives input as an expression, e.g. "A B C"
	DiscreteDistribution* maximize(const vector<string> variableList) const;
	DiscreteDistribution* maximize(string expr) const;	// Gives input as an expression, e.g. "A B C"
	double difference(const DiscreteDistribution* distribution);
	bool multiply(const DiscreteDistribution* table, bool normalize = true);
	bool normalize();
	bool makeUniform();
	bool makeRandom();
	bool makeOne();
	bool replace(const DiscreteDistribution& src);
	bool replaceNoCheck(const DiscreteDistribution& src);

	/* Other miscellaneous operations */
	double maximum() const;
	double minimum() const;
	bool multiply(double value);
	bool setValue(double value);
	bool replaceValue(double fromValue, double toValue);
	bool exponent();
	bool inverse();
	
	/* Printing and debugging */
	int maximumIndex() const;
	vector<string> getVariableList() const;
	int dataIndexOf(const vector<pair> pairList) const;
	int dataIndexOf(const vector<int> stateList) const;
	vector<int> convExprToStateList(string expr) const;
	vector<int> convIndexToStateList(int index) const;
	bool writeFile(string fileName) const;
	string toString() const;
	string toStringDetail() const;
	string toStringSimple() const;
	string toStringData() const;
	
protected:
	
	bool mInitialized;			// This is the boolean that controls the adding of variables to a distribution
								// table. When a table is created, there is no data. The user can add in the 
								// variables and the number of states per variable. But when they call
								// initializeTable(), space will be created and allocated to the table and from
								// there on, no more variables can be added.
	double* mData;
	int mDimension;
	int mDataSize;
	vector<string> mVariableList;
	vector<int> mSizeList;
	vector<int> mOffsetList;	// Used for the computation of index
	
	void copyFrom(const DiscreteDistribution& src);
	
};

}

#endif
