/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: discretedistribution.cpp
**  
** Description:
** The test file for all the discrete distribution related methods.
**  
** Last modified: 
** 25 October 2006
** 
** Comments:
**
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <vector>
#include <map>

/* Application Includes */
#include "discretedistribution_test.h"
#include "discretedistribution.h"

using namespace distribution;

/** Main function to call for the test of the discrete distribution table.
* It will call the rest of the test, such as basic test, test on operations such as normalization etc.
**/
void testDiscreteDistribution() {

	testDiscreteDistributionBasic();
	testDiscreteDistributionOperations();
	
}

void testDiscreteDistributionBasic() {

	cout << "==================================" << endl;
	cout << "In testDiscreteDistributionBasic()" << endl;
	cout << "==================================" << endl;

	DiscreteDistribution* ddist;
	
	cout << "Test 1 : Creation of table" << endl;
	ddist = new DiscreteDistribution();
	cout << ddist->toString() << endl;
	
	cout << "Test 2 : Add in variables" << endl;
	ddist->addVariable("x1", 5);
	ddist->addVariable("x2", 4);
	ddist->addVariable("x3", 3);
	ddist->addVariable("x4", 5);
	cout << ddist->toString() << endl;
	
	cout << "Test 3 : Remove variables" << endl;
	ddist->removeVariable("x3");
	ddist->removeVariable("x7");		// Supposed to fail
	cout << ddist->toString() << endl;
	
	cout << "Test 4 : Initializing distribution" << endl;
	ddist->initializeTable();
	ddist->addVariable("x6", 4);		// Supposed to fail
	ddist->removeVariable("x1");		// Supposed to fail
	cout << ddist->toString() << endl;
	
	cout << "Test 5 : Getting a value" << endl;
	vector<distribution::pair> pairList;
	distribution::pair x1; x1.variable = "x1"; x1.state = 3;
	distribution::pair x2; x2.variable = "x2"; x2.state = 2;
	distribution::pair x4; x4.variable = "x4"; x4.state = 1;
	pairList.push_back(x1); pairList.push_back(x2); pairList.push_back(x4);
	cout << "P(x1 = 3, x2 = 2, x4 = 1) = " << ddist->getProbability(pairList) << endl;
	cout << endl;
	
	cout << "Test 6 : Getting a value using string expression" << endl;
	cout << "P(x1 = 3, x2 = 2, x4 = 1) = " << ddist->getProbability("x1 3 x2 3 x4 1") << endl;
	cout << endl;
	
	cout << "Test 7 : Getting a value after making it uniform" << endl;
	ddist->makeUniform();
	cout << "P(x1 = 3, x2 = 2, x4 = 1) = " << ddist->getProbability("x1 3 x2 3 x4 1") << endl;
	cout << endl;
	
	cout << "Test 8 : Getting a value after randomizing it" << endl;
	ddist->makeRandom();
	cout << "P(x1 = 3, x2 = 2, x4 = 1) = " << ddist->getProbability("x1 3 x2 3 x4 1") << endl;
	cout << endl;
	
	delete ddist;
	
}

void testDiscreteDistributionOperations() {

	cout << "=======================================" << endl;
	cout << "In testDiscreteDistributionOperations()" << endl;
	cout << "=======================================" << endl;

	DiscreteDistribution* ddist = new DiscreteDistribution();
		
	cout << "Test 1 : Creating a univariate table" << endl;
	ddist->addVariable("x1", 4);
	ddist->initializeTable();
	ddist->setProbability("x1 0", 0.13);
	ddist->setProbability("x1 1", 0.43);
	ddist->setProbability("x1 2", 0.16);
	ddist->setProbability("x1 3", 0.40);
	cout << ddist->toStringDetail() << endl;
	cout << "Normalize distribution..." << endl;
	ddist->normalize();
	cout << ddist->toStringDetail() << endl;
	
	DiscreteDistribution* ddist_2 = new DiscreteDistribution();
	
	cout << "Test 2 : Creating a second univariate table" << endl;
	ddist_2->addVariable("x1", 4);
	ddist_2->initializeTable();
	ddist_2->setProbability("x1 0", 0.1);
	ddist_2->setProbability("x1 1", 0.1);
	ddist_2->setProbability("x1 2", 0.1);
	ddist_2->setProbability("x1 3", 0.7);
	cout << ddist_2->toStringDetail() << endl;
	
	cout << "Test 3 : Testing multiplication" << endl;
	ddist->multiply(ddist_2);
	cout << ddist->toStringDetail() << endl;
	
	cout << "Test 4 : Testing marginalization over a bivariate distribution" << endl;
	DiscreteDistribution* ddist_3 = new DiscreteDistribution();
	DiscreteDistribution* ddist_4;
	ddist_3->addVariable("x1", 3);
	ddist_3->addVariable("x2", 3);
	ddist_3->initializeTable();
	// Setting the values
	ddist_3->setProbability("x1 0 x2 0", 0.2);
	ddist_3->setProbability("x1 0 x2 1", 0.15);
	ddist_3->setProbability("x1 0 x2 2", 0.3);
	ddist_3->setProbability("x1 1 x2 0", 0.4);
	ddist_3->setProbability("x1 1 x2 1", 0.1);
	ddist_3->setProbability("x1 1 x2 2", 0.05);
	ddist_3->setProbability("x1 2 x2 0", 0.1);
	ddist_3->setProbability("x1 2 x2 1", 0.15);
	ddist_3->setProbability("x1 2 x2 2", 0.25);
	cout << ddist_3->toStringDetail() << endl;
	cout << "Normalize distribution..." << endl;
	ddist_3->normalize();
	cout << ddist_3->toStringDetail() << endl;
	cout << "Marginalize over x2" << endl;
	ddist_4 = ddist_3->marginalize("x2");
	cout << ddist_4->toStringDetail() << endl;
	
	cout << "Test 5 : Testing marginalization over a trivariate distribution" << endl;
	DiscreteDistribution* ddist_5 = new DiscreteDistribution();
	DiscreteDistribution* ddist_5b;
	ddist_5->addVariable("x1", 3);
	ddist_5->addVariable("x2", 3);
	ddist_5->addVariable("x3", 3);
	ddist_5->initializeTable();
	// Setting the values
	ddist_5->setProbability("x1 0 x2 0 x3 0", 0.02);
	ddist_5->setProbability("x1 0 x2 0 x3 1", 0.01);
	ddist_5->setProbability("x1 0 x2 0 x3 2", 0.02);
	ddist_5->setProbability("x1 0 x2 1 x3 0", 0.05);
	ddist_5->setProbability("x1 0 x2 1 x3 1", 0.01);
	ddist_5->setProbability("x1 0 x2 1 x3 2", 0.01);
	ddist_5->setProbability("x1 0 x2 2 x3 0", 0.05);
	ddist_5->setProbability("x1 0 x2 2 x3 1", 0.01);
	ddist_5->setProbability("x1 0 x2 2 x3 2", 0.03);
	ddist_5->setProbability("x1 1 x2 0 x3 0", 0.01);
	ddist_5->setProbability("x1 1 x2 0 x3 1", 0.2);
	ddist_5->setProbability("x1 1 x2 0 x3 2", 0.01);
	ddist_5->setProbability("x1 1 x2 1 x3 0", 0.2);
	ddist_5->setProbability("x1 1 x2 1 x3 1", 0.02);
	ddist_5->setProbability("x1 1 x2 1 x3 2", 0.01);
	ddist_5->setProbability("x1 1 x2 2 x3 0", 0.05);
	ddist_5->setProbability("x1 1 x2 2 x3 1", 0.01);
	ddist_5->setProbability("x1 1 x2 2 x3 2", 0.1);
	ddist_5->setProbability("x1 2 x2 0 x3 0", 0.02);
	ddist_5->setProbability("x1 2 x2 0 x3 1", 0.01);
	ddist_5->setProbability("x1 2 x2 0 x3 2", 0.01);
	ddist_5->setProbability("x1 2 x2 1 x3 0", 0.05);
	ddist_5->setProbability("x1 2 x2 1 x3 1", 0.01);
	ddist_5->setProbability("x1 2 x2 1 x3 2", 0.01);
	ddist_5->setProbability("x1 2 x2 2 x3 0", 0.05);
	ddist_5->setProbability("x1 2 x2 2 x3 1", 0.01);
	ddist_5->setProbability("x1 2 x2 2 x3 2", 0.01);
	ddist_5b = ddist_5->marginalize("x3 x1");
	cout << "After marginalization over x1 and x3..." <<  endl;
	cout << ddist_5b->toStringDetail() << endl;
	cout << "Multiplication with the marginalized distribution..." << endl;
	ddist_5->multiply(ddist_5b, false);
	cout << ddist_5->toStringDetail() << endl;
	
	delete ddist;
	delete ddist_2;
	delete ddist_3;
	delete ddist_4;
	delete ddist_5;
	delete ddist_5b;

}

