/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: factorgraph.cpp
**  
** Description:
** The source file containing the structure of the factor graphs, including its components
** such as the nodes.
**  
** Last modified: 
** 04 December 2006
** 
** Comments:
**
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <sstream>
#include <vector>
#include <map>

/* Application Includes */
#include "factorgraph.h"

using namespace std;

namespace factorgraph {

/** Checks if the id is valid
* Checks if a given Id is valid, according to standard variable rules.
* 1) It must not be empty
* 2) It must start with an alphabet or underscore "_"
* 3) It must not contain any invalid characters, such as ", ", "?", " " etc.
**/
bool isValidId(string id) {

	// Checks for empty string
	if (id == "") return false;
	
	char c;
	
	// Checks the first charactor
	c = id[0];
	if (c != '_' && !isalpha(c)) return false;
	
	// Checks the rest of the characters. The rest can only
	// be alphanumeric or '_'
	int length = strlen(id.c_str());
	for (int i = 1; i < length; i++) {
		c = id[i];
		if (c != '_' && !isalnum(c)) return false;
	}
	
	return true;
}

/** Default constructor
**/
FactorGraph::FactorGraph(string id, string label) : mLabel(label) {
	
	// Checks if the Id is valid. If it is valid then set it. Otherwise put an empty id
	if (isValidId(id)) {
		mId = id;
	} else {
		mId = "";
	}

	#ifdef FACTORGRAPH_DEBUG
		cout << "[Info] FactorGraph::FactorGraph() - Creating factorGraph:id=" << id << ",label=" << label << endl;
	#endif

}

/** Copy constructor
* Copies all the contents (By creating a new instance of them) from src to the new instance.
**/
FactorGraph::FactorGraph(const FactorGraph& src) {
	copyFrom(src);
}

/** Assignment constructor
**/
FactorGraph& FactorGraph::operator=(const FactorGraph& src) {
	
	// Checks for self assignment
	if (this == &src) {
		return (*this);	
	}
	
	// Clears the node list and connection list
	removeAll();
	
	// Copy contents
	copyFrom(src);
	
	return (*this);

}

/** Default Destructor
**/
FactorGraph::~FactorGraph() {
	
	#ifdef FACTORGRAPH_DEBUG
		cout << "[Info] FactorGraph::~FactorGraph() - Destroying FactorGraph:id=" << mId << ",label=" << mLabel << endl;
	#endif

	// Clears everything
	removeAll();
}

/** Checks if the Factor Graph is consistent
* Performs a series of checks, one of which is to check if it is bipartite, i.e. no two nodes
* of the same type are supposed to be connected together
**/
bool FactorGraph::isConsistent() const {
	// Todo
	return true;
}

/** Checks if the node with the Id exists in this factor graph
**/
bool FactorGraph::nodeExists(string id) const {
	map<string, FactorGraphNode*>::const_iterator iter = mNodeList.find(id);
	if (iter != mNodeList.end()) {
		return true;
	} else {
		return false;
	}
}

/** Returns the number of nodes
**/
int FactorGraph::nodeSize() const {
	return mNodeList.size();
}

/** Creates a default node and adds it to the factor graph
* The node will be default, i.e. its Id will be generated and its name will take
* its Id.
**/
FactorGraphNode* FactorGraph::addNode(nodeType type) {

	// Creates new node with name and Id being the same
	string id = generateNodeId(type);
	FactorGraphNode* node = new FactorGraphNode(id, id, type);
	
	// Adds the node in
	addNode(node);
	
	return node;
}

/** Creates a new node with the given Id, name and type and adds it to the factor graph
* If, however, the Id is not valid, a node WILL NOT be created at all.
**/
FactorGraphNode* FactorGraph::addNode(string id, string label, nodeType type) {
	
	// Checks for Id validity and whether it exists in the factor graph
	if (!isValidId(id) && nodeExists(id)) {
		return 0;
	}
	
	// Creates a new node with Id and label and type
	FactorGraphNode* node = new FactorGraphNode(id, label, type);
	
	// Adds the node in
	addNode(node);
	
	return node;
}

/** Adds a node to the factor graph
**/
FactorGraphNode* FactorGraph::addNode(FactorGraphNode* node) {

	// If checks if the node is already added
	if (node->getParent() == this) {
		return node;
	}
	
	// If the node has an earlier parent, remove it first
	if (node->getParent() != 0) {
		node->getParent()->removeNode(node);
	}
	
	// Checks if Id exists already in the factor graph. If it is, then do
	// not add the node in
	if (nodeExists(node->getId())) {
		return 0;
	} else {
		mNodeList[node->getId()] = node;
		node->mParent = this;
		return node;
	}
	
}

/** Finds the node with the given Id and returns it
**/
FactorGraphNode* FactorGraph::getNode(string id) {
	map<string, FactorGraphNode*>::iterator iter = mNodeList.find(id);
	if (iter == mNodeList.end()) {
		return 0;
	} else {
		return iter->second;
	}
}

/** Finds a node and remove it from the factor graph
**/
FactorGraphNode* FactorGraph::removeNode(string id) {
	map<string, FactorGraphNode*>::iterator iter = mNodeList.find(id);
	if (iter == mNodeList.end()) {
		return 0;
	} else {
		return removeNode(iter->second);
	}
}

/** Removes a node from the factor graph
**/
FactorGraphNode* FactorGraph::removeNode(FactorGraphNode* node) {
	
	// Checks if the parent is indeed the current instance
	if (node->getParent() != this) {
		return 0;
	}
	
	// Removes it from the connection list, i.e. any connection with this node will be
	// disconnected.
	multimap<string, string>::iterator iter = mConnectionList.begin();
	while (iter != mConnectionList.end()) {
		if (iter->first == node->getId() || iter->second == node->getId()) {
			mConnectionList.erase(iter++);
		} else {
			iter++;
		}
	}
	
	// Remove the node from node list
	mNodeList.erase(node->getId());
	
	// Remove the parent and returns the node
	node->mParent = 0;
	return node;
	
}

/** Removes the node from a factor graph and deletes it off
**/
bool FactorGraph::deleteNode(string id) {
	
	FactorGraphNode* node = removeNode(id);
	
	if (node != 0) {
		delete node;
		return true;
	} else {
		return false;
	}
}

/** Returns the vector of nodes
**/
vector<FactorGraphNode*> FactorGraph::getNodes() {
	vector<FactorGraphNode*> nodeList;
	map<string, FactorGraphNode*>::iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		nodeList.push_back(iter->second);
		iter++;
	}
	return nodeList;
}

/** Returns the vector of node Id
**/
vector<string> FactorGraph::getNodesId() const {
	vector<string> idList;
	map<string, FactorGraphNode*>::const_iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		idList.push_back(iter->first);
		iter++;
	}
	return idList;
}

/** Returns the vector of nodes of type: Factor
**/
vector<FactorGraphNode*> FactorGraph::getFactors() {
	vector<FactorGraphNode*> nodeList;
	map<string, FactorGraphNode*>::iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		if (iter->second->getType() == nodeType_factor) {
			nodeList.push_back(iter->second);
		}
		iter++;
	}
	return nodeList;
}

/** Returns a vector of node Id for the factor nodes
**/
vector<string> FactorGraph::getFactorsId() const {
	vector<string> idList;
	map<string, FactorGraphNode*>::const_iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		if (iter->second->getType() == nodeType_factor) {
			idList.push_back(iter->first);
		}
		iter++;
	}
	return idList;
}

/** Returns the vector of nodes of type: Variable
**/
vector<FactorGraphNode*> FactorGraph::getVariables() {
	vector<FactorGraphNode*> nodeList;
	map<string, FactorGraphNode*>::iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		if (iter->second->getType() == nodeType_variable) {
			nodeList.push_back(iter->second);
		}
		iter++;
	}
	return nodeList;
}

/** Returns a vector of node Id for the variable nodes
**/
vector<string> FactorGraph::getVariablesId() const {
	vector<string> idList;
	map<string, FactorGraphNode*>::const_iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		if (iter->second->getType() == nodeType_variable) {
			idList.push_back(iter->first);
		}
		iter++;
	}
	return idList;
}

/** Returns the neighbors of the given node in a const vector
**/
vector<FactorGraphNode*> FactorGraph::getNeighbors(FactorGraphNode* node) {
	vector<FactorGraphNode*> neighborList;
	
	// Checks if the node belongs to the factor graph
	if (node->getParent() != this || !nodeExists(node->getId())) {
		return neighborList;
	} 
	
	multimap<string, string>::const_iterator startIter, endIter;
	startIter = mConnectionList.lower_bound(node->getId());
	endIter = mConnectionList.upper_bound(node->getId());
	
	for (startIter; startIter != endIter; startIter++) {
		neighborList.push_back(getNode(startIter->second));
	}
	
	return neighborList;
}

/** Return the neighbors of the node of the given Id
**/
vector<FactorGraphNode*> FactorGraph::getNeighbors(string id) {

	// Finds the node using the Id
	FactorGraphNode* node = getNode(id);
	
	if (node == 0) {
		vector<FactorGraphNode*> neighborList;
		return neighborList;
	} else {
		return getNeighbors(node);
	}
}

/** Return the Ids of the neighbours of the given node
**/
vector<string> FactorGraph::getNeighborsId(FactorGraphNode* node) const {
	vector<string> idList;
	
	// Checks if the node belongs to the factor graph
	if (node->getParent() != this || !nodeExists(node->getId())) {
		return idList;
	}
	
	multimap<string, string>::const_iterator startIter, endIter;
	startIter = mConnectionList.lower_bound(node->getId());
	endIter = mConnectionList.upper_bound(node->getId());
	
	for (startIter; startIter != endIter; startIter++) {
		idList.push_back(startIter->second);
	}
	
	return idList;
}

/** Returns the Ids of the neighnors of the given Id
**/
vector<string> FactorGraph::getNeighborsId(string id) const {

	// Finds the node using the Id
	map<string, FactorGraphNode*>::const_iterator iter = mNodeList.find(id);
	if (iter == mNodeList.end()) {
		vector<string> idList;
		return idList;
	} else {
		return getNeighborsId(iter->second);
	}

}

/** Connects the two given nodes
**/	
bool FactorGraph::connect(FactorGraphNode* sourceNode, FactorGraphNode* targetNode) {
	
	// Checks if the two nodes are valid or if they are the same and that they have the correct parents
	if (sourceNode == 0 || targetNode == 0 || sourceNode == targetNode) {
		return false;
	}
	if (sourceNode->getParent() != this || targetNode->getParent() != this) {
		return false;
	}
	
	// Checks if they are already connected. If it is, just return true
	if (isConnected(sourceNode, targetNode)) {
		return true;
	}
	
	// Adds the entry in the connection list
	mConnectionList.insert(make_pair(sourceNode->getId(), targetNode->getId()));
	mConnectionList.insert(make_pair(targetNode->getId(), sourceNode->getId()));
	return true;
}

/** Connects the two nodes given their Id
**/
bool FactorGraph::connect(string sourceId, string targetId) {
	FactorGraphNode* sourceNode = getNode(sourceId);
	FactorGraphNode* targetNode = getNode(targetId);
	
	if (sourceNode != 0 && targetNode != 0) {
		return connect(sourceNode, targetNode);
	} else {
		return false;
	}
}

/** Checks if the two nodes are connected
**/
bool FactorGraph::isConnected(FactorGraphNode* sourceNode, FactorGraphNode* targetNode) {
	
	// Checks if the source node and target node are valid and correct parents
	if (sourceNode == 0 || targetNode == 0 || sourceNode == targetNode) {
		return false;
	}
	
	if (sourceNode->getParent() != this || targetNode->getParent() != this) {
		return false;
	}
	
	// Checks for connection. Since it is symmetric, we just have to check for one of them
	bool isConnect = false;
	map<string, string>::const_iterator startIter, endIter;
	startIter = mConnectionList.lower_bound(sourceNode->getLabel());
	endIter = mConnectionList.upper_bound(sourceNode->getLabel());
	for (startIter; startIter != endIter; startIter++) {
		if (startIter->second == targetNode->getId()) {
			isConnect = true;
			break;
		}
	}
	return isConnect;
	
}

/** Checks if the two nodes are connected given the Ids
**/
bool FactorGraph::isConnected(string sourceId, string targetId) {
	FactorGraphNode* sourceNode = getNode(sourceId);
	FactorGraphNode* targetNode = getNode(targetId);
	
	if (sourceNode != 0 && targetNode != 0) {
		return isConnected(sourceNode, targetNode);
	} else {
		return false;
	}
}
	
/** Disconnects two given nodes
* However they must be connected first before this can be done.
**/
bool FactorGraph::disconnect(FactorGraphNode* sourceNode, FactorGraphNode* targetNode) {

	// Note: There is no need to check for source Node and target Node being valid as this is done in isConnected()
	
	if (isConnected(sourceNode, targetNode)) {
		multimap<string, string>::iterator iter = mConnectionList.begin();
		while (iter != mConnectionList.end()) {
			if ((iter->first == sourceNode->getId() && iter->second == targetNode->getId()) ||
				(iter->first == targetNode->getId() && iter->second == sourceNode->getId())) {
				mConnectionList.erase(iter++);
			} else {
				iter++;
			}
		}
		return true;
	} else {
		return false;
	}
}

/** Disconnects two nodes given their Id
**/
bool FactorGraph::disconnect(string sourceId, string targetId) {
	
	FactorGraphNode* sourceNode = getNode(sourceId);
	FactorGraphNode* targetNode = getNode(targetId);
	
	if (sourceNode != 0 && targetNode != 0) {
		return disconnect(sourceNode, targetNode);
	} else {
		return false;
	}
	
}

/** Disconnects all the nodes
* It does this by clearing the connection list
**/
void FactorGraph::disconnectAll() {
	mConnectionList.clear();
}

/** Disconnects all nodes and deletes all the nodes
**/
void FactorGraph::removeAll() {
	mConnectionList.clear();
	
	// Deletes the memory off the node list
	map<string, FactorGraphNode*>::const_iterator iter = mNodeList.begin();
	map<string, FactorGraphNode*>::const_iterator tempIter;
	while (iter != mNodeList.end()) {
		tempIter = iter;
		iter++;
		delete tempIter->second;
	}
	mNodeList.clear();
}

/** Renames the node with oldId to newId
**/
bool FactorGraph::renameNode(string oldId, string newId) {
	
	FactorGraphNode* node = getNode(oldId);
	
	if (node != 0) {
		return renameNode(node, newId);
	} else {
		return false;
	}

}

/** Renames the node to a new Id
**/
bool FactorGraph::renameNode(FactorGraphNode* node, string newId) {
	
	// Checks if the node is valid, and if the parent is correct
	if (node == 0 || node->getParent() != this) {
		return false;
	}
	
	// Checks if the new Id is valid
	if (!isValidId(newId) || nodeExists(newId)) {
		return false;
	}
	
	// Removes the old entry from the node list and adds in the new one
	map<string, FactorGraphNode*>::iterator iter = mNodeList.find(node->getId());
	if (iter != mNodeList.end()) { 
		mNodeList.erase(iter);
		mNodeList[newId] = node;
	}
	
	// Changes the entries in the connection list. Erase the entries
	// where the oldId is the key. Then look for the rest to find the rest where the element
	// is oldId, rename them to thew new Id and add the new Id in.
	vector<string> tempList;
	mConnectionList.erase(node->getId());
	multimap<string, string>::iterator connIter = mConnectionList.begin();
	while (connIter != mConnectionList.end()) {
		if (connIter->second == node->getId()) {
			connIter->second = newId;
			tempList.push_back(connIter->first);
		}
		connIter++;
	}
	// For each entry in the temporary list, adds into the connection list
	vector<string>::const_iterator tempIter = tempList.begin();
	while (tempIter != tempList.end()) {
		mConnectionList.insert(make_pair(newId, *tempIter));
		tempIter++;
	}
	
	// Changes the Id of the node
	node->mId = newId;
	
	return true;
}

/** Returns the Id of the factor graph
**/
string FactorGraph::getId() const {
	return mId;
}

/** Sets the Id of the factor graph
* The Id is checked for validity, i.e it follows the variable naming convention
**/
string FactorGraph::setId(string id) {
	if (isValidId(id)) {
		mId = id;
		return id;
	} else {
		return "";
	}
}

/** Returns the label of the factor graph
**/
string FactorGraph::getLabel() const {
	return mLabel;
}

/** Sets the label of the factor graph
**/
void FactorGraph::setLabel(string label) {
	mLabel = label;
}

/** Returns a string representation of the factor graph
**/
string FactorGraph::toString() const {
	ostringstream str;
	
	str << "---FactorGraph---" << endl;
	str << "mId: " << mId << endl;
	str << "mLabel: " << mLabel << endl;
	str << "List of nodes : ";
	map<string, FactorGraphNode*>::const_iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		str << iter->first << " ";
		iter++;
	}
	str << endl;
	str << "List of edges : " << endl;
	multimap<string, string>::const_iterator connIter = mConnectionList.begin();
	while (connIter != mConnectionList.end()) {
		str << "[" << connIter->first << ", " << connIter->second << "]" << endl;
		connIter++;
	}

	return str.str();
}

/** Generates a unique ID
* This method will generate a random ID with the format 'f/v_#' where
* # is an increasing number. The current version of this method does not 
* check for maximal number of elements so there is a chance (although minute)
* of it going into endless loop.
**/
string FactorGraph::generateNodeId(nodeType type) {
	
	string prefix;
	ostringstream suffix;
	string tempStr;
	bool generated;
		
	// Checks the type first to determine the prefix
	if (type == nodeType_factor) {
		prefix = "f_";
	} else {
		prefix = "v_";
	}
	
	generated = false;
	int index = 1;
	while (!generated) {
	
		// Generates name
		suffix << index;
		tempStr = prefix + suffix.str();
		suffix.str("");
		index++;
		
		// Checks if it exists
		map<string, FactorGraphNode*>::iterator iter = mNodeList.find(tempStr);
		if (iter == mNodeList.end()) {
			generated = true;
		} else {
			// Found tempStr, increase index and carry on.
		}
	}
	
	return tempStr;

}

/** Helper function for copying contents
* It copies everything but assumes that allocated memory (if any)
* has already been deleted.
**/
void FactorGraph::copyFrom(const FactorGraph& src) {

	mId = src.mId;
	mLabel = src.mLabel;
	
	// Copies the node list. Note that the nodes are being cloned into separate copies
	vector<FactorGraphNode*> nodeList = ((FactorGraph&)src).getNodes();
	vector<FactorGraphNode*>::iterator iter = nodeList.begin();
	FactorGraphNode* node;
	while (iter != nodeList.end()) {
		node = new FactorGraphNode(**iter);
		this->addNode(node);
		iter++;
	}
	
	// Copies the connection list
	multimap<string, string>::const_iterator connIter = src.mConnectionList.begin();
	while (connIter != src.mConnectionList.end()) {
		mConnectionList.insert(make_pair(connIter->first, connIter->second));
		connIter++;
	}
	
}

/** Default Constructor
**/
FactorGraphNode::FactorGraphNode(string id, string label, nodeType type) : mParent(0), mLabel(label), mType(type) {

	// Checks if the Id is valid. If it is valid then set it. Otherwise put an empty id
	if (isValidId(id)) {
		mId = id;
	} else {
		mId = "";
	}

	#ifdef FACTORGRAPH_DEBUG
		cout << "[Info] FactorGraphNode::FactorGraphNode() - Creating factorNode:id=" << id << ",label=" << label << ",type=" << type << endl;
	#endif

}

/** Copy constructor
* Copies all the contents (other than the parent) from src to the new instance.
**/
FactorGraphNode::FactorGraphNode(const FactorGraphNode& src) {
	copyFrom(src);
}

/** Assignment constructor
**/
FactorGraphNode& FactorGraphNode::operator=(const FactorGraphNode& src) {
	
	// Checks for self assignment
	if (this == &src) {
		return (*this);	
	}
	
	// Remove from parent if parent exists
	if (mParent != 0) {
		mParent->removeNode(this);
	}
	
	// Copy contents
	copyFrom(src);
	
	return (*this);
}

/** Default destructor
**/
FactorGraphNode::~FactorGraphNode() {

	#ifdef FACTORGRAPH_DEBUG
		cout << "[Info] FactorGraphNode::~FactorGraphNode() - Destroying factorNode:id=" << mId << ",label=" << mLabel << ",type=" << mType << endl;
	#endif

	// Remove from factor graph if parent exists
	if (mParent != 0) {
		mParent->removeNode(this);
	}
}

/** Returns the Id of the node
**/
string FactorGraphNode::getId() {
	return mId;
}

/** Sets the Id of the node.
**/
string FactorGraphNode::setId(string id) {

	// If the parent is 0, simply check for valid id and sets it
	if (mParent == 0) {
		if (isValidId(id)) {
			mId = id;
			return id;
		} else {
			return "";
		}
	} else {
		if (mParent->renameNode(this, id)) {
			return id;
		} else {
			return "";
		}
	}
}

/** Returns the label for the node
**/
string FactorGraphNode::getLabel() {
	return mLabel;
}

/** Sets the label for the node
* There are no limitations for the label. It can contain any characters and even spaces.
**/
void FactorGraphNode::setLabel(string label) {
	mLabel = label;
}

/** Returns the type of the node
* The type can either be a factor node or a variable node.
**/
nodeType FactorGraphNode::getType() {
	return mType;
}

/** Sets the type of the node
* Note that consistency is not managed here. The final check have to be done by the FactorGraph parent.
**/
void FactorGraphNode::setType(nodeType type) {
	mType = type;
}

/** Returns the parent for the node
**/
FactorGraph* FactorGraphNode::getParent() {
	return mParent;
}

/** Sets the parent for the node
* If the parents are the same, do nothing. If the old parent is not NULL, then
* have to call removeNode to break all the links. If the new parent is not NULL,
* then have to call addNodes to set the links. Note that if the Id is not valid for
* the new parent, it will not be added to the new parent and becomes an orphan node
**/
void FactorGraphNode::setParent(FactorGraph* parent) {
	
	// Checks if the parent is the same as the given one
	if (mParent == parent) {
		return;
	}
	
	// Checks if the old parent is 0. Make sure that
	// after removing the node, mParent = 0
	if (mParent != 0) {
		mParent->removeNode(this);
	}
	
	// Checks if the new parent is 0. Make sure that
	// after adding the node, mParent = parent
	if (parent != 0) {
		parent->addNode(this);
	}
}

/** Returns a const vector of neighbor nodes
* If the parent exists, use the parent to obtain the list of neighbors and
* return it. Otherwise return an empty vector.
**/
vector<FactorGraphNode*> FactorGraphNode::getNeighbors() {
	// If it has a parent graph, use it to get the neighbors
	if (mParent != 0) {
		return mParent->getNeighbors(this);
	} else {
		vector<FactorGraphNode*> neighborList;
		return neighborList;
	}
}

/** Returns a string representation of the factor node.
* The string will contain all the details such as the id, name, parent name etc.
**/
string FactorGraphNode::toString() {
	ostringstream str;
	
	str << "---FactorGraphNode---" << endl;
	str << "mId: " << mId << endl;
	str << "mLabel: " << mLabel << endl;
	str << "mType: " << mType << endl;
	if (mParent == 0) {
		str << "mParent: " << mParent << endl;
	} else {
		str << "mParent: " << mParent->getLabel() << endl;
	}
	
	return str.str();
}

/** Helper function for copying contents
* It copies everything but assumes that allocated memory (if any)
* has already been deleted.
* Note that the parent is not being copied as it would violate the unique ID requirement.
**/
void FactorGraphNode::copyFrom(const FactorGraphNode& src) {
	mId = src.mId;
	mLabel = src.mLabel;
	mType = src.mType;
}

}
