/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: factorgraph.h
**  
** Description:
** The header file containing the structure of the factor graphs, including its components
** such as the nodes.
**  
** Last modified: 
** 04 December 2006
** 
** Comments:
** 1)  For the factor graph implementation, note that the edges are implicitly defined, 
**      i.e. the connection list in the FactorGraph class will store the mapping list. Note that it
**      is supposed to be symmetric. [23 Oct 2006]
** 2) The nodes and factor graphs are only supposed to behave as they are - Factor Graph.
**      Further extension such as domains (continuous or discrete) will be put as separate
**      classes. Then other utility classes such as Belief Propagation class will be defined separately
**      so as to separate the concepts of factor graphs (or any other probabilistic models for that
**      matter) from the different possible algorithms.  [23 Oct 2006]
** 3) Replaced all the FactorNode with FactorGraphNode. This is to cater for the introduction of
**      FactorTree, in which the tree nodes will be called FactorTreeNode [04 Dec 2006]
****************************************************************************/

#ifndef FACTORGRAPH_H_
#define FACTORGRAPH_H_

/* For debugging. Comment out to suppress debugging lines */
#define FACTORGRAPH_DEBUG

/* Standard Includes */
#include <iostream>
#include <vector>
#include <map>
 
using namespace std;
 
namespace factorgraph {

typedef enum {
	nodeType_factor,
	nodeType_variable
} nodeType;

bool isValidId(string id);
class FactorGraphNode;

class FactorGraph {

public:

	/* Constructors and destructors */
	FactorGraph(string id = "", string label = "");
	FactorGraph(const FactorGraph& src);	
	FactorGraph& operator=(const FactorGraph& src);
	virtual ~FactorGraph(); 
	
	/* Main methods to add and remove nodes */
	bool isConsistent() const;
	bool nodeExists(string id) const;
	int nodeSize() const;
	FactorGraphNode* addNode(nodeType type = nodeType_factor);
	FactorGraphNode* addNode(string id, string label = "", nodeType type = nodeType_factor);
	FactorGraphNode* addNode(FactorGraphNode* node);
	FactorGraphNode* getNode(string id);
	FactorGraphNode* removeNode(string id);
	FactorGraphNode* removeNode(FactorGraphNode* node);
	bool deleteNode(string id);
	vector<FactorGraphNode*> getNodes();
	vector<string> getNodesId() const;
	vector<FactorGraphNode*> getFactors();
	vector<string> getFactorsId() const;
	vector<FactorGraphNode*> getVariables();
	vector<string> getVariablesId() const;
	vector<FactorGraphNode*> getNeighbors(FactorGraphNode* node);
	vector<FactorGraphNode*> getNeighbors(string id);
	vector<string> getNeighborsId(FactorGraphNode* node) const;
	vector<string> getNeighborsId(string id) const;
	bool connect(FactorGraphNode* sourceNode, FactorGraphNode* targetNode);
	bool connect(string sourceId, string targetId);
	bool isConnected(FactorGraphNode* sourceNode, FactorGraphNode* targetNode);
	bool isConnected(string sourceId, string targetId);
	bool disconnect(FactorGraphNode* sourceNode, FactorGraphNode* targetNode);
	bool disconnect(string sourceId, string targetId);
	void disconnectAll();
	void removeAll();
	bool renameNode(string oldId, string newId);
	bool renameNode(FactorGraphNode* node, string newId);

	/* Methods to set and get attributes */
	string getId() const;
	string setId(string id);
	string getLabel() const;
	void setLabel(string label);
	
	/* Printing and debugging */
	string toString() const;
	
protected:

	/* Attributes */
	string mId;
	string mLabel;
	map<string, FactorGraphNode*> mNodeList;
	multimap<string, string> mConnectionList;

	/* Protected methods */
	string generateNodeId(nodeType type = nodeType_factor);

private:
	void copyFrom(const FactorGraph& src);
	
};

class FactorGraphNode {

public:

	/* Constructors and destructors */
	FactorGraphNode(string id = "", string label = "", nodeType type = nodeType_factor);
	FactorGraphNode(const FactorGraphNode& src);
	FactorGraphNode& operator=(const FactorGraphNode& src);
	virtual ~FactorGraphNode(); 

	/* Methods to set and get attributes */
	string getId();
	string setId(string id);
	string getLabel();
	void setLabel(string label);
	nodeType getType();
	void setType(nodeType type);
	FactorGraph* getParent();
	void setParent(FactorGraph* parent);
	vector<FactorGraphNode*> getNeighbors();
	
	/* Printing and debugging */
	string toString();
	
protected:

	/* Parent */
	FactorGraph* mParent;
	
	/* Attributes */
	string mId;
	string mLabel;
	nodeType mType;

private:

	/* Friend class */
	friend class FactorGraph;
	void copyFrom(const FactorGraphNode& src);
	
};

}

#endif /* FACTORGRAPH_H_ */
