/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: factortree.cpp
**  
** Description:
** The source file containing the structure of the factor tree, including its components
** such as the nodes.
**  
** Last modified: 
** 04 December 2006
** 
** Comments:
**
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <sstream>
#include <vector>
#include <map>

/* Application Includes */
#include "factortree.h"

using namespace std;

namespace factortree {

/** Checks if the id is valid
* Checks if a given Id is valid, according to standard variable rules.
* 1) It must not be empty
* 2) It must start with an alphabet or underscore "_"
* 3) It must not contain any invalid characters, such as ", ", "?", " " etc.
**/
bool isValidId(string id) {

	// Checks for empty string
	if (id == "") return false;
	
	char c;
	
	// Checks the first charactor
	c = id[0];
	if (c != '_' && !isalpha(c)) return false;
	
	// Checks the rest of the characters. The rest can only
	// be alphanumeric or '_'
	int length = strlen(id.c_str());
	for (int i = 1; i < length; i++) {
		c = id[i];
		if (c != '_' && !isalnum(c)) return false;
	}
	
	return true;
}

/** Default constructor
**/
FactorTree::FactorTree(string id, string label) : mLabel(label), mRootId("") {
	
	// Checks if the Id is valid. If it is valid then set it. Otherwise put an empty id
	if (isValidId(id)) {
		mId = id;
	} else {
		mId = "";
	}

	#ifdef FACTORTREE_DEBUG
		cout << "[Info] FactorTree::FactorTree() - Creating FactorTree:id=" << id << ",label=" << label << endl;
	#endif

}

/** Copy constructor
* Copies all the contents (By creating a new instance of them) from src to the new instance.
**/
FactorTree::FactorTree(const FactorTree& src) {
	copyFrom(src);
}

/** Assignment constructor
**/
FactorTree& FactorTree::operator=(const FactorTree& src) {
	
	// Checks for self assignment
	if (this == &src) {
		return (*this);	
	}
	
	// Clears the node list and connection list
	removeAll();
	
	// Copy contents
	copyFrom(src);
	
	return (*this);

}

/** Default Destructor
**/
FactorTree::~FactorTree() {
	
	#ifdef FACTORTREE_DEBUG
		cout << "[Info] FactorTree::~FactorTree() - Destroying FactorTree:id=" << mId << ",label=" << mLabel << endl;
	#endif

	// Clears everything
	removeAll();
}

/** Checks if the Factor Tree is consistent
* Performs a series of checks, one of which is to check if it is bipartite, i.e. no two nodes
* of the same type are supposed to be connected together
* Then is to check if the underlying graph is singly connected, and check if there is a valid
* root which is a variable node
**/
bool FactorTree::isConsistent() const {
	// Todo
	return true;
}

/** Checks if the node with the Id exists in this factor tree
**/
bool FactorTree::nodeExists(string id) const {
	map<string, FactorTreeNode*>::const_iterator iter = mNodeList.find(id);
	if (iter != mNodeList.end()) {
		return true;
	} else {
		return false;
	}
}

/** Checks if the node is the root of the factor tree
**/
bool FactorTree::isRoot(FactorTreeNode* node) const {
	return isRoot(node->getId());
}

/** Checks if the node with the given id is the root of the factor tree
**/
bool FactorTree::isRoot(string id) const {
	if (mRootId == id) {
		return true;
	} else {
		return false;
	}
}

/** Checks if the node is a leaf of the factor tree
**/
bool FactorTree::isLeaf(FactorTreeNode* node) const {
	return isLeaf(node->getId());
}

/** Checks if the node with the given id is a leaf  of the factor tree
* It does this by checking the child connection list.
**/
bool FactorTree::isLeaf(string id) const {
	multimap<string, string>::const_iterator iter = mChildConnectionList.find(id);
	if (iter == mChildConnectionList.end()) {
		return true;
	} else {
		return false;
	}
}

/** Returns the number of nodes
**/
int FactorTree::nodeSize() const {
	return mNodeList.size();
}

/** Sets the node as the root node
* If the node is not in the tree, adds it in before setting it as the root
**/
FactorTreeNode* FactorTree::setRoot(FactorTreeNode* node) {
	
	// Check the type of the node. If it is a factor node, return an 0 as only
	// variable nodes can be root
	if (node->getType() == nodeType_factor) {
		#ifdef FACTORTREE_DEBUG
			cout << "[Error] FactorTree::setRoot(1) - Node " << node->getId() << " is not a variable node" << endl;
		#endif
		return 0;
	} 
	
	// Checks if the node is in the factor tree. If it is not, add it in
	if (!nodeExists(node->getId())) {
		if (addNode(node) == 0) {
			#ifdef FACTORTREE_DEBUG
				cout << "[Error] FactorTree::setRoot(1) - Node " << node->getId() << " cannot be added to the tree" << endl;
			#endif
			return 0;
		}
	} else if (mNodeList[node->getId()]->getType() != node->getType()) {
		#ifdef FACTORTREE_DEBUG
			cout << "[Error] FactorTree::setRoot(1) - Node " << node->getId() << " type mismatch" << endl;
		#endif
		return 0;
	}
	
	// Sets this node as the root
	mRootId = node->getId();
	
	return node;
}

/** Sets the node with the id as the root node. 
* If the node does not exist, create one and sets it as the root
**/
FactorTreeNode* FactorTree::setRoot(string id) {

	// Checks if the node is in the factor tree. If it is not, add it in
	if (!nodeExists(id)) {
		if (addNode(id, id, nodeType_variable) == 0) {
			#ifdef FACTORTREE_DEBUG
				cout << "[Error] FactorTree::setRoot(2) - Node " << id << " cannot be added to the tree" << endl;
			#endif
			return 0;
		}
	} else if (mNodeList[id]->getType() != nodeType_variable) {
		#ifdef FACTORTREE_DEBUG
			cout << "[Error] FactorTree::setRoot(2) - Node " << id << " type mismatch" << endl;
		#endif
		return 0;
	}
	
	// Sets this node as the root
	mRootId = id;
	
	return mNodeList[id];
}

/** Creates a default node and adds it to the factor graph
* The node will be default, i.e. its Id will be generated and its name will take
* its Id.
**/
FactorTreeNode* FactorTree::addNode(nodeType type) {

	// Creates new node with name and Id being the same
	string id = generateNodeId(type);
	FactorTreeNode* node = new FactorTreeNode(id, id, type);
	
	// Adds the node in
	addNode(node);
	
	return node;
}

/** Creates a new node with the given Id, name and type and adds it to the factor graph
* If, however, the Id is not valid, a node WILL NOT be created at all.
**/
FactorTreeNode* FactorTree::addNode(string id, string label, nodeType type) {
	
	// Checks for Id validity and whether it exists in the factor graph
	if (!isValidId(id) && nodeExists(id)) {
		return 0;
	}
	
	// Creates a new node with Id and label and type
	FactorTreeNode* node = new FactorTreeNode(id, label, type);
	
	// Adds the node in
	addNode(node);
	
	return node;
}

/** Adds a node to the factor graph
**/
FactorTreeNode* FactorTree::addNode(FactorTreeNode* node) {

	// If checks if the node is already added
	if (node->getParent() == this) {
		return node;
	}
	
	// If the node has an earlier parent, remove it first
	if (node->getParent() != 0) {
		node->getParent()->removeNode(node);
	}
	
	// Checks if Id exists already in the factor graph. If it is, then do
	// not add the node in
	if (nodeExists(node->getId())) {
		return 0;
	} else {
		mNodeList[node->getId()] = node;
		node->mParent = this;
		return node;
	}
	
}

/** Finds the node with the given Id and returns it
**/
FactorTreeNode* FactorTree::getNode(string id) {
	map<string, FactorTreeNode*>::iterator iter = mNodeList.find(id);
	if (iter == mNodeList.end()) {
		return 0;
	} else {
		return iter->second;
	}
}

/** Finds a node and remove it from the factor tree
**/
FactorTreeNode* FactorTree::removeNode(string id) {
	map<string, FactorTreeNode*>::iterator iter = mNodeList.find(id);
	if (iter == mNodeList.end()) {
		return 0;
	} else {
		return removeNode(iter->second);
	}
}

/** Removes a node from the factor tree
**/
FactorTreeNode* FactorTree::removeNode(FactorTreeNode* node) {
	
	// Checks if the parent is indeed the current instance
	if (node->getParent() != this) {
		return 0;
	}
	
	// Removes it from both the connection list, i.e. any connection with this node will be
	// disconnected.
	multimap<string, string>::iterator childIter = mChildConnectionList.begin();
	while (childIter != mChildConnectionList.end()) {
		if (childIter->first == node->getId() || childIter->second == node->getId()) {
			mChildConnectionList.erase(childIter++);
		} else {
			childIter++;
		}
	}
	map<string, string>::iterator parentIter = mParentConnectionList.begin();
	while (parentIter != mParentConnectionList.end()) {
		if (parentIter->first == node->getId() || parentIter->second == node->getId()) {
			mParentConnectionList.erase(parentIter++);
		} else {
			parentIter++;
		}
	}
	// Checks if it is the root. If it is, set the root to empty string
	if (mRootId == node->getId()) {
		mRootId = "";
	}
	
	// Remove the node from node list
	mNodeList.erase(node->getId());
	
	// Remove the parent and returns the node
	node->mParent = 0;
	return node;
	
}

/** Removes the node from a factor graph and deletes it off
**/
bool FactorTree::deleteNode(string id) {
	
	FactorTreeNode* node = removeNode(id);
	
	if (node != 0) {
		delete node;
		return true;
	} else {
		return false;
	}
}

/** Returns the vector of nodes
**/
vector<FactorTreeNode*> FactorTree::getNodes() {
	vector<FactorTreeNode*> nodeList;
	map<string, FactorTreeNode*>::iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		nodeList.push_back(iter->second);
		iter++;
	}
	return nodeList;
}

/** Returns the vector of node Id
**/
vector<string> FactorTree::getNodesId() const {
	vector<string> idList;
	map<string, FactorTreeNode*>::const_iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		idList.push_back(iter->first);
		iter++;
	}
	return idList;
}

/** Returns the vector of nodes of type: Factor
**/
vector<FactorTreeNode*> FactorTree::getFactors() {
	vector<FactorTreeNode*> nodeList;
	map<string, FactorTreeNode*>::iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		if (iter->second->getType() == nodeType_factor) {
			nodeList.push_back(iter->second);
		}
		iter++;
	}
	return nodeList;
}

/** Returns a vector of node Id for the factor nodes
**/
vector<string> FactorTree::getFactorsId() const {
	vector<string> idList;
	map<string, FactorTreeNode*>::const_iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		if (iter->second->getType() == nodeType_factor) {
			idList.push_back(iter->first);
		}
		iter++;
	}
	return idList;
}

/** Returns the vector of nodes of type: Variable
**/
vector<FactorTreeNode*> FactorTree::getVariables() {
	vector<FactorTreeNode*> nodeList;
	map<string, FactorTreeNode*>::iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		if (iter->second->getType() == nodeType_variable) {
			nodeList.push_back(iter->second);
		}
		iter++;
	}
	return nodeList;
}

/** Returns a vector of node Id for the variable nodes
**/
vector<string> FactorTree::getVariablesId() const {
	vector<string> idList;
	map<string, FactorTreeNode*>::const_iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		if (iter->second->getType() == nodeType_variable) {
			idList.push_back(iter->first);
		}
		iter++;
	}
	return idList;
}

/** Returns the root node, if available. Else return a 0
**/
FactorTreeNode* FactorTree::getRoot() {
	return getNode(mRootId);
}

/** Returns the rood node id
**/
string FactorTree::getRootId() const {
	return mRootId;
}

/** Returns a vector of leaf nodes
**/
vector<FactorTreeNode*> FactorTree::getLeaves() {
	vector<string> idList = getLeavesId();
	vector<FactorTreeNode*> nodeList;
	
	vector<string>::const_iterator iter = idList.begin();
	while (iter != idList.end()) {
		nodeList.push_back(getNode(*iter));
		iter++;
	}
	
	return nodeList;
}

/** Returns a vector of leaf node ids
**/
vector<string> FactorTree::getLeavesId() const {
	vector<string> idList;
	
	string currNodeId;
	map<string, FactorTreeNode*>::const_iterator nodeIter = mNodeList.begin();
	multimap<string, string>::const_iterator childIter;
	while (nodeIter != mNodeList.end()) {
		childIter = mChildConnectionList.find(nodeIter->first);
		// If the node is a leaf then it has no children and find() will return no result.
		if (childIter == mChildConnectionList.end()) {
			idList.push_back(nodeIter->first);
		}
		nodeIter++;
	}
	
	return idList;
}

/** Returns the parent of the given node
**/
FactorTreeNode* FactorTree::getParent(FactorTreeNode* node) {

	// Checks if the node belongs in the factor tree
	if (node->getParent() != this) {
		return 0;
	}

	string parentId = getParentId(node);
	if (parentId != "") {
		return getNode(parentId);
	} else {
		return 0;
	}
}

/** Returns the parent of the given node id
**/
FactorTreeNode* FactorTree::getParent(string id) {
	string parentId = getParentId(id);
	if (parentId != "") {
		return getNode(parentId);
	} else {
		return 0;
	}
}

/** Returns the parent id of the given node
**/
string FactorTree::getParentId(FactorTreeNode* node) const {

	// Checks if the node belongs in the factor tree
	if (node->getParent() != this) {
		return "";
	}

	return getParentId(node->getId());
}

/** Returns the parent id of the node with given id
**/
string FactorTree::getParentId(string id) const {
	map<string, string>::const_iterator iter = mParentConnectionList.find(id);
	if (iter != mParentConnectionList.end()) {
		return iter->second;
	}
	return "";
}

/** Returns the children of the given node
**/
vector<FactorTreeNode*> FactorTree::getChildren(FactorTreeNode* node) {

	vector<FactorTreeNode*> nodeList;
	
	// Checks if the node belongs in the factor tree
	if (node->getParent() != this) {
		return nodeList;
	}

	nodeList = getChildren(node->getId());
	return nodeList;
}

/** Returns the children of the given node
**/
vector<FactorTreeNode*> FactorTree::getChildren(string id) {
	vector<string> idList;
	vector<FactorTreeNode*> nodeList;
	
	idList = getChildrenId(id);
	vector<string>::const_iterator iter = idList.begin();
	while (iter != idList.end()) {
		nodeList.push_back(getNode(*iter));
		iter++;
	}
	
	return nodeList;
}

/** Returns the children of the given node in a const vector
**/
vector<string> FactorTree::getChildrenId(FactorTreeNode* node) const {
	vector<string> idList;
	
	// Checks if the node belongs in the factor tree
	if (node->getParent() != this) {
		return idList;
	}
	
	idList = getChildrenId(node->getId());
	return idList;
}

/** Returns the children of the given node Id in a const vector
**/
vector<string> FactorTree::getChildrenId(string id) const {
	vector<string> idList;

	multimap<string, string>::const_iterator startIter, endIter;
	startIter = mChildConnectionList.lower_bound(id);
	endIter = mChildConnectionList.upper_bound(id);
	
	for (startIter; startIter != endIter; startIter++) {
		idList.push_back(startIter->second);
	}
	
	return idList;
}

/** Connects the two given nodes
**/	
bool FactorTree::connect(FactorTreeNode* parentNode, FactorTreeNode* childNode) {
	
	// Checks if the two nodes are valid or if they are the same and that they have the correct parents
	if (parentNode == 0 || childNode == 0 || parentNode == childNode) {
		return false;
	}
	if (parentNode->getParent() != this || childNode->getParent() != this) {
		return false;
	}
	
	// Checks if they are already connected. If it is, just return true
	if (isParent(parentNode, childNode)) {
		return true;
	}

	// Checks if the child node has no parents
	if (getParentId(childNode) != "") {
		return false;
	}
	
	// Adds the entry in the connection list
	mChildConnectionList.insert(make_pair(parentNode->getId(), childNode->getId()));
	mParentConnectionList.insert(make_pair(childNode->getId(), parentNode->getId()));
	return true;
}

/** Connects the two nodes given their Id
**/
bool FactorTree::connect(string parentId, string childId) {
	FactorTreeNode* parentNode = getNode(parentId);
	FactorTreeNode* childNode = getNode(childId);
	
	if (parentNode != 0 && childNode != 0) {
		return connect(parentNode, childNode);
	} else {
		return false;
	}
}

/** Checks if a node is the parent of the other given node
**/
bool FactorTree::isParent(FactorTreeNode* parentNode, FactorTreeNode* childNode) {

	// Checks if the source node and target node are valid and correct parents
	if (parentNode == 0 || childNode == 0 || parentNode == childNode) {
		return false;
	}
	
	if (parentNode->getParent() != this || childNode->getParent() != this) {
		return false;
	}

	return isParent(parentNode->getId(), childNode->getId());
}

/** Checks if a node is the parent of the other given node
**/
bool FactorTree::isParent(string parentId, string childId) {

	map<string, string>::const_iterator iter = mParentConnectionList.find(childId);
	if (iter != mParentConnectionList.end() && iter->second == parentId) {
		return true;
	} else {
		return false;
	}
}
	
/** Disconnects two given nodes
* However they must be connected first before this can be done.
**/
bool FactorTree::disconnect(FactorTreeNode* parentNode, FactorTreeNode* childNode) {

	// Note: There is no need to check for parent Node and child Node being valid as this is done in isParent()
	
	if (isParent(parentNode, childNode)) {
	
		multimap<string, string>::iterator startIter, endIter;
		startIter = mChildConnectionList.lower_bound(parentNode->getId());
		endIter = mChildConnectionList.upper_bound(parentNode->getId());
	
		for (startIter; startIter != endIter; startIter++) {
			if (startIter->second == childNode->getId()) {
				mChildConnectionList.erase(startIter);
				break;
			}
		}
	
		map<string, string>::iterator iter = mParentConnectionList.find(childNode->getId());
		if (iter != mParentConnectionList.end()) {
			mParentConnectionList.erase(iter);
		}
		
		return true;
	} else {
		return false;
	}
}

/** Disconnects two nodes given their Id
**/
bool FactorTree::disconnect(string parentId, string childId) {
	
	FactorTreeNode* parentNode = getNode(parentId);
	FactorTreeNode* childNode = getNode(childId);
	
	if (parentNode != 0 && childNode != 0) {
		return disconnect(parentNode, childNode);
	} else {
		return false;
	}
	
}

/** Disconnects all the nodes
* It does this by clearing the connection list, both parent, child and
* also unsets the root node
**/
void FactorTree::disconnectAll() {
	mChildConnectionList.clear();
	mParentConnectionList.clear();
	mRootId = "";
}

/** Disconnects all nodes and deletes all the nodes
**/
void FactorTree::removeAll() {
	mChildConnectionList.clear();
	mParentConnectionList.clear();
	mRootId = "";
	
	// Deletes the memory off the node list
	map<string, FactorTreeNode*>::const_iterator iter = mNodeList.begin();
	map<string, FactorTreeNode*>::const_iterator tempIter;
	while (iter != mNodeList.end()) {
		tempIter = iter;
		iter++;
		delete tempIter->second;
	}
	mNodeList.clear();
}

/** Renames the node with oldId to newId
**/
bool FactorTree::renameNode(string oldId, string newId) {
	
	FactorTreeNode* node = getNode(oldId);
	
	if (node != 0) {
		return renameNode(node, newId);
	} else {
		return false;
	}

}

/** Renames the node to a new Id
**/
bool FactorTree::renameNode(FactorTreeNode* node, string newId) {
	
	// Checks if the node is valid, and if the parent is correct
	if (node == 0 || node->getParent() != this) {
		return false;
	}
	
	// Checks if the new Id is valid
	if (!isValidId(newId) || nodeExists(newId)) {
		return false;
	}
	
	string oldId = node->getId();
	vector<string> childList = getChildrenId(oldId);
	string parentId = getParentId(oldId);
	
	// Removes the old entry from the node list and adds in the new one
	map<string, FactorTreeNode*>::iterator iter = mNodeList.find(node->getId());
	if (iter != mNodeList.end()) { 
		mNodeList.erase(iter);
		mNodeList[newId] = node;
	}
	
	// Changes the entries in the connection list. Erase the entries
	// where the oldId is the key. Then look for the rest to find the rest where the element
	// is oldId, rename them to thew new Id and add the new Id in.
	multimap<string, string>::iterator childConnIter = mChildConnectionList.begin();
	while (childConnIter != mChildConnectionList.end()) {
		if (childConnIter->second == oldId) {
			childConnIter->second = newId;
		}
		childConnIter++;
	}
	mChildConnectionList.erase(oldId);
	vector<string>::const_iterator childIter = childList.begin();
	while (childIter != childList.end()) {
		mChildConnectionList.insert(make_pair(newId, *childIter));
		childIter++;
	}
	
	// Perform the same thing for the parent list
	map<string, string>::iterator parentConnIter = mParentConnectionList.begin();
	while (parentConnIter != mParentConnectionList.end()) {
		if (parentConnIter->second == oldId) {
			parentConnIter->second = newId;
		}
		parentConnIter++;
	}
	if (parentId != "") {
		mParentConnectionList.erase(oldId);
		mParentConnectionList.insert(make_pair(newId, parentId));
	}
	
	// Changes the Id of the node
	node->mId = newId;
	
	return true;
}

/** Returns the Id of the factor graph
**/
string FactorTree::getId() const {
	return mId;
}

/** Sets the Id of the factor graph
* The Id is checked for validity, i.e it follows the variable naming convention
**/
string FactorTree::setId(string id) {
	if (isValidId(id)) {
		mId = id;
		return id;
	} else {
		return "";
	}
}

/** Returns the label of the factor graph
**/
string FactorTree::getLabel() const {
	return mLabel;
}

/** Sets the label of the factor graph
**/
void FactorTree::setLabel(string label) {
	mLabel = label;
}

/** Returns a string representation of the factor graph
**/
string FactorTree::toString() const {
	ostringstream str;
	
	str << "---FactorTree---" << endl;
	str << "mId: " << mId << endl;
	str << "mLabel: " << mLabel << endl;
	str << "mRoot: " << mRootId << endl;
	str << "List of nodes : ";
	map<string, FactorTreeNode*>::const_iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		str << iter->first << " ";
		iter++;
	}
	str << endl;
	str << "List of edges : " << endl;
	multimap<string, string>::const_iterator connIter = mChildConnectionList.begin();
	while (connIter != mChildConnectionList.end()) {
		str << "[" << connIter->first << ", " << connIter->second << "]" << endl;
		connIter++;
	}

	return str.str();
}

/** Generates a unique ID
* This method will generate a random ID with the format 'f/v_#' where
* # is an increasing number. The current version of this method does not 
* check for maximal number of elements so there is a chance (although minute)
* of it going into endless loop.
**/
string FactorTree::generateNodeId(nodeType type) {
	
	string prefix;
	ostringstream suffix;
	string tempStr;
	bool generated;
		
	// Checks the type first to determine the prefix
	if (type == nodeType_factor) {
		prefix = "f_";
	} else {
		prefix = "v_";
	}
	
	generated = false;
	int index = 1;
	while (!generated) {
	
		// Generates name
		suffix << index;
		tempStr = prefix + suffix.str();
		suffix.str("");
		index++;
		
		// Checks if it exists
		map<string, FactorTreeNode*>::iterator iter = mNodeList.find(tempStr);
		if (iter == mNodeList.end()) {
			generated = true;
		} else {
			// Found tempStr, increase index and carry on.
		}
	}
	
	return tempStr;

}

/** Helper function for copying contents
* It copies everything but assumes that allocated memory (if any)
* has already been deleted.
**/
void FactorTree::copyFrom(const FactorTree& src) {

	mId = src.mId;
	mLabel = src.mLabel;
	mRootId = src.mRootId;
	
	// Copies the node list. Note that the nodes are being cloned into separate copies
	vector<FactorTreeNode*> nodeList = ((FactorTree&)src).getNodes();
	vector<FactorTreeNode*>::iterator iter = nodeList.begin();
	FactorTreeNode* node;
	while (iter != nodeList.end()) {
		node = new FactorTreeNode(**iter);
		this->addNode(node);
		iter++;
	}
	
	// Copies the connection list
	multimap<string, string>::const_iterator connIter = src.mChildConnectionList.begin();
	while (connIter != src.mChildConnectionList.end()) {
		mChildConnectionList.insert(make_pair(connIter->first, connIter->second));
		connIter++;
	}
	map<string, string>::const_iterator parentConnIter = src.mParentConnectionList.begin();
	while (parentConnIter != src.mParentConnectionList.end()) {
		mParentConnectionList.insert(make_pair(parentConnIter->first, parentConnIter->second));
		parentConnIter++;
	}
	
}

/** Default Constructor
**/
FactorTreeNode::FactorTreeNode(string id, string label, nodeType type) : mParent(0), mLabel(label), mType(type) {

	// Checks if the Id is valid. If it is valid then set it. Otherwise put an empty id
	if (isValidId(id)) {
		mId = id;
	} else {
		mId = "";
	}

	#ifdef FACTORTREE_DEBUG
		cout << "[Info] FactorTreeNode::FactorTreeNode() - Creating factorNode:id=" << id << ",label=" << label << ",type=" << type << endl;
	#endif

}

/** Copy constructor
* Copies all the contents (other than the parent) from src to the new instance.
**/
FactorTreeNode::FactorTreeNode(const FactorTreeNode& src) {
	copyFrom(src);
}

/** Assignment constructor
**/
FactorTreeNode& FactorTreeNode::operator=(const FactorTreeNode& src) {
	
	// Checks for self assignment
	if (this == &src) {
		return (*this);	
	}
	
	// Remove from parent if parent exists
	if (mParent != 0) {
		mParent->removeNode(this);
	}
	
	// Copy contents
	copyFrom(src);
	
	return (*this);
}

/** Default destructor
**/
FactorTreeNode::~FactorTreeNode() {

	#ifdef FACTORTREE_DEBUG
		cout << "[Info] FactorTreeNode::~FactorTreeNode() - Destroying FactorTreeNode:id=" << mId << ",label=" << mLabel << ",type=" << mType << endl;
	#endif

	// Remove from factor graph if parent exists
	if (mParent != 0) {
		mParent->removeNode(this);
	}
}

/** Returns the Id of the node
**/
string FactorTreeNode::getId() {
	return mId;
}

/** Sets the Id of the node.
**/
string FactorTreeNode::setId(string id) {

	// If the parent is 0, simply check for valid id and sets it
	if (mParent == 0) {
		if (isValidId(id)) {
			mId = id;
			return id;
		} else {
			return "";
		}
	} else {
		if (mParent->renameNode(this, id)) {
			return id;
		} else {
			return "";
		}
	}
}

/** Returns the label for the node
**/
string FactorTreeNode::getLabel() {
	return mLabel;
}

/** Sets the label for the node
* There are no limitations for the label. It can contain any characters and even spaces.
**/
void FactorTreeNode::setLabel(string label) {
	mLabel = label;
}

/** Returns the type of the node
* The type can either be a factor node or a variable node.
**/
nodeType FactorTreeNode::getType() {
	return mType;
}

/** Sets the type of the node
* Note that consistency is not managed here. The final check have to be done by the FactorTree parent.
**/
void FactorTreeNode::setType(nodeType type) {
	mType = type;
}

/** Returns the parent for the node
**/
FactorTree* FactorTreeNode::getParent() {
	return mParent;
}

/** Sets the parent for the node
* If the parents are the same, do nothing. If the old parent is not NULL, then
* have to call removeNode to break all the links. If the new parent is not NULL,
* then have to call addNodes to set the links. Note that if the Id is not valid for
* the new parent, it will not be added to the new parent and becomes an orphan node
**/
void FactorTreeNode::setParent(FactorTree* parent) {
	
	// Checks if the parent is the same as the given one
	if (mParent == parent) {
		return;
	}
	
	// Checks if the old parent is 0. Make sure that
	// after removing the node, mParent = 0
	if (mParent != 0) {
		mParent->removeNode(this);
	}
	
	// Checks if the new parent is 0. Make sure that
	// after adding the node, mParent = parent
	if (parent != 0) {
		parent->addNode(this);
	}
}

/** Returns a pointer to the parent node (not parent tree)
**/
FactorTreeNode* FactorTreeNode::getParentNode() {
	// If it has a parent tree, use it to get its parent node
	if (mParent != 0) {
		mParent->getParent(this);
	} else {
		return 0;
	}
}

/** Returns a vector of children nodes
* If the parent exists, use the parent to obtain the list of children nodes and return it.
* Otherwise, return an empty vector
**/
vector<FactorTreeNode*> FactorTreeNode::getChildren() {
	// If it has a parent tree, use it to get the children
	if (mParent != 0) {
		return mParent->getChildren(this);
	} else {
		vector<FactorTreeNode*> childrenList;
		return childrenList;
	}
}

/** Returns a string representation of the factor node.
* The string will contain all the details such as the id, name, parent name etc.
**/
string FactorTreeNode::toString() {
	ostringstream str;
	
	str << "---FactorTreeNode---" << endl;
	str << "mId: " << mId << endl;
	str << "mLabel: " << mLabel << endl;
	str << "mType: " << mType << endl;
	if (mParent == 0) {
		str << "mParent: " << mParent << endl;
	} else {
		str << "mParent: " << mParent->getLabel() << endl;
	}
	
	return str.str();
}

/** Helper function for copying contents
* It copies everything but assumes that allocated memory (if any)
* has already been deleted.
* Note that the parent is not being copied as it would violate the unique ID requirement.
**/
void FactorTreeNode::copyFrom(const FactorTreeNode& src) {
	mId = src.mId;
	mLabel = src.mLabel;
	mType = src.mType;
}

}
