/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: factortree.h
**  
** Description:
** The header file containing the structure of the factor tree, including its components
** such as the nodes.
**  
** Last modified: 
** 04 December 2006
** 
** Comments:
** 1) Essentially this factor tree is not much different from the factor graph, with the exception
**     that (i) it must of course be a tree (ii) there is one variable node which is a root
** 2) The edges are still implicitly defined, but this time, all the edges are only to its children
**      while there is one edge reserved for its parents.
** 3) While there can be many nodes without parents, only one of them (or none) can be the
**     root node, which, in this case, will be flagged out by the method isConsistent() to be
**     an error as a Factor Tree can only have one node that has no parents.
****************************************************************************/

#ifndef FACTORTREE_H_
#define FACTORTREE_H_

/* For debugging. Comment out to suppress debugging lines */
#define FACTORTREE_DEBUG

/* Standard Includes */
#include <iostream>
#include <vector>
#include <map>
 
using namespace std;
 
namespace factortree {

typedef enum {
	nodeType_factor,
	nodeType_variable
} nodeType;

bool isValidId(string id);
class FactorTreeNode;

class FactorTree {

public:

	/* Constructors and destructors */
	FactorTree(string id = "", string label = "");
	FactorTree(const FactorTree& src);	
	FactorTree& operator=(const FactorTree& src);
	virtual ~FactorTree(); 
	
	/* Main methods to add and remove nodes */
	bool isConsistent() const;
	bool nodeExists(string id) const;
	bool isRoot(FactorTreeNode* node) const;
	bool isRoot(string id) const;
	bool isLeaf(FactorTreeNode* node) const;
	bool isLeaf(string id) const;
	int nodeSize() const;
	
	FactorTreeNode* setRoot(FactorTreeNode* node);
	FactorTreeNode* setRoot(string id);
	FactorTreeNode* addNode(nodeType type = nodeType_factor);
	FactorTreeNode* addNode(string id, string label = "", nodeType type = nodeType_factor);
	FactorTreeNode* addNode(FactorTreeNode* node);
	FactorTreeNode* getNode(string id);
	FactorTreeNode* removeNode(string id);
	FactorTreeNode* removeNode(FactorTreeNode* node);
	bool deleteNode(string id);
	
	vector<FactorTreeNode*> getNodes();
	vector<string> getNodesId() const;
	vector<FactorTreeNode*> getFactors();
	vector<string> getFactorsId() const;
	vector<FactorTreeNode*> getVariables();
	vector<string> getVariablesId() const;
	FactorTreeNode* getRoot();
	string getRootId() const;
	vector<FactorTreeNode*> getLeaves();
	vector<string> getLeavesId() const;
	FactorTreeNode* getParent(FactorTreeNode* node);
	FactorTreeNode* getParent(string id);
	string getParentId(FactorTreeNode* node) const;
	string getParentId(string id) const;
	vector<FactorTreeNode*> getChildren(FactorTreeNode* node);
	vector<FactorTreeNode*> getChildren(string id);
	vector<string> getChildrenId(FactorTreeNode* node) const;
	vector<string> getChildrenId(string id) const;
	
	bool connect(FactorTreeNode* parentNode, FactorTreeNode* childNode);
	bool connect(string parentId, string childId);
	bool isParent(FactorTreeNode* parentNode, FactorTreeNode* childNode);
	bool isParent(string parentId, string childId);
	bool disconnect(FactorTreeNode* parentNode, FactorTreeNode* childNode);
	bool disconnect(string parentId, string childId);
	void disconnectAll();
	void removeAll();
	bool renameNode(string oldId, string newId);
	bool renameNode(FactorTreeNode* node, string newId);

	/* Methods to set and get attributes */
	string getId() const;
	string setId(string id);
	string getLabel() const;
	void setLabel(string label);
	
	/* Printing and debugging */
	string toString() const;
	
protected:

	/* Attributes */
	string mId;
	string mLabel;
	string mRootId;
	map<string, FactorTreeNode*> mNodeList;
	multimap<string, string> mChildConnectionList;
	map<string, string> mParentConnectionList;

	/* Protected methods */
	string generateNodeId(nodeType type = nodeType_factor);

private:
	void copyFrom(const FactorTree& src);
	
};

class FactorTreeNode {

public:

	/* Constructors and destructors */
	FactorTreeNode(string id = "", string label = "", nodeType type = nodeType_factor);
	FactorTreeNode(const FactorTreeNode& src);
	FactorTreeNode& operator=(const FactorTreeNode& src);
	virtual ~FactorTreeNode(); 

	/* Methods to set and get attributes */
	string getId();
	string setId(string id);
	string getLabel();
	void setLabel(string label);
	nodeType getType();
	void setType(nodeType type);
	FactorTree* getParent();
	void setParent(FactorTree* parent);
	FactorTreeNode* getParentNode();
	vector<FactorTreeNode*> getChildren();
	
	/* Printing and debugging */
	string toString();
	
protected:

	/* Parent */
	FactorTree* mParent;
	
	/* Attributes */
	string mId;
	string mLabel;
	nodeType mType;

private:

	/* Friend class */
	friend class FactorTree;
	void copyFrom(const FactorTreeNode& src);
	
};

}

#endif /* FACTORTREE_H_ */
