/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: maxproductpropagator_test.cpp
**  
** Description:
** The source file storing all the specification for a test file for the simple max product belief propagation.
**  
** Last modified: 
** 06 December 2006
** 
** Comments:
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <sstream>
#include <math.h>

/* Application Includes */
#include "factorgraph.h"
#include "discretedistribution.h"
#include "discdomain.h"
#include "random.h"
#include "maxproductpropagator.h"
#include "maxproductpropagator_test.h"
// #include "maxproductpropagator_3step_test.h"

using namespace std;
using namespace factorgraph;
using namespace distribution;
using namespace propagator;
using namespace random;
using namespace domain;

/** Test for the belief propagation on a very simple pathway that has all nodes
* being observed. The sampling of data are all done in this step.
**/
void testMaxProductPropagator_SimplePathway() {

	cout << "===========================================" << endl;
	cout << "In testMaxProductPropagator_SimplePathway()" << endl;
	cout << "===========================================" << endl;
	
	MaxProductPropagator* bp = new MaxProductPropagator();
	FactorGraph* graph = new FactorGraph();
	
	cout << "Setting up factor graph..." << endl;
	
	// Sets up the graph and adds in the nodes
	graph->addNode("k1", "k1", nodeType_variable);
	graph->addNode("k2", "k2", nodeType_variable);
	graph->addNode("k3", "k3", nodeType_variable);
	graph->addNode("k4", "k4", nodeType_variable);
	graph->addNode("k5", "k5", nodeType_variable);
	graph->addNode("k6", "k6", nodeType_variable);

	graph->addNode("f1_1", "f1_1", nodeType_factor);
	graph->addNode("f1_2", "f1_2", nodeType_factor);
	graph->addNode("f1_3", "f1_3", nodeType_factor);
	graph->addNode("f1_4", "f1_4", nodeType_factor);
		
	graph->addNode("f2_1", "f2_1", nodeType_factor);
	graph->addNode("f2_2", "f2_2", nodeType_factor);
	graph->addNode("f2_3", "f2_3", nodeType_factor);
	graph->addNode("f2_4", "f2_4", nodeType_factor);

	graph->addNode("f3_1", "f3_1", nodeType_factor);
	graph->addNode("f3_2", "f3_2", nodeType_factor);
	graph->addNode("f3_3", "f3_3", nodeType_factor);
	graph->addNode("f3_4", "f3_4", nodeType_factor);

	graph->addNode("f4_1", "f4_1", nodeType_factor);
	graph->addNode("f4_2", "f4_2", nodeType_factor);
	graph->addNode("f4_3", "f4_3", nodeType_factor);
	graph->addNode("f4_4", "f4_4", nodeType_factor);
	
	// Connects the nodes
	graph->connect("f1_1", "k1");
	graph->connect("f1_1", "k2");

	graph->connect("f1_2", "k1");
	graph->connect("f1_2", "k2");

	graph->connect("f1_3", "k1");
	graph->connect("f1_3", "k2");

	graph->connect("f1_4", "k1");
	graph->connect("f1_4", "k2");

	graph->connect("f2_1", "k2");
	graph->connect("f2_1", "k3");
	graph->connect("f2_1", "k4");

	graph->connect("f2_2", "k2");
	graph->connect("f2_2", "k3");
	graph->connect("f2_2", "k4");

	graph->connect("f2_3", "k2");
	graph->connect("f2_3", "k3");
	graph->connect("f2_3", "k4");

	graph->connect("f2_4", "k2");
	graph->connect("f2_4", "k3");
	graph->connect("f2_4", "k4");

	graph->connect("f3_1", "k3");
	graph->connect("f3_1", "k5");
	
	graph->connect("f3_2", "k3");
	graph->connect("f3_2", "k5");

	graph->connect("f3_3", "k3");
	graph->connect("f3_3", "k5");

	graph->connect("f3_4", "k3");
	graph->connect("f3_4", "k5");
	
	graph->connect("f4_1", "k4");
	graph->connect("f4_1", "k6");

	graph->connect("f4_2", "k4");
	graph->connect("f4_2", "k6");

	graph->connect("f4_3", "k4");
	graph->connect("f4_3", "k6");

	graph->connect("f4_4", "k4");
	graph->connect("f4_4", "k6");
	
	cout << "Factor graph..." << endl;	
	cout << graph->toString() << endl;	
	
	// Loads factor graph with propagator
	bp->loadFactorGraph(graph);
	
	/////////////////////////////////////////////////////////////////
	// The pathway that we are going to take is based on a simple pathway with branch.
	// The set of differential equations that define the pathway is as follows
	// 	0 = x1_dot - (k1 - k2.x1)
	// 	0 = x2_dot - (k2.x1 - [k3 + k4].x2)
	// 	0 = x3_dot - (k3.x2 - k5.x3)
	// 	0 = x4_dot - (k4.x2 - k6.x4)
	//
	// For this example we are going to assume that we know ALL the points
	// i.e. the variable nodes corresponding to the observed nodes will not be 
	// updated. 
	// The data are as follows
	// Time		2		4		6		8
	// x1			0.431		0.537	0.563	0.569
	// x1_dot		0.0982	0.0241	0.00591	0.00145
	// x2			0.248	0.487	0.6		0.64
	// x2_dot		0.153		0.084	0.034	0.0126
	// x3			0.086	0.341		0.606	0.805
	// x3_dot		0.0981	0.141		0.118		0.0803
	// x4			0.0181	0.0760	0.142		0.199
	// x4_dot		0.0211	0.0335	0.0315	0.0246
	//
	// and the nominal parameters that we want to estimate for are
	// k1 : 0.4
	// k2 : 0.7
	// k3 : 0.5 
	// k4 : 0.1
	// k5 : 0.3
	// k6 : 0.2
	//
	/////////////////////////////////////////////////////////////////
	
	// Sample the various points for a number of times
	double k1, k2, k3, k4, k5, k6;
	double x1;
	double x1_dot;
	double x2;
	double x2_dot;
	double x3;
	double x3_dot;
	double x4;
	double x4_dot;
	
	DiscreteDistribution* dist_k1 = new DiscreteDistribution();
	DiscreteDistribution* dist_k2 = new DiscreteDistribution();
	DiscreteDistribution* dist_k3 = new DiscreteDistribution();
	DiscreteDistribution* dist_k4 = new DiscreteDistribution();	
	DiscreteDistribution* dist_k5 = new DiscreteDistribution();
	DiscreteDistribution* dist_k6 = new DiscreteDistribution();
	
	DiscreteDistribution* dist_f1_1 = new DiscreteDistribution();
	DiscreteDistribution* dist_f1_2 = new DiscreteDistribution();
	DiscreteDistribution* dist_f1_3 = new DiscreteDistribution();
	DiscreteDistribution* dist_f1_4 = new DiscreteDistribution();
	DiscreteDistribution* dist_f2_1 = new DiscreteDistribution();
	DiscreteDistribution* dist_f2_2 = new DiscreteDistribution();
	DiscreteDistribution* dist_f2_3 = new DiscreteDistribution();
	DiscreteDistribution* dist_f2_4 = new DiscreteDistribution();
	DiscreteDistribution* dist_f3_1 = new DiscreteDistribution();
	DiscreteDistribution* dist_f3_2 = new DiscreteDistribution();
	DiscreteDistribution* dist_f3_3 = new DiscreteDistribution();
	DiscreteDistribution* dist_f3_4 = new DiscreteDistribution();
	DiscreteDistribution* dist_f4_1 = new DiscreteDistribution();
	DiscreteDistribution* dist_f4_2 = new DiscreteDistribution();
	DiscreteDistribution* dist_f4_3 = new DiscreteDistribution();
	DiscreteDistribution* dist_f4_4 = new DiscreteDistribution();
	
	// Distributions for the parameters - uniform
	dist_k1->addVariable("k1", 10);
	dist_k1->initializeTable();
	dist_k1->makeUniform();

	dist_k2->addVariable("k2", 10);
	dist_k2->initializeTable();
	dist_k2->makeUniform();

	dist_k3->addVariable("k3", 10);
	dist_k3->initializeTable();
	dist_k3->makeUniform();

	dist_k4->addVariable("k4", 10);
	dist_k4->initializeTable();
	dist_k4->makeUniform();

	dist_k5->addVariable("k5", 10);
	dist_k5->initializeTable();
	dist_k5->makeUniform();

	dist_k6->addVariable("k6", 10);
	dist_k6->initializeTable();
	dist_k6->makeUniform();

	// 	0 = x1_dot - (k1 - k2.x1)
	// 	0 = x2_dot - (k2.x1 - [k3 + k4].x2)
	// 	0 = x3_dot - (k3.x2 - k5.x3)
	// 	0 = x4_dot - (k4.x2 - k6.x4)
	
	// Factor distributions
	dist_f1_1->addVariable("k1", 10);
	dist_f1_1->addVariable("k2", 10);
	dist_f1_1->initializeTable();

	dist_f1_2->addVariable("k1", 10);
	dist_f1_2->addVariable("k2", 10);
	dist_f1_2->initializeTable();
	
	dist_f1_3->addVariable("k1", 10);
	dist_f1_3->addVariable("k2", 10);
	dist_f1_3->initializeTable();

	dist_f1_4->addVariable("k1", 10);
	dist_f1_4->addVariable("k2", 10);
	dist_f1_4->initializeTable();

	dist_f2_1->addVariable("k2", 10);
	dist_f2_1->addVariable("k3", 10);
	dist_f2_1->addVariable("k4", 10);
	dist_f2_1->initializeTable();

	dist_f2_2->addVariable("k2", 10);
	dist_f2_2->addVariable("k3", 10);
	dist_f2_2->addVariable("k4", 10);
	dist_f2_2->initializeTable();

	dist_f2_3->addVariable("k2", 10);
	dist_f2_3->addVariable("k3", 10);
	dist_f2_3->addVariable("k4", 10);
	dist_f2_3->initializeTable();

	dist_f2_4->addVariable("k2", 10);
	dist_f2_4->addVariable("k3", 10);
	dist_f2_4->addVariable("k4", 10);
	dist_f2_4->initializeTable();

	dist_f3_1->addVariable("k3", 10);
	dist_f3_1->addVariable("k5", 10);
	dist_f3_1->initializeTable();

	dist_f3_2->addVariable("k3", 10);
	dist_f3_2->addVariable("k5", 10);
	dist_f3_2->initializeTable();

	dist_f3_3->addVariable("k3", 10);
	dist_f3_3->addVariable("k5", 10);
	dist_f3_3->initializeTable();

	dist_f3_4->addVariable("k3", 10);
	dist_f3_4->addVariable("k5", 10);
	dist_f3_4->initializeTable();

	dist_f4_1->addVariable("k4", 10);
	dist_f4_1->addVariable("k6", 10);
	dist_f4_1->initializeTable();

	dist_f4_2->addVariable("k4", 10);
	dist_f4_2->addVariable("k6", 10);
	dist_f4_2->initializeTable();

	dist_f4_3->addVariable("k4", 10);
	dist_f4_3->addVariable("k6", 10);
	dist_f4_3->initializeTable();

	dist_f4_4->addVariable("k4", 10);
	dist_f4_4->addVariable("k6", 10);
	dist_f4_4->initializeTable();
	
	// Seed the Random number generator
	MersenneTwister::init(time(NULL));
	
	double score_f1_1;
	double score_f1_2;
	double score_f1_3;
	double score_f1_4;
	double score_f2_1;
	double score_f2_2;
	double score_f2_3;
	double score_f2_4;
	double score_f3_1;
	double score_f3_2;
	double score_f3_3;
	double score_f3_4;
	double score_f4_1;
	double score_f4_2;
	double score_f4_3;
	double score_f4_4;

	// x1			0.431		0.537	0.563	0.569
	// x1_dot		0.0982	0.0241	0.00591	0.00145
	// x2			0.248	0.487	0.6		0.64
	// x2_dot		0.153		0.084	0.034	0.0126
	// x3			0.086	0.341		0.606	0.805
	// x3_dot		0.0981	0.141		0.118		0.0803
	// x4			0.0181	0.0760	0.142		0.199
	// x4_dot		0.0211	0.0335	0.0315	0.0246
	
	double obs[8][4];
	obs[0][0] = 0.431;	obs[0][1] = 0.537;	obs[0][2] = 0.563;	obs[0][3] = 0.569;
	obs[1][0] = 0.0982;	obs[1][1] = 0.0241;	obs[1][2] = 0.00591;	obs[1][3] = 0.00145;
	obs[2][0] = 0.248;	obs[2][1] = 0.487;	obs[2][2] = 0.6;	obs[2][3] = 0.64;
	obs[3][0] = 0.153;	obs[3][1] = 0.084;	obs[3][2] = 0.034;	obs[3][3] = 0.0126;
	obs[4][0] = 0.086;	obs[4][1] = 0.341;	obs[4][2] = 0.606;	obs[4][3] = 0.805;
	obs[5][0] = 0.0981;	obs[5][1] = 0.141;	obs[5][2] = 0.118;	obs[5][3] = 0.0803;
	obs[6][0] = 0.0181;	obs[6][1] = 0.0760;	obs[6][2] = 0.142;	obs[6][3] = 0.199;
	obs[7][0] = 0.0211;	obs[7][1] = 0.0335;	obs[7][2] = 0.0315;	obs[7][3] = 0.0246;
	
	int i_x1 = 0;
	int i_x1_dot = 1;
	int i_x2 = 2;
	int i_x2_dot = 3;
	int i_x3 = 4;
	int i_x3_dot = 5;
	int i_x4 = 6;
	int i_x4_dot = 7;
	
	string f1_1_expr;
	string f1_2_expr;
	string f1_3_expr;
	string f1_4_expr;
	string f2_1_expr;
	string f2_2_expr;
	string f2_3_expr;
	string f2_4_expr;
	string f3_1_expr;
	string f3_2_expr;
	string f3_3_expr;
	string f3_4_expr;
	string f4_1_expr;
	string f4_2_expr;
	string f4_3_expr;
	string f4_4_expr;
	
	// Declare domains for the parameters
	DiscDomain dom(0.0, 1.0);
	dom.repartition(10);
	
	cout << "Start sampling..." << endl;
	
	for (int i = 0; i < 100000; i++) {
		
		// Randomly generates values from between [0, 1] for k1, k2, k3, k4, k5 and k6
		k1 = dom.random();
		k2 = dom.random();
		k3 = dom.random();
		k4 = dom.random();
		k5 = dom.random();
		k6 = dom.random();
		
		// Printouts for sanity check
		if (i % 10000 == 0) {
			cout << "Iteration " << i << " : " << k1 << " " << k2 << " " << k3 << " " << k4 << " " << k5 << " " << k6 << endl;
		}
		
		// Computes the score
		score_f1_1 = samplef1(obs[i_x1_dot][0], obs[i_x1][0], k1, k2);
		score_f1_2 = samplef1(obs[i_x1_dot][1], obs[i_x1][1], k1, k2);
		score_f1_3 = samplef1(obs[i_x1_dot][2], obs[i_x1][2], k1, k2);
		score_f1_4 = samplef1(obs[i_x1_dot][3], obs[i_x1][3], k1, k2);
		
		score_f2_1 = samplef2(obs[i_x2_dot][0], obs[i_x1][0], obs[i_x2][0], k2, k3, k4);
		score_f2_2 = samplef2(obs[i_x2_dot][1], obs[i_x1][1], obs[i_x2][1], k2, k3, k4);
		score_f2_3 = samplef2(obs[i_x2_dot][2], obs[i_x1][2], obs[i_x2][2], k2, k3, k4);
		score_f2_4 = samplef2(obs[i_x2_dot][3], obs[i_x1][3], obs[i_x2][3], k2, k3, k4);
		
		score_f3_1 = samplef3(obs[i_x3_dot][0], obs[i_x2][0], obs[i_x3][0], k3, k5);
		score_f3_2 = samplef3(obs[i_x3_dot][1], obs[i_x2][1], obs[i_x3][1], k3, k5);
		score_f3_3 = samplef3(obs[i_x3_dot][2], obs[i_x2][2], obs[i_x3][2], k3, k5);
		score_f3_4 = samplef3(obs[i_x3_dot][3], obs[i_x2][3], obs[i_x3][3], k3, k5);
		
		score_f4_1 = samplef4(obs[i_x4_dot][0], obs[i_x2][0], obs[i_x4][0], k4, k6);
		score_f4_2 = samplef4(obs[i_x4_dot][1], obs[i_x2][1], obs[i_x4][1], k4, k6);
		score_f4_3 = samplef4(obs[i_x4_dot][2], obs[i_x2][2], obs[i_x4][2], k4, k6);
		score_f4_4 = samplef4(obs[i_x4_dot][3], obs[i_x2][3], obs[i_x4][3], k4, k6);
		
		// Increments the distribution tables
		f1_1_expr = "k1 " + itos(dom.indexOf(k1)) + " k2 " + itos(dom.indexOf(k2));
		f1_2_expr = "k1 " + itos(dom.indexOf(k1)) + " k2 " + itos(dom.indexOf(k2));
		f1_3_expr = "k1 " + itos(dom.indexOf(k1)) + " k2 " + itos(dom.indexOf(k2));
		f1_4_expr = "k1 " + itos(dom.indexOf(k1)) + " k2 " + itos(dom.indexOf(k2));
		f2_1_expr = "k2 " + itos(dom.indexOf(k2)) + " k3 " + itos(dom.indexOf(k3)) + " k4 " + itos(dom.indexOf(k4));
		f2_2_expr = "k2 " + itos(dom.indexOf(k2)) + " k3 " + itos(dom.indexOf(k3)) + " k4 " + itos(dom.indexOf(k4));
		f2_3_expr = "k2 " + itos(dom.indexOf(k2)) + " k3 " + itos(dom.indexOf(k3)) + " k4 " + itos(dom.indexOf(k4));
		f2_4_expr = "k2 " + itos(dom.indexOf(k2)) + " k3 " + itos(dom.indexOf(k3)) + " k4 " + itos(dom.indexOf(k4));
		f3_1_expr = "k3 " + itos(dom.indexOf(k3)) + " k5 " + itos(dom.indexOf(k5));
		f3_2_expr = "k3 " + itos(dom.indexOf(k3)) + " k5 " + itos(dom.indexOf(k5));
		f3_3_expr = "k3 " + itos(dom.indexOf(k3)) + " k5 " + itos(dom.indexOf(k5));
		f3_4_expr = "k3 " + itos(dom.indexOf(k3)) + " k5 " + itos(dom.indexOf(k5));
		f4_1_expr = "k4 " + itos(dom.indexOf(k4)) + " k6 " + itos(dom.indexOf(k6));
		f4_2_expr = "k4 " + itos(dom.indexOf(k4)) + " k6 " + itos(dom.indexOf(k6));
		f4_3_expr = "k4 " + itos(dom.indexOf(k4)) + " k6 " + itos(dom.indexOf(k6));
		f4_4_expr = "k4 " + itos(dom.indexOf(k4)) + " k6 " + itos(dom.indexOf(k6));
		
		dist_f1_1->incProbability(f1_1_expr, score_f1_1);
		dist_f1_2->incProbability(f1_2_expr, score_f1_2);
		dist_f1_3->incProbability(f1_3_expr, score_f1_3);
		dist_f1_4->incProbability(f1_4_expr, score_f1_4);

		dist_f2_1->incProbability(f2_1_expr, score_f2_1);
		dist_f2_2->incProbability(f2_2_expr, score_f2_2);
		dist_f2_3->incProbability(f2_3_expr, score_f2_3);
		dist_f2_4->incProbability(f2_4_expr, score_f2_4);

		dist_f3_1->incProbability(f3_1_expr, score_f3_1);
		dist_f3_2->incProbability(f3_2_expr, score_f3_2);
		dist_f3_3->incProbability(f3_3_expr, score_f3_3);
		dist_f3_4->incProbability(f3_4_expr, score_f3_4);

		dist_f4_1->incProbability(f4_1_expr, score_f4_1);
		dist_f4_2->incProbability(f4_2_expr, score_f4_2);
		dist_f4_3->incProbability(f4_3_expr, score_f4_3);
		dist_f4_4->incProbability(f4_4_expr, score_f4_4);
		
	}
	
	cout << "Done sampling..." << endl;
	
	// Normalize the factor distributions
	dist_f1_1->normalize();
	dist_f1_2->normalize();
	dist_f1_3->normalize();
	dist_f1_4->normalize();

	dist_f2_1->normalize();
	dist_f2_2->normalize();
	dist_f2_3->normalize();
	dist_f2_4->normalize();

	dist_f3_1->normalize();
	dist_f3_2->normalize();
	dist_f3_3->normalize();
	dist_f3_4->normalize();

	dist_f4_1->normalize();
	dist_f4_2->normalize();
	dist_f4_3->normalize();
	dist_f4_4->normalize();
	
	// Associates distributions
	bp->setDistribution("k1", dist_k1);
	bp->setDistribution("k2", dist_k2);
	bp->setDistribution("k3", dist_k3);
	bp->setDistribution("k4", dist_k4);
	bp->setDistribution("k5", dist_k5);
	bp->setDistribution("k6", dist_k6);
	
	bp->setDistribution("f1_1", dist_f1_1);
	bp->setDistribution("f1_2", dist_f1_2);
	bp->setDistribution("f1_3", dist_f1_3);
	bp->setDistribution("f1_4", dist_f1_4);

	bp->setDistribution("f2_1", dist_f2_1);
	bp->setDistribution("f2_2", dist_f2_2);
	bp->setDistribution("f2_3", dist_f2_3);
	bp->setDistribution("f2_4", dist_f2_4);

	bp->setDistribution("f3_1", dist_f3_1);
	bp->setDistribution("f3_2", dist_f3_2);
	bp->setDistribution("f3_3", dist_f3_3);
	bp->setDistribution("f3_4", dist_f3_4);

	bp->setDistribution("f4_1", dist_f4_1);
	bp->setDistribution("f4_2", dist_f4_2);
	bp->setDistribution("f4_3", dist_f4_3);
	bp->setDistribution("f4_4", dist_f4_4);
	
	// Propagates the beliefs
	bp->propagate();
	
	// Unloads graph
	bp->unloadFactorGraph();
	
	// Deletes graph and propagator
	delete graph;
	delete bp;

}

/** Sample for f1
**/
double samplef1(double x1_dot, double x1, double k1, double k2) {
	double difference = fabs(x1_dot - (k1 - (k2 * x1))); 
	double cost = exp(-difference);
	return cost;
}

/** Sample for f2
**/
double samplef2(double x2_dot, double x1, double x2, double k2, double k3, double k4) {
	double difference = fabs(x2_dot - ((k2 * x1) - ((k3 + k4) * x2)));
	double cost = exp(-difference);
	return cost;
}

/** Sample for f3
**/
double samplef3(double x3_dot, double x2, double x3, double k3, double k5) {
	double difference = fabs(x3_dot - ((k3 * x2) - (k5 * x3)));
	double cost = exp(-difference);
	return cost;
}

/** Sample for f4
**/
double samplef4(double x4_dot, double x2, double x4, double k4, double k6) {
	double difference = fabs(x4_dot - ((k4 * x2) - (k6 * x4)));
	double cost = exp(-difference);
	return cost;
}

/** Converts int to string
**/
string itos(int i) {
	stringstream s;
	s << i;
	return s.str();
}
