/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: sumproductpropagator.cpp
**  
** Description:
** The header file storing all the specification for a class that performs belief propagation. 
**  
** Last modified: 
** 04 December 2006
** 
** Comments:
****************************************************************************/

/* Standard Includes */
#include <iostream>

/* Application Includes */
#include "sumproductpropagator.h"
#include "factorgraph.h"
#include "discretedistribution.h"

using namespace std;
using namespace factorgraph;
using namespace distribution;

namespace propagator {

/** Default constructor
**/
SumProductPropagator::SumProductPropagator() : 
	mFactorGraph(0), mTolerance(SUMPRODUCTPROPAGATOR_TOLERANCE), mMaxIterations(SUMPRODUCTPROPAGATOR_MAXITERATIONS) {
	
	#ifdef SUMPRODUCTPROPAGATOR_DEBUG
		cout << "[Info] SumProductPropagator::SumProductPropagator() - Creating SumProductPropagator" << endl;
	#endif

}

/** Copy constructor
**/
SumProductPropagator::SumProductPropagator(const SumProductPropagator& src) {
	copyFrom(src);
}

/** Assignment constructor
**/
SumProductPropagator& SumProductPropagator::operator=(const SumProductPropagator& src) {
	
	// Checks for self assignment
	if (this == &src) {
		return (*this);	
	}

	// Unloads the current factor graph
	unloadFactorGraph();
	
	// Copy contents
	copyFrom(src);
	
	return (*this);
	
}

/** Default destructor
**/
SumProductPropagator::~SumProductPropagator() {

	#ifdef SUMPRODUCTPROPAGATOR_DEBUG
		cout << "[Info] SumProductPropagator::~SumProductPropagator() - Destroying SumProductPropagator" << endl;
	#endif

	unloadFactorGraph();
}

/** Loads a factor graph
**/
bool SumProductPropagator::loadFactorGraph(const FactorGraph* graph) {
	
	if (graph == 0) {
		#ifdef SUMPRODUCTPROPAGATOR_DEBUG
			cout << "[Error] SumProductPropagator::loadFactorGraph() - Invalid factor graph" << endl;
		#endif
		return false;
	}

	// Unloads the current factor graph
	unloadFactorGraph();
	
	// Sets the factor graph
	mFactorGraph = graph;
	
}

/** Unloads a factor graph from the propagator
* The maps, queues etc. are all cleared
**/
void SumProductPropagator::unloadFactorGraph() {

	// Clears all the contents
	mFactorGraph = 0;
	map<string, DiscreteDistribution*>::iterator mapIter = mBelief.begin();
	while (mapIter != mBelief.end()) {
		delete mapIter->second;
		mapIter++;
	}
	mBelief.clear();
	mFixedList.clear();
	mBeliefQueue.clear();	// The belief queue does not require explicit deletion as the instances are the same as those in the history list. 
							// Deletion is done via the history list instead
	map<std::pair<string, string>, DiscreteDistribution*>::iterator histIter = mMessageHistory.begin();
	while (histIter != mMessageHistory.end()) {
		delete histIter->second;
		histIter++;
	}
	mMessageHistory.clear();

}

/** Set a node with the given id, with a certain distribution
* The old distribution will be deleted.
**/
bool SumProductPropagator::setDistribution(string id, DiscreteDistribution* distribution, bool isFixed) {
	
	// Checks if the factor graph is loaded
	if (mFactorGraph == 0) {
		#ifdef SUMPRODUCTPROPAGATOR_DEBUG
			cout << "[Error] SumProductPropagator::setDistribution() - Factor graph is not loaded" << endl;
		#endif
		return false;
	}
	
	// Checks if the id is valid
	if (!mFactorGraph->nodeExists(id)) {
		#ifdef SUMPRODUCTPROPAGATOR_DEBUG
			cout << "[Error] SumProductPropagator::setDistribution() - FactorGraphNode " << id << " does not exist" << endl;
		#endif
		return false;
	}
	
	// Checks if the current map already has an associated distribution
	// If it has, the old one will be deleted and the new one will be associated. Otherwise
	// it will add one to it
	map<string, DiscreteDistribution*>::iterator iter = mBelief.find(id);
	if (iter != mBelief.end()) {
		delete iter->second;
		iter->second = distribution;
	} else {
		mBelief[id] = distribution;
	}
	
	mFixedList[id] = isFixed;
	return true;
}

/** Retrieves the distribution of the node given the id
**/
DiscreteDistribution* SumProductPropagator::getDistribution(string id) {

	// Checks if the factor graph is loaded
	if (mFactorGraph == 0) {
		#ifdef SUMPRODUCTPROPAGATOR_DEBUG
			cout << "[Error] SumProductPropagator::getDistribution() - Factor graph is not loaded" << endl;
		#endif
		return 0;
	}
	
	// Checks if the id is valid
	if (!mFactorGraph->nodeExists(id)) {
		#ifdef SUMPRODUCTPROPAGATOR_DEBUG
			cout << "[Error] SumProductPropagator::getDistribution() - FactorGraphNode " << id << " does not exist" << endl;
		#endif
		return 0;
	}

	// Finds the distribution
	map<string, DiscreteDistribution*>::const_iterator iter = mBelief.find(id);
	if (iter != mBelief.end()) {
		return iter->second;
	} else {
		#ifdef SUMPRODUCTPROPAGATOR_DEBUG
			cout << "[Info] SumProductPropagator::getDistribution() - Distribution for FactorGraphNode " << id << " is not set" << endl;
		#endif
		return 0;
	}
	
}

/** Checks if a variable has a fixed distribution
**/
bool SumProductPropagator::isFixed(string id) {

	// Checks if the factor graph is loaded
	if (mFactorGraph == 0) {
		#ifdef SUMPRODUCTPROPAGATOR_DEBUG
			cout << "[Error] SumProductPropagator::isFixed() - Factor graph is not loaded" << endl;
		#endif
		return false;
	}
	
	// Checks if the id is valid
	if (!mFactorGraph->nodeExists(id)) {
		#ifdef SUMPRODUCTPROPAGATOR_DEBUG
			cout << "[Error] SumProductPropagator::isFixed() - FactorGraphNode " << id << " does not exist" << endl;
		#endif
		return false;
	}

	map<string, bool>::const_iterator iter = mFixedList.find(id);
	if (iter == mFixedList.end()) {
		#ifdef SUMPRODUCTPROPAGATOR_DEBUG
			cout << "[Info] SumProductPropagator::isFixed() - Distribution for FactorGraphNode " << id << " is not set" << endl;
		#endif
		return false;
	} else {
		return iter->second;
	}
}

/** Sets the tolerance level for halting propagation
* On error, returns -1.0
**/
double SumProductPropagator::setTolerance(double tolerance) {
	if (tolerance < 0.0) {
		return -1.0;
	} else {
		mTolerance = tolerance;
		return mTolerance;
	}
}

/** Gets the tolerance
**/
double SumProductPropagator::getTolerance() {
	return mTolerance;
}

/** Sets the maximum number of iterations for propagation
* On error, returns -1
**/
int SumProductPropagator::setMaxIterations(int iterations) {
	if (iterations < 1) {
		return -1;
	} else {
		mMaxIterations = iterations;
		return mMaxIterations;
	}
}

/** Gets the maximum number of iterations
**/
int SumProductPropagator::getMaxIterations() {
	return mMaxIterations;
}

/** Propagates the beliefs throughout the network
* This algorithm follows a two phase approach for each iteration
* 1) Performs a computation on all the factor nodes
* 2) Updates each variable nodes
**/	
bool SumProductPropagator::propagate() {
	
	// Checks if the factor graph is already set, and that all the 
	// nodes are properly set up with their respective distributions
	// i.e. for each factor node, we have to get the variable list and the size and check with their connected
	// variable node
	if (mFactorGraph == 0) {
		#ifdef SUMPRODUCTPROPAGATOR_DEBUG
			cout << "[Error] SumProductPropagator::propagate() - Factor graph is not loaded" << endl;
		#endif
		return false;
	}
	
	if (!mFactorGraph->isConsistent()) {
		#ifdef SUMPRODUCTPROPAGATOR_DEBUG
			cout << "[Error] SumProductPropagator::propagate() - Factor graph is not consistent" << endl;
		#endif
		return false;
	}
	
	string currFactor, currVariable;
	map<string, DiscreteDistribution*> bufferMap;
	vector<string> neighborList;
	vector<string> marginalList;
	vector<string> factorList = mFactorGraph->getFactorsId();
	vector<string> variableList = mFactorGraph->getVariablesId();
	DiscreteDistribution* factorDistribution;
	DiscreteDistribution* variableDistribution;
	DiscreteDistribution* marginalDistribution;
	
	// Checks the factor list to ensure that all the distributions are present
	for (int i = 0; i < factorList.size(); i++) {
	
		currFactor = factorList.at(i);
		factorDistribution = getDistribution(currFactor);
		
		if (factorDistribution == 0 || !factorDistribution->isInitialized()) {
			#ifdef SUMPRODUCTPROPAGATOR_DEBUG
				cout << "[Error] SumProductPropagator::propagate() - Distribution for factor " << factorList.at(i) << " is not set" << endl;
			#endif
			return false;
		}
	}
	
	#ifdef SUMPRODUCTPROPAGATOR_DEBUG
		cout << "[Info] SumProductPropagator::propagate() - Distribution for factor nodes checked" << endl;
	#endif
	
	// Checks the variable list to ensure that all the distributions are present, and that it is compatible with its
	// connected factors
	for (int i = 0; i < variableList.size(); i++) {
		currVariable = variableList.at(i);
		variableDistribution = getDistribution(currVariable);
		if (variableDistribution == 0 || !variableDistribution->isInitialized() || variableDistribution->getDimension() != 1) {
			#ifdef SUMPRODUCTPROPAGATOR_DEBUG
				cout << "[Error] SumProductPropagator::propagate() - Distribution for variable " << currVariable << " is not set" << endl;
			#endif
			return false;
		}
		
		neighborList = mFactorGraph->getNeighborsId(currVariable);
		
		for (int j = 0; j < neighborList.size(); j++) {
			factorDistribution = getDistribution(neighborList.at(j));
			if (variableDistribution->getSize(currVariable) != factorDistribution->getSize(currVariable) ||
				variableDistribution->getSize(currVariable) == -1)  {
				#ifdef SUMPRODUCTPROPAGATOR_DEBUG
					cout << "[Error] SumProductPropagator::propagate() - Incompatible distribution between variable " << 
						    currVariable << " and factor " << neighborList.at(j) << endl;
					return false;
				#endif
			}
		}
	}

	#ifdef SUMPRODUCTPROPAGATOR_DEBUG
		cout << "[Info] SumProductPropagator::propagate() - Distribution for variable nodes checked" << endl;
	#endif

	// Sets up the history list. This step is separate from the checking as it has to be emptied and deleted eventually
	// The history list is initially set up as the same distribution as the variable nodes
	for (int i = 0; i < variableList.size(); i++) {
		currVariable = variableList.at(i);
		neighborList = mFactorGraph->getNeighborsId(currVariable);
		variableDistribution = getDistribution(currVariable);
		for (int j = 0; j < neighborList.size(); j++) {
			mMessageHistory[make_pair(currVariable, neighborList.at(j))] = 
				new DiscreteDistribution(*variableDistribution);
			mMessageHistory[make_pair(neighborList.at(j), currVariable)] = 
				new DiscreteDistribution(*variableDistribution);
		}
		
		// Sets up the belief queue
		mBeliefQueue[currVariable];
	}

	// Sets up the buffer list. This step is separate from the checking as it has to be emptied and deleted eventually
	// The buffer map is the working buffers for the distributions of the various factor nodes
	for (int i = 0; i < factorList.size(); i++) {
		currFactor = factorList.at(i);
		factorDistribution = getDistribution(currFactor);
		bufferMap[currFactor] = new DiscreteDistribution(*factorDistribution);
	}
	
	// while max number of iterations not reached or it has not converged
	bool converged;
	int currIteration = 0;

	#ifdef SUMPRODUCTPROPAGATOR_DEBUG
		cout << "[Info] SumProductPropagator::propagate() - History and queue set up" << endl;
	#endif
	
	while (currIteration < mMaxIterations && !converged) {

		#ifdef SUMPRODUCTPROPAGATOR_DEBUG
			cout << "[Info] SumProductPropagator::propagate() - Iteration " << currIteration << endl;
		#endif
	
		// Sets converged to true first
		converged = true;
	
		// For each factor nodes, get the beliefs from the variable nodes
		for (int i = 0; i < factorList.size(); i++) {
		
			currFactor = factorList.at(i);
			factorDistribution = getDistribution(currFactor);
			
			// Get the neighbors
			neighborList = mFactorGraph->getNeighborsId(currFactor);
			
			// For each variable neighbor, get a copy of its distribution and update the history list
			for (int j = 0; j < neighborList.size(); j++) {
				currVariable = neighborList.at(j);
				variableDistribution = getDistribution(currVariable);
				mMessageHistory[make_pair(currVariable, currFactor)]->replace(*variableDistribution);
			}
			
			// For each variable neighbor, generates a message from its n-1 variable nodes, generates by multiplying by the rest of the distributions
			// i.e. if factor i has variable nodes a, b and c, it will generate for a, the product of b and c with the distribution and normalize it along b and c
			// to generate the message for a. Then it will repeat this step for b and c. Note that if the variable node is fixed, this step can be safely skipped
			for (int j = 0; j < neighborList.size(); j++) {
			
				currVariable = neighborList.at(j);

				if (!mFixedList[currVariable]) {
				
					// Replaces the distribution in the buffer map by the original distribution first
					bufferMap[currFactor]->replaceNoCheck(*factorDistribution);
					marginalList.clear();

					for (int k = 0; k < neighborList.size(); k++) {
						if (k != j) {
							bufferMap[currFactor]->multiply(mMessageHistory[make_pair(neighborList.at(k), currFactor)], false);
							marginalList.push_back(neighborList.at(k));
						}
					}

					bufferMap[currFactor]->normalize();
					marginalDistribution = bufferMap[currFactor]->marginalize(marginalList);
				
					// Checks the marginalDistribution with the one against the history list from <factor, variable>. If it is within the tolerance
					// nothing will be done. Otherwise, it will add into the queue, update history and set converged to false
					double tolerance = marginalDistribution->difference(mMessageHistory[make_pair(currFactor, currVariable)]);

					#ifdef SUMPRODUCTPROPAGATOR_DEBUG
						cout << "[Info] SumProductPropagator::propagate() - Tolerance for <" 
							 << currFactor << ", " << currVariable << "> : " << tolerance << endl;
					#endif

					if (tolerance < mTolerance) {
						// Do nothing
					} else {
						converged = false;
						// Copies the contents of the marginal distribution to the message history
						mMessageHistory[make_pair(currFactor, currVariable)]->replaceNoCheck(*marginalDistribution);
						queue<DiscreteDistribution*> msgQueue = mBeliefQueue[currVariable];
						msgQueue.push(mMessageHistory[make_pair(currFactor, currVariable)]);
						mBeliefQueue[currVariable] = msgQueue;
					}

					// Deletes the marginal as its values are already copied to the history
					delete marginalDistribution;

				} else {
					// Variable is fixed. Need not process
				}
			}
		}

		// For each variable node, get the beliefs off the belief queue and performs multiplication
		for (int i = 0; i < variableList.size(); i++) {
			currVariable = variableList.at(i);

			if (!mFixedList[currVariable]) {

				variableDistribution = getDistribution(currVariable);
			
				while(!mBeliefQueue[currVariable].empty()) {
					queue<DiscreteDistribution*> msgQueue = mBeliefQueue[currVariable];
					DiscreteDistribution* msg = msgQueue.front();
					msgQueue.pop();
					mBeliefQueue[currVariable] = msgQueue;
					variableDistribution->multiply(msg, false);
				}
				variableDistribution->normalize();
				
			} else {
				// Node is fixed. Do not update
			}
		}

		#ifdef SUMPRODUCTPROPAGATOR_DEBUG
			cout << "[Info] SumProductPropagator::propagate() - Current beliefs : " << endl;
			for (int i = 0; i < variableList.size(); i++) {
				currVariable = variableList.at(i);
				cout << mBelief[currVariable]->toStringSimple() << endl;
			}
		#endif
					
		currIteration++;
	}
	
	// Deallocates all the memory by wiping off the history list and 
	// queue list and the buffer map
	mBeliefQueue.clear();	
	map<std::pair<string, string>, DiscreteDistribution*>::iterator histIter = mMessageHistory.begin();
	while (histIter != mMessageHistory.end()) {
		delete histIter->second;
		histIter++;
	}
	mMessageHistory.clear();
	map<string, DiscreteDistribution*>::iterator buffIter = bufferMap.begin();
	while (buffIter != bufferMap.end()) {
		delete buffIter->second;
		buffIter++;
	}
	bufferMap.clear();
	
	return true;
	
}

/** Helper function for copying contents
* Note that items such as the history and queue list
* will not be copied
**/
void SumProductPropagator::copyFrom(const SumProductPropagator& src) {
	mFactorGraph = src.mFactorGraph;
	mTolerance = src.mTolerance;
	mMaxIterations = src.mMaxIterations;
	
	// Copy the contents (new instance) of the distribution
	map<string, DiscreteDistribution*>::const_iterator iter = src.mBelief.begin();
	while (iter != src.mBelief.end()) {
		DiscreteDistribution* distribution = new DiscreteDistribution(*(iter->second));
		mBelief[iter->first] = distribution;
		iter++;
	}
	
	// Copy the fix list
	map<string, bool>::const_iterator fixIter = src.mFixedList.begin();
	while (fixIter != src.mFixedList.end()) {
		mFixedList[fixIter->first] = fixIter->second;
		fixIter++;
	}
}

}
