/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: sumproductpropagator.h
**  
** Description:
** The header file storing all the specification for a class that performs sum product belief propagation. 
**  
** Last modified: 
** 26 November 2006
** 
** Comments:
** 1) Note that this propagator does not care about evidences, domains etc. All that it sees
**     and cares about are a) Underlying Factor graph, and b) The discrete distribution objects
**     that it will use to propagation (i.e populate queues) and update with
** 2) The propagator does not 'own' the factor graph. It only 'owns' the distribution and hence
**     it will not be responsible for cleaning up the factor graph.
** 3) The algorithm for this propagator is very simple. It performs a naive propagation using
**      the below algorithm [29 Oct 2006]:
**      While not converged or max iterations not reached
**         For each factor compute beliefs for each variable
**         For each variable, update the beliefs and compute for convergence
**         Increment iteration count
** 4) Added a fixed list to indicate whether or not a variable is fixed
****************************************************************************/

#ifndef SUMPRODUCTPROPAGATOR_H_
#define SUMPRODUCTPROPAGATOR_H_

/* For debugging. Comment out to suppress debugging lines */
#define SUMPRODUCTPROPAGATOR_DEBUG

/* Constants */
#define SUMPRODUCTPROPAGATOR_TOLERANCE 1e-5
#define SUMPRODUCTPROPAGATOR_MAXITERATIONS 10000

/* Standard Includes */
#include <iostream>
#include <vector>
#include <map>
#include <queue>

/* Application Includes */
#include "factorgraph.h"
#include "discretedistribution.h"

using namespace std;
using namespace factorgraph;
using namespace distribution;

namespace propagator {

class SumProductPropagator {

public:

	/* Constructors and destructors */
	SumProductPropagator();
	SumProductPropagator(const SumProductPropagator& src);	
	SumProductPropagator& operator=(const SumProductPropagator& src);
	virtual ~SumProductPropagator(); 

	/* Main methods to set up the propagator */
	bool loadFactorGraph(const FactorGraph* graph);
	void unloadFactorGraph();
	bool setDistribution(string id, DiscreteDistribution* distribution, bool isFixed = false);
	DiscreteDistribution* getDistribution(string id);
	bool isFixed(string id);
	double setTolerance(double tolerance);
	double getTolerance();
	int setMaxIterations(int iterations);
	int getMaxIterations();
	
	/* Method to start the propagation */
	bool propagate();
	
protected:

	const FactorGraph* mFactorGraph;
	map<string, DiscreteDistribution*> mBelief;								// Messages here will be DESTROYED AFTER THE WHOLE THING
	map<string, bool> mFixedList;											// Map to whether or not a variable is to be updated
	map<string, queue<DiscreteDistribution*> > mBeliefQueue;				// Messages here will not be destroyed
	map<std::pair<string, string>, DiscreteDistribution*> mMessageHistory;	// Messages here will BE DESTROYED
	
	/* Inference specific attributes */
	double mTolerance;
	int mMaxIterations;
	
	void copyFrom(const SumProductPropagator& src);
	
};

}

#endif
