/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: matrix.cpp
**  
** Description:
** The source file for the matrix classes
**  
** Last modified: 
** 02 March 2007
** 
** Comments:
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <iomanip>
#include <vector>
#include <sstream>
#include <math.h>

/* Application Includes */
#include "matrix.h"

using namespace std;

namespace matrix {

/** Default constructor
* Note that the default value of rows and cols are 1. If given less than that, 
* It will print out a warning and sets them as 1
**/
Matrix::Matrix(int rows, int cols) : mBufferData(0) {
	
	if (rows < 1) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] Matrix::Matrix() - Rows less than 1" << endl;
		#endif
		rows = 1;
	}
	
	if (cols < 1) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] Matrix::Matrix() - Cols less than 1" << endl;
		#endif
		cols = 1;
	}

	#ifdef MATRIX_DEBUG
		cout << "[Info] Matrix::Matrix() - Creating " << rows << " x " << cols << " Matrix..." << endl;
	#endif
	
	// Assign the number of rows and cols and allocates the correct amount of memory
	mRows = rows;
	mCols = cols;
	mData = new double*[mRows];
	for (int i = 0; i < mRows; i++) {
		mData[i] = new double[mCols];
		for (int j = 0; j < mCols; j++) {
			mData[i][j] = 0.0;
		}
	}
	
}

/** Copy constructor
**/
Matrix::Matrix(const Matrix& src) {
	copyFrom(src);
}

/** Assignment constructor
**/
Matrix& Matrix::operator=(const Matrix& src) {

	// Checks for self assignment
	if (this == &src) {
		return (*this);	
	}
	
	// Clears all the entries in the data matrix
	for (int i = 0; i < mRows; i++) {
		delete[] mData[i];
	}
	delete[] mData;

	// Clears the buffer data if it is not 0
	if (mBufferData != 0) {
		for (int i = 0; i < mRows; i++) {
			delete[] mBufferData[i];
		}
		delete[] mBufferData;
	}
	
	// Copy contents
	copyFrom(src);
	
	return (*this);
}

/** Destructor
**/
Matrix::~Matrix() {

	#ifdef MATRIX_DEBUG
		cout << "[Info] Matrix::~Matrix() - Destroying Matrix" << endl;
	#endif

	// Clears all the entries in the data matrix
	for (int i = 0; i < mRows; i++) {
		delete[] mData[i];
	}
	delete[] mData;

	// Clears all the entries in the buffer array
	if (mBufferData != 0) {
		for (int i = 0; i < mRows; i++) {
			delete[] mBufferData[i];
		}
		delete[] mBufferData;
	}
}

/** Clears the matrix
* This method sets all the entries to 0
**/
void Matrix::clear() {
	setValue(0.0);
}

/** Sets all the entries to zeros
* Essentially this is the same as clear().
**/
void Matrix::zeros() {
	setValue(0.0);
}

/** Sets all the entries to ones
**/
void Matrix::ones() {
	setValue(1.0);
}

/** Sets all the entries to the given value
**/
void Matrix::setValue(double value) {
	for (int i = 0; i < mRows; i++) {
		for (int j = 0; j < mCols; j++) {
			mData[i][j] = value;
		}
	}
}

/** Sets the diagonal values to be 1 and the rest to be 
* zero
**/
void Matrix::setIdentity() {
	for (int i = 0; i < mRows; i++) {
		for (int j = 0; j < mCols; j++) {
			if (i == j) { 
				mData[i][j] = 1.0;
			} else {
				mData[i][j] = 0.0;
			}
		}
	}
}

/** Fills the diagonal with the given value
**/
void Matrix::fillDiagonal(double value) {
	
	int size = (mRows < mCols) ? mRows : mCols;
	
	for (int i = 0; i < size; i++) {
		mData[i][i] = value;
	}
	
}

/** Swap the rows indicated by r1 and r2
**/
bool Matrix::swapRow(int r1, int r2) {

	// If any one of them are invalid, return false
	if (r1 < 0 || r2 < 0 || r1 >= mRows || r2 >= mRows) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] Matrix::swapRow() - Array index out of bounds" << endl;
		#endif
		return false;
	}

	// If r1 and r2 are the same, return true
	if (r1 == r2) {
		return true;
	}
	
	// Performs the swap
	double temp;
	for (int i = 0; i < mCols; i++) {
		temp = mData[r1][i];
		mData[r1][i] = mData[r2][i];
		mData[r2][i] = temp;
	}
	
	return true;
}

/** Swap the cols indicated by c1 and c2
**/
bool Matrix::swapCol(int c1, int c2) {

	// If any one of them are invalid, return false
	if (c1 < 0 || c2 < 0 || c1 >= mCols || c2 >= mCols) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] Matrix::swapCol() - Array index out of bounds" << endl;
		#endif
		return false;
	}

	// If c1 and c2 are the same, return true
	if (c1 == c2) {
		return true;
	}
	
	// Performs the swap
	double temp;
	for (int i = 0; i < mRows; i++) {
		temp = mData[i][c1];
		mData[i][c1] = mData[i][c2];
		mData[i][c2] = temp;
	}
	
	return true;

}

/** Adds up matrix1 and matrix2 and puts the contents into the current matrix
**/
bool Matrix::add(const Matrix& matrix1, const Matrix& matrix2) {

	if (!isDimensionSame(*this, matrix1) || !isDimensionSame(*this, matrix2)) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] Matrix::add() - Matrices are of different dimensions" << endl;
		#endif
		return false;
	}
	
	// Does an addition
	for (int i = 0; i < mRows; i++) {
		for (int j = 0; j < mCols; j++) {
			mData[i][j] = matrix1.mData[i][j] + matrix2.mData[i][j];
		}
	}
	
	return true;
}

/** Subtract matrix1 and matrix2 and puts the contents into the current matrix
**/
bool Matrix::subtract(const Matrix& matrix1, const Matrix& matrix2) {

	if (!isDimensionSame(*this, matrix1) || !isDimensionSame(*this, matrix2)) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] Matrix::subtract() - Matrices are of different dimensions" << endl;
		#endif
		return false;
	}
	
	// Does an subtraction
	for (int i = 0; i < mRows; i++) {
		for (int j = 0; j < mCols; j++) {
			mData[i][j] = matrix1.mData[i][j] - matrix2.mData[i][j];
		}
	}
	
	return true;
}

/** Multiplies matrix1 and matrix2 and puts the contents into the current matrix
**/
bool Matrix::multiply(const Matrix& matrix1, const Matrix& matrix2) {

	if (matrix1.cols() != matrix2.rows() || mRows != matrix1.rows() || mCols != matrix2.cols()) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] Matrix::multiply() - Matrices are Incompatible" << endl;
		#endif
		return false;
	}
	
	// Does a multiplication
	double value;
	for (int i = 0; i < mRows; i++) {
		for (int j = 0; j < matrix2.cols(); j++) {
			value = matrix1.mData[i][0] * matrix2.mData[0][j];
			for (int k = 1; k < matrix1.cols(); k++) {
				value = value + (matrix1.mData[i][k] * matrix2.mData[k][j]);
			}
			mData[i][j] = value;
		}
	}
	
	return true;
}

/** Multiplies the matrix with a column vector and places the results in another vector
**/
bool Matrix::multiplyVector(vector<double>& result, const vector<double>& colVector) const {

	if (mCols != colVector.size()) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] Matrix::multiplyVector() - Matrice and column vector are Incompatible" << endl;
		#endif
		return false;
	}

	// Checks the size of the result. If not the same then resize it
	if (result.size() != mRows) {
		result.resize(mRows);
	}
	
	// Does a multiplication
	double value;
	for (int i = 0; i < mRows; i++) {
		value = mData[i][0] * colVector[0];
		for (int j = 1; j < mCols; j++) {
			value = value + mData[i][j] * colVector[j];
		}
		result[i] = value;
	}
	
	return true;
}

/** Invert the matrix
**/
bool Matrix::inverse() {

	// First checks if the matrix is a square matrix.
	if (mRows != mCols) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] Matrix::inverse() - Matrix is not a square matrix" << endl;
		#endif
		return false;
	}

	// Checks if the buffer data is initialized. If it is not, initialize it
	if (mBufferData == 0) {
		mBufferData = new double*[mRows];
		for (int i = 0; i < mRows; i++) {
			mBufferData[i] = new double[mCols];
		}
	}
		
	// Declares the auxiliary variables
	vector<int> indexArray;
	vector<double> col;
	double sign;

	luDecompose(*this, indexArray, sign);

	// Transpose the indexArray to get the correct pivoting
	vector<int> p;
	p.resize(indexArray.size());
	for (int i = 0; i < indexArray.size(); i++) {
		p[indexArray[i]] = i;
	}
	
	// Occupies col with 0s
	for (int i = 0; i < mRows; i++) {
		col.push_back(0.0);
	}
	
	for (int j = 0; j < mCols; j++) {
		for (int i = 0; i < mRows; i++) {
			col[i] = 0.0;
		}
		col[j] = 1.0;
		solve(*this, p, col);
		for (int i = 0; i < mRows; i++) {
			mBufferData[i][j] = col[i];
		}
	}

	// Copies the contents from the buffer array
	for (int i = 0; i < mRows; i++) {
		for (int j = 0; j < mCols; j++) {
			mData[i][j] = mBufferData[i][j];
		}
	}
	
	return true;
}

/** Resize the matrix and resets all the values to 0.0
**/
bool Matrix::resize(int rows, int cols) {

	if (rows < 1 || cols < 1) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] Matrix::resize() - Invalid size" << endl;
		#endif
		return false;
	}
	
	// Deletes off the current matrix entries
	for (int i = 0; i < mRows; i++) {
		delete[] mData[i];
	}
	delete[] mData;
	
	// Resets the rows and cols
	mRows = rows;
	mCols = cols;
	mData = new double*[mRows];
	for (int i = 0; i < mRows; i++) {
		mData[i] = new double[mCols];
		for (int j = 0; j < mCols; j++) {
			mData[i][j] = 0.0;
		}
	}
	
	// Deletes off the buffer entries
	if (mBufferData != 0) {
		for (int i = 0; i < mRows; i++) {
			delete[] mBufferData[i];
		}
		delete[] mBufferData;
		mBufferData = 0;
	}

}

/** Index operator
* If row and col is out of bounds, return NaN [sqrt(1)]
**/
double& Matrix::operator() (int row, int col) {

	if (row < 0 || row >= mRows || col < 0 || col >= mCols) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] Matrix::operator() - Array index out of bounds (" << row << ", " << col << ")" << endl;
		#endif
		mBuffer = sqrt(-1);
		return mBuffer;
	}
	
	return mData[row][col];
}

/** Index operator but one that ensures
* the matrix is not changed in any way
**/
double Matrix::at(int row, int col) const {

	if (row < 0 || row >= mRows || col < 0 || col >= mCols) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] Matrix::at() - Array index out of bounds (" << row << ", " << col << ")" << endl;
		#endif
		return sqrt(-1);
	}
	
	return mData[row][col];
}

/** Returns a string representation of the matrix
**/
string Matrix::toString(int precision) const {
	ostringstream str;
	
	str << "---Matrix---" << endl;
	str << "mRows: " << mRows << endl;
	str << "mCols: " << mCols << endl;
	
	if (precision >= 0) {
		str.precision(precision);
		str << showpoint;
		str << fixed;
	}
		
	for (int i = 0; i < mRows; i++) {
		str << mData[i][0];
		for (int j = 1; j < mCols; j++) {
			str << ", " << mData[i][j];
		}
		str << endl;
	}
	
	return str.str();
}

/** Helper function for copying contents
* It copies everything but assumes that allocated memory (if any)
* has already been deleted. The buffer data is not copied over
**/
void Matrix::copyFrom(const Matrix& src) {
	mRows = src.mRows;
	mCols = src.mCols;
	mData = new double*[mRows];
	for (int i = 0; i < mRows; i++) {
		mData[i] = new double[mCols];
		for (int j = 0; j < mCols; j++) {
			mData[i][j] = src.mData[i][j];
		}
	}
	mBufferData = 0;
}

/** Checks if the matrix is a square matrix
**/
bool isSquare(const Matrix& matrix) {
	if (matrix.rows() == matrix.cols()) {
		return true;
	} else {
		return false;
	}
}

/** Checks if the matrix has an inverse
* It does this by obtaining the determinant of the matrix. If the determinant is
* less than MATRIX_TINY, then it is considered singular
**/
bool isSingular(const Matrix& matrix) {
	// Todo
	return false;
}

/** Checks if two matrices are compatible
* The column of the first matrix is the same as the row 
* of the second matrix
**/
bool isCompatible(const Matrix& matrix1, const Matrix& matrix2) {
	if (matrix1.cols() == matrix2.rows()) {
		return true;
	} else {
		return false;
	}
}

/** Checks if two matrices are of the same dimension
**/ 
bool isDimensionSame(const Matrix& matrix1, const Matrix& matrix2) {
	if (matrix1.rows() == matrix2.rows() &&
	    matrix1.cols() == matrix2.cols()) {
		return true;
	} else {
		return false;
	}
}

/** Does an addition operation on all the elements in the matrices and returns a pointer to a new matrix. 
* The matrices must be of the same dimension
**/
Matrix* operator+(const Matrix& matrix1, const Matrix& matrix2) {
	
	if (!isDimensionSame(matrix1, matrix2)) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] matrix::operator+() - Matrices are of different dimensions" << endl;
		#endif
		return 0;
	}
	
	// Creates a new matrix
	Matrix* matrix = new Matrix(matrix1.rows(), matrix1.cols());
	for (int i = 0; i < matrix1.rows(); i++) {
		for (int j = 0; j < matrix1.cols(); j++) {
			(*matrix)(i, j) = matrix1.at(i, j) + matrix2.at(i, j);
		}
	}
	
	return matrix;
}

/** Does an subtraction operation on all the elements in the matrices and returns a pointer to a new matrix. 
* The matrices must be of the same dimension
**/
Matrix* operator-(const Matrix& matrix1, const Matrix& matrix2) {

	if (!isDimensionSame(matrix1, matrix2)) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] matrix::operator-() - Matrices are of different dimensions" << endl;
		#endif
		return 0;
	}
	
	// Creates a new matrix
	Matrix* matrix = new Matrix(matrix1.rows(), matrix1.cols());
	for (int i = 0; i < matrix1.rows(); i++) {
		for (int j = 0; j < matrix1.cols(); j++) {
			(*matrix)(i, j) = matrix1.at(i, j) - matrix2.at(i, j);
		}
	}
	
	return matrix;
}

/** Does a matrix multiplication and returns a pointer to the new matrix
* This is the most rudimentary algorithm. Enhancements will be added later
**/ 
Matrix* operator*(const Matrix& matrix1, const Matrix& matrix2) {

	if (!isCompatible(matrix1, matrix2)) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] matrix::operator*() - Matrices are not compatible" << endl;
		#endif
		return 0;
	}
	
	// Creates a new matrix
	Matrix* matrix = new Matrix(matrix1.rows(), matrix2.cols());
	double value;
	for (int i = 0; i < matrix->rows(); i++) {
		for (int j = 0; j < matrix->cols(); j++) {
			value = matrix1.at(i, 0) * matrix2.at(0, j);
			for (int index = 1; index < matrix1.cols(); index++) {
				value = value + (matrix1.at(i, index) * matrix2.at(index, j));
			}
			(*matrix)(i, j) = value;
		}
	}
	
	return matrix;
}

/** Checks if two matrices are the same, i.e. same dimension and every element is the same.
**/
bool operator==(const Matrix& matrix1, const Matrix& matrix2) {
	
	// Checks if their dimensions are the same first
	if (!isDimensionSame(matrix1, matrix2)) {
		return false;
	}
	
	// Checks element by element, returning a false upon the first inequality
	for (int i = 0; i < matrix1.rows(); i++) {
		for (int j = 0; j < matrix1.cols(); j++) {
			if (matrix1.at(i, j) != matrix2.at(i, j)) {
				return false;
			}
		}
	}
	
	return true;
}

/** Transposes a matrix and returns a new matrix
**/
Matrix* transpose(const Matrix& matrix) {

	// Creates a new matrix
	Matrix* transposeMatrix = new Matrix(matrix.cols(), matrix.rows());
	
	// Copies the items
	for (int i = 0; i < matrix.rows(); i++) {
		for (int j = 0; j < matrix.cols(); j++) {
			(*transposeMatrix)(j, i) = matrix.at(i, j);
		}
	}
	
	return transposeMatrix;
}

/** Performs an inverse operation on the given matrix
* Note that the matrix must be a square matrix
**/
bool inverse(Matrix& matrix) {

	// First checks if the matrix is a square matrix.
	if (!isSquare(matrix)) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] matrix::inverse() - Input matrix is not a square matrix" << endl;
		#endif
		return false;
	}

	// Declares the auxiliary variables
	vector<int> indexArray;
	vector<double> col;
	double sign;

	luDecompose(matrix, indexArray, sign);

	// Transpose the indexArray to get the correct pivoting
	vector<int> p;
	p.resize(indexArray.size());
	for (int i = 0; i < indexArray.size(); i++) {
		p[indexArray[i]] = i;
	}
	
	// Occupies col with 0s
	for (int i = 0; i < matrix.rows(); i++) {
		col.push_back(0.0);
	}
	
	// Creates a temporary matrix which is a copy of the input
	Matrix tmpMatrix(matrix);

	for (int j = 0; j < tmpMatrix.cols(); j++) {
		for (int i = 0; i < tmpMatrix.rows(); i++) {
			col[i] = 0.0;
		}
		col[j] = 1.0;
		solve(tmpMatrix, p, col);
		for (int i = 0; i < tmpMatrix.rows(); i++) {
			matrix(i, j) = col[i];
		}
	}
	
	return true;
}

/** Performs a back substitution
* and overwrites the vector b
**/
bool solve(const Matrix& matrix, const vector<int>& indexArray, vector<double>& b) {

	// First checks if the matrix is a square matrix.
	if (!isSquare(matrix)) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] matrix::backSub() - Input matrix is not a square matrix" << endl;
		#endif
		return false;
	}

	// Checks that the indexArray is compatible
	if (indexArray.size() != matrix.rows() || b.size() != matrix.rows()) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] matrix::solve() - indexArray and/or b is not compatible with the matrix" << endl;
		#endif
		return false;
	}

	// Checks that there are no 0 elements along the diagonal
	for (int i = 0; i < matrix.rows(); i++) {
		if (matrix.at(i, i) == 0) {
			#ifdef MATRIX_DEBUG 
				cout << "[Error] matrix::solve() - LU matrix has a 0 diagonal entry" << endl;
			#endif
			return false;
		}
	}
	
	// Transpose the index array and use it to fill up c
	// later on
	vector<double> c;
	for (int i = 0; i < b.size(); i++) {
		c.push_back(b[i]);
	}
	
	for (int i = 0; i < b.size(); i++) {
		b[indexArray[i]] = c[i];
	}
	
	double sum;
	
	// Solve for c using forward substitution, i.e. L.c = P.b
	for (int i = 0; i < b.size(); i++) {
		sum = b[i];
		for (int j = 0; j < i; j++) {
			sum -= matrix.at(i, j) * c[j];
		}
		c[i] = sum;
	}
		
	// Perform back substitution U.x = c
	for (int i = b.size() - 1; i >= 0; i--) {
		sum = c[i];
		for (int j = b.size() - 1; j > i; j--) {
			sum -= matrix.at(i, j) * b[j];
		}
		sum = sum / matrix.at(i, i);
		b[i] = sum;
	}
	
	return true;
}

/** Performs an LU Decomposition on the given matrix
* and overwrites the matrix. The returned matrix is a combined matrix of the upper (U) and 
* lower (L) triangular matrices. It is based on the Crout's algorithm of LU Decomposition
* with partial pivoting
* This algorithm is obtained from the GSL (GNU Scientific Library) with one modification from
* Numerical Recipes, i.e. if the matrix is singular, it will replace the diagonal element by a very
* small number
**/
bool luDecompose(Matrix& matrix, vector<int>& indexArray, double& sign) {

	// First checks if the matrix is a square matrix.
	if (!isSquare(matrix)) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] matrix::luDecompose() - Input matrix is not a square matrix" << endl;
		#endif
		return false;
	}
	
	// Checks if the index array is of the correct size. If not, resize it
	if (indexArray.size() != matrix.rows()) {
		indexArray.clear();
		for (int i = 0; i < matrix.rows(); i++) indexArray.push_back(i);
	} else {
		for (int i = 0; i < matrix.rows(); i++) indexArray[i] = i;
	}

	sign = 1.0;
	int pivot;
	int size = matrix.rows();
	int temp;
	double max;
	double data;
	double dataDiag;
	
	for (int j = 0; j < size - 1; j++) {

		// Find the maximum in the jth column
		pivot = j;
		max = fabs(matrix.at(j, j));
		for (int i = j + 1; i < size; i++) {
			data = fabs(matrix.at(i, j));
			if (data > max) {
				max = data;
				pivot = i;
			}
		}
		// Test if a swap is needed
		if (pivot != j) {
			matrix.swapRow(pivot, j);
			temp = indexArray[pivot];
			indexArray[pivot] = indexArray[j];
			indexArray[j] = temp;
			sign = -1.0 * sign;
		}
		
		// If the diagonal entry is 0.0, then the matrix is non-invertible, i.e. singular matrix
		// To prevent this from becoming an issue, we instead let it be a very small number
		if (matrix.at(j, j) != 0.0) {
			dataDiag = matrix.at(j, j);
		} else {
			dataDiag = MATRIX_TINY;
		}
		
		for (int i = j + 1; i < size; i++) {
			matrix(i, j) = matrix.at(i, j) / matrix.at(j, j);
			for (int k = j + 1; k < size; k++) {
				matrix(i, k) = matrix.at(i, k) - matrix.at(i, j) * matrix.at(j, k);
			}
		}
		
	}
	
	return true;
}

/** Obtains the determinant of a matrix
* If the matrix is invalid, i.e. not a square matrix, it will return NaN
**/
double determinant(const Matrix& matrix) {

	// First checks if the matrix is a square matrix.
	if (!isSquare(matrix)) {
		#ifdef MATRIX_DEBUG 
			cout << "[Error] matrix::determinant() - Input matrix is not a square matrix" << endl;
		#endif
		return sqrt(-1);
	}
	
	Matrix LUMatrix(matrix);
	vector<int> indexArray;
	double sign;
	double det = 1.0;
	
	// Performs a LU Decomposition of the matrix
	luDecompose(LUMatrix, indexArray, sign);
	
	// Multiply the diagonals of the U matrix
	for (int i = 0; i < LUMatrix.size(); i++) {
		det *= LUMatrix.at(i, i);
	}
	
	return det;
}

}
