/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: matrix.h
**  
** Description:
** The header file storing the interface of a double valued matrix object, as well as the two default subclasses - 
** Sparse and Dense matrix. The base class is an abstract class and it should not be instantiated.
** Rather, it serves as a interface for the two underlying matrix i.e. dense and sparse matrix. 
** This is such that users do not have to be too overly
** concerned about the underlying implementation, and that they can use both easily using by learning
** one set of functions
**  
** Last modified: 
** 02 March 2007
** 
** Comments:
** 1) The default matrix is a dense matrix. Sparse matrix will be implemented later
****************************************************************************/

#ifndef MATRIX_H_
#define MATRIX_H_

/* For debugging. Comment out to suppress debugging lines */
// #define MATRIX_DEBUG

/* Definition of constants to be used */
#define MATRIX_TINY 1.0e-20

/* Standard Includes */
#include <iostream>
#include <vector>

using namespace std;

namespace matrix {

class Matrix {

public:
	
	/* Constructor and Destructor */
	Matrix(int rows = 1, int cols = 1);
	Matrix(const Matrix& src);	
	Matrix& operator=(const Matrix& src);
	virtual ~Matrix();

	/* Operations */
	void clear();
	void zeros();
	void ones();
	void setValue(double value);
	void setIdentity();
	void fillDiagonal(double value);
	bool resize(int rows, int cols);
	bool swapRow(int r1, int r2);
	bool swapCol(int c1, int c2);

	/* Basic Operations */
	bool add(const Matrix& matrix1, const Matrix& matrix2);
	bool subtract(const Matrix& matrix1, const Matrix& matrix2);
	bool multiply(const Matrix& matrix1, const Matrix& matrix2);
	bool multiplyVector(vector<double>& result, const vector<double>& colVector) const;

	/* Advance operations */
	bool inverse();
		
	/* Accessor functions */
	int rows() const { return mRows; } 
	int cols() const { return mCols; } 
	int size() const { return mRows * mCols; }
	double& operator() (int row, int col);
	double at(int row, int col) const;
	
	/* Printing and debugging */
	string toString(int precision = -1) const;
	
protected:
	
	/* Attributes */
	int mRows;
	int mCols;
	double ** mData;
	double mBuffer;		// This mBuffer is only used by the operator() in case an error occurs. 
	
	void copyFrom(const Matrix& src);

private:
	
	/* Miscellaneous attributes to aid computation */
	double** mBufferData;

};

/***************************************
** Functions to be performed on the matrices
***************************************/
/* Enquiry functions */
bool isSquare(const Matrix& matrix);
bool isSingular(const Matrix& matrix);
bool isCompatible(const Matrix& matrix1, const Matrix& matrix2);
bool isDimensionSame(const Matrix& matrix1, const Matrix& matrix2);

/* Operations */
Matrix* operator+(const Matrix& matrix1, const Matrix& matrix2);
Matrix* operator-(const Matrix& matrix1, const Matrix& matrix2);
Matrix* operator*(const Matrix& matrix1, const Matrix& matrix2);
Matrix* operator/(const Matrix& matrix1, const Matrix& matrix2);
bool operator==(const Matrix& matrix1, const Matrix& matrix2);
Matrix* transpose(const Matrix& matrix);
bool inverse(Matrix& matrix);
bool solve(const Matrix& matrix, const vector<int>& indexArray, vector<double>& b);
bool luDecompose(Matrix& matrix, vector<int>& indexArray, double& sign);
double determinant(const Matrix& matrix);

}

#endif
