/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: ode_test.cpp
**  
** Description:
** This is a set of test functions to test the functionality of the ODE system
**
** Last modified: 
** 26 February 2007
** 
** Comments:
**
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <vector>

/* Application Includes */
#include "odecommon.h"
#include "odeparser.h"
#include "odetokenizer.h"
#include "odesystem.h"
#include "odenode.h"
#include "odefunction.h"
#include "odetrace.h"
#include "odesolver.h"
#include "ode_test.h"

using namespace std;
using namespace ode;

void testODEMain() {
	
	// Calls the testing methods
	// testODE();
	// testODEParser();
	// testODETokenizer();
	// testODENode();
	// testODESystem();
	testODESolver();
}

void testODE() {

	cout << "============" << endl;
	cout << "In testODE()" << endl;
	cout << "============" << endl;

	cout << "Test Set 1: Test isValidSymbol()" << endl;
	cout << "Checking 'v1'    : " << (isValidSymbol("v1") ? "true" : "false") << endl;
	cout << "Checking '_11'   : " << (isValidSymbol("_11") ? "true" : "false") << endl;
	cout << "Checking '+'     : " << (isValidSymbol("+") ? "true" : "false") << endl;
	cout << "Checking '12.3'  : " << (isValidSymbol("12.3") ? "true" : "false") << endl;
	cout << "Checking 'cos'   : " << (isValidSymbol("cos") ? "true" : "false") << endl;
	cout << "Checking 'isVal' : " << (isValidSymbol("isValidVar") ? "true" : "false") << endl;
	cout << "Checking 'e'     : " << (isValidSymbol("e") ? "true" : "false") << endl;
	cout << "Checking 'e+4'   : " << (isValidSymbol("e+4") ? "true" : "false") << endl;
	cout << endl;
	
	cout << "Test Set 2: Test isNumeric()" << endl;
	cout << "Checking '1234'   : " << (isNumeric("1234") ? "true" : "false") << endl;
	cout << "Checking '12.3'   : " << (isNumeric("12.3") ? "true" : "false") << endl;
	cout << "Checking '-5.0'   : " << (isNumeric("-5.0") ? "true" : "false") << endl;
	cout << "Checking '+5.0'   : " << (isNumeric("+5.0") ? "true" : "false") << endl;
	cout << "Checking '5.6e4'  : " << (isNumeric("5.6e4") ? "true" : "false") << endl;
	cout << "Checking '5.6e+4' : " << (isNumeric("5.6e+4") ? "true" : "false") << endl;
	cout << "Checking '5.6e-4' : " << (isNumeric("5.6e-4") ? "true" : "false") << endl;
	cout << "Checking '5.6e+-4': " << (isNumeric("5.6e+-4") ? "true" : "false") << endl;
	cout << "Checking 'e5'     : " << (isNumeric("e5") ? "true" : "false") << endl;
	cout << "Checking 'test'   : " << (isNumeric("test") ? "true" : "false") << endl;
	cout << "Checking '+'      : " << (isNumeric("+") ? "true" : "false") << endl;
	cout << endl;
		
	cout << "Test Set 3: Test toDouble()" << endl;
	cout << "Converting '1234'   : " << toDouble("1234") << endl;
	cout << "Converting '12.3'   : " << toDouble("12.3") << endl;
	cout << "Converting '-5.0'   : " << toDouble("-5.0") << endl;
	cout << "Converting '+5.0'   : " << toDouble("+5.0") << endl;
	cout << "Converting '5.6e4'  : " << toDouble("5.6e4") << endl;
	cout << "Converting '5.6e+4' : " << toDouble("5.6e+4") << endl;
	cout << "Converting '5.6e-4' : " << toDouble("5.6e-4") << endl;
	cout << "Converting '5.6e+-4': " << toDouble("5.6e+-4") << endl;
	cout << "Converting 'e5'     : " << toDouble("e5") << endl;
	cout << "Converting 'test'   : " << toDouble("test") << endl;
	cout << "Converting '+'      : " << toDouble("+") << endl;
	cout << endl;

	cout << "Test Set 4: Test isOperator()" << endl;
	cout << "Checking '+'  : " << (isOperator("+") ? "true" : "false") << endl;
	cout << "Checking '-'  : " << (isOperator("-") ? "true" : "false") << endl;
	cout << "Checking '*'  : " << (isOperator("*") ? "true" : "false") << endl;
	cout << "Checking '/'  : " << (isOperator("/") ? "true" : "false") << endl;
	cout << "Checking '^'  : " << (isOperator("^") ? "true" : "false") << endl;
	cout << "Checking '5'  : " << (isOperator("5") ? "true" : "false") << endl;
	cout << endl;
	
	cout << "Test Set 5: Test isFunction()" << endl;
	cout << "Checking 'sin'  : " << (isFunction("sin") ? "true" : "false") << endl;
	cout << "Checking '67'   : " << (isFunction("67") ? "true" : "false") << endl;
	cout << "Checking '*'    : " << (isFunction("*") ? "true" : "false") << endl;
	cout << "Checking 'tanh' : " << (isFunction("tanh") ? "true" : "false") << endl;
	cout << "Checking 'exp'  : " << (isFunction("exp") ? "true" : "false") << endl;
	cout << "Checking 'test' : " << (isFunction("test") ? "true" : "false") << endl;
	cout << endl;
	
	ODEFunction* fun;
	cout << "Test Set 6: Test createFunction()" << endl;
	cout << "Creating '+':" << endl;
	fun = createFunction("+");
	cout << "2 + 3 = " << fun->f(2, 3) << endl;
	if (fun != 0) delete fun;
	cout << "Creating '-':" << endl;
	fun = createFunction("-");
	cout << "2 - 3 = " << fun->f(2, 3) << endl;
	if (fun != 0) delete fun;
	cout << "Creating '*':" << endl;
	fun = createFunction("*");
	cout << "2 * 3 = " << fun->f(2, 3) << endl;
	if (fun != 0) delete fun;
	cout << "Creating '/':" << endl;
	fun = createFunction("/");
	cout << "2 / 3 = " << fun->f(2, 3) << endl;
	if (fun != 0) delete fun;
	cout << "Creating '^':" << endl;
	fun = createFunction("^");
	cout << "2 ^ 3 = " << fun->f(2, 3) << endl;
	if (fun != 0) delete fun;
	fun = createFunction("test");
	cout << "test = " << fun << endl;
	if (fun != 0) delete fun;
	fun = createFunction("ln");
	cout << "ln(5)  = " << fun->f(5) << endl;
	if (fun != 0) delete fun;

}

void testODEParser() {

	cout << "==================" << endl;
	cout << "In testODEParser()" << endl;
	cout << "==================" << endl;

	ODEParser parser;
	ODENode* node;
	
	cout << "Test 1: Parsing equation : 1 + 2" << endl;
	node = parser.parse("1 + 2");
	if (node != 0) {
		cout << node->toString() << endl;
		delete node;
	} else {
		cout << "Node not created" << endl;
	}
	cout << endl;

	cout << "Test 1: Parsing equation : (1 + 2)" << endl;
	node = parser.parse("(1 + 2)");
	if (node != 0) {
		cout << node->toString() << endl;
		delete node;
	} else {
		cout << "Node not created" << endl;
	}
	cout << endl;
	
	cout << "Test 2: Parsing equation : sin(3) + 5" << endl;
	node = parser.parse("sin(3) + 5");
	if (node != 0) {
		cout << node->toString() << endl;
		delete node;
	} else {
		cout << "Node not created" << endl;
	}
	cout << endl;

	cout << "Test 3: Parsing equation : sin(3 + 5)" << endl;
	node = parser.parse("sin(3 + 5)");
	if (node != 0) {
		cout << node->toString() << endl;
		delete node;
	} else {
		cout << "Node not created" << endl;
	}
	cout << endl;
	
	cout << "Test 4: Parsing equation : sin(cos(5))" << endl;
	node = parser.parse("sin(cos(5))");
	if (node != 0) {
		cout << node->toString() << endl;
		delete node;
	} else {
		cout << "Node not created" << endl;
	}
	cout << endl;

	cout << "Test 5: Parsing equation : k1 * x1 * x2 - k2 * x2" << endl;
	node = parser.parse("k1 * x1 * x2 - k2 * x2");
	if (node != 0) {
		cout << node->toString() << endl;
		delete node;
	} else {
		cout << "Node not created" << endl;
	}
	cout << endl;

	cout << "Test 6: Parsing equation : k1 * x1 * x2 / (KM + x2)" << endl;
	node = parser.parse("k1 * x1 * x2 / (KM + x2)");
	if (node != 0) {
		cout << node->toString() << endl;
		delete node;
	} else {
		cout << "Node not created" << endl;
	}
	cout << endl;
	
}

void testODETokenizer() {

	cout << "==================" << endl;
	cout << "In testODETokenizer()" << endl;
	cout << "==================" << endl;

	ODETokenizer tokenizer;
	
	cout << "Test 1: Tokenize 3 * 4 * sin(5 + 6) / 7 + v1" << endl;
	tokenizer.tokenize("3 * 4 * sin(5 + 6) / 7 + v1");
	if (tokenizer.hasMoreTokens()) cout << tokenizer.nextToken();
	while (tokenizer.hasMoreTokens()) cout << ", " << tokenizer.nextToken();
	cout << endl;
	
}

void testODENode() {

	cout << "==================" << endl;
	cout << "In testODENode()" << endl;
	cout << "==================" << endl;

	ODEParser parser;
	ODENode* node;
	
	cout << "Test 1: Parsing and evaluating sin(3) + 5" << endl;
	node = parser.parse("sin(3) + 5");
	if (node != 0) {
		cout << node->toString() << endl;
		cout << "Ans : " << node->evaluate() << endl;
		delete node;
	} else {
		cout << "Node not created" << endl;
	}
	cout << endl;
	
	cout << "Test 2: Parsing and evaluating 5*3*5/(4+3)" << endl;
	node = parser.parse("5*3*5/(4+3)");
	if (node != 0) {
		cout << node->toString() << endl;
		cout << "Ans : " << node->evaluate() << endl;
		delete node;
	} else {
		cout << "Node not created" << endl;
	}
	cout << endl;
	
}

void testODEFunction() {
}

void testODESystem() {

	cout << "==================" << endl;
	cout << "In testODENode()" << endl;
	cout << "==================" << endl << endl;

	cout << "=======================================" << endl;
	cout << "This is a more extensive test of the core functionalities" << endl;
	cout << "of a system of ordinary differential equations" << endl;
	
	cout << "Test set 1: Creating a system of equations of the following form " << endl;
	cout << "y1 = 2" << endl;
	cout << "y2 = 1.5 * y1 - k1 * y2" << endl;
	cout << "y3 = 0.75 * y2 - k2 * y3" << endl;
	cout << "y4 = u1 * y4 + 5.0" << endl;
	cout << "=======================================" << endl;
	
	ODESystem system("System");
	system.addDVariable("y1");
	system.addDVariable("y2");
	system.addDVariable("y3");
	system.addDVariable("y4");
	system.addIVariable("u1");
	system.addParameter("k1");
	system.addParameter("k2");
	system.addNode("y1", "2");
	system.addNode("y2", "1.5 * y1 - k1 * y2");
	system.addNode("y3", "0.75 * y2 - k2 * y3");
	system.addNode("y4", "u1 * y4 + 5.0");

	cout << "Testing basic functions" << endl;
	system.setParameter("k1", 0.5);
	system.setParameter("k2", 1.5);
	ODEFunction* fn_sin = new ODEFunction_sin();
	system.addFunction("u1", fn_sin);
	
	cout << system.toString() << endl;
	
	cout << "Getting the results" << endl;
	vector<double> result;
	vector<double> y;
	vector<double> u;
	
	result.push_back(0.0); result.push_back(0.0); result.push_back(0.0); 
	y.push_back(1.0); y.push_back(2.0); y.push_back(3.0); y.push_back(1.5);
	system.getIVariableAt(u, 0.5);
	
	system.f(result, y, u);
	cout << result[0];
	for (int i = 0; i < result.size(); i++) {
		cout << ", " << result[i];
	}
	cout << endl;
}

void testODESolver() {

	cout << "==================" << endl;
	cout << "In testODESolver()" << endl;
	cout << "==================" << endl << endl;

	ODESolverExpEuler solver;
	// ODESolverImpEuler solver;
	vector<double> p;
	vector<double> y;
	ODETrace* trace;

	cout << "=======================================" << endl;
	cout << "This is the test of a complete simple system" << endl;
	cout << "using the explicit euler simulation " << endl;
	cout << "scheme" << endl;
	cout << "=======================================" << endl;
	
	cout << "Test set 1: Creating a system of equations of the following form " << endl;
	cout << "y1 = 2 - 1.5 * y1" << endl;
	cout << "y2 = 1.5 * y1 - k1 * y2" << endl;
	cout << "y3 = k1 * y2 - k2 * y3" << endl;
	cout << "=======================================" << endl;
	
	// Creates the system
	ODESystem system1("System 1");
	system1.addDVariable("y1");
	system1.addDVariable("y2");
	system1.addDVariable("y3");
	system1.addParameter("k1");
	system1.addParameter("k2");
	system1.addNode("y1", "2 - 1.5 * y1");
	system1.addNode("y2", "1.5 * y1 - k1 * y2");
	system1.addNode("y3", "k1 * y2 - k2 * y3");

	// Sets the parameters
	p.push_back(0.5); p.push_back(2.0);
	y.push_back(2.0); y.push_back(0.0); y.push_back(0.0);
	
	// Loads the system, parameters and simulates
	system1.setParameter(p);
	solver.loadODESystem(&system1);
	trace = solver.solve(y);
	
	// Saves the trace and delete it
	if (trace != 0) {
		trace->save("testODESolver_1_log.txt");
		delete trace;
	} else {
		cout << "Test fail: Unable to simulate" << endl;
	}

	// Clears the stuff
	solver.unloadODESystem();
	p.clear();
	y.clear();
	
	cout << "Test set 2: Stiff ODE " << endl;
	cout << "y1 = -15 * y1" << endl;
	cout << "=======================================" << endl;

	// Creates the system
	ODESystem system2("System 2");
	system2.addDVariable("y1");
	system2.addNode("y1", "- 15 * y1");
	
	// Sets the initial values
	y.push_back(1.0);
	
	// Loads the system, parameters and simulates
	system2.setParameter(p);
	solver.loadODESystem(&system2);
	solver.setAttribute("step", 0.20);
	solver.setAttribute("logStep", 0.20);
	trace = solver.solve(y);
	
	// Saves the trace and delete it
	if (trace != 0) {
		trace->save("testODESolver_2_log.txt");
		delete trace;
	} else {
		cout << "Test fail: Unable to simulate" << endl;
	}

	// Clears the stuff
	solver.unloadODESystem();
	p.clear();
	y.clear();
	
}

