/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: odecommon.cpp
**  
** Description:
** This is the implementation of the common functionalities that will be used in the ode package
**
** Last modified: 
** 06 April 2007
** 
** Comments:
**
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <sstream>
#include <vector>
#include <math.h>

/* Application Includes */
#include "odecommon.h"
#include "odefunction.h"

using namespace std;

namespace ode {

/** Checks if a given token is a valid symbol
* This is the same as the valid symbol of the ode.h
**/
bool isValidSymbol(string token) {

	// Checks for empty string
	if (token == "") return false;
	
	char c;
	
	// Checks the first charactor
	c = token[0];
	if (c != '_' && !isalpha(c)) return false;
	
	// Checks the rest of the characters. The rest can only
	// be alphanumeric or '_'
	int length = strlen(token.c_str());
	for (int i = 1; i < length; i++) {
		c = token[i];
		if (c != '_' && !isalnum(c)) return false;
	}
	
	// Does a final check to determine if it is not a predefined function
	if (isFunction(token)) {
		return false;
	}
	
	return true;

}

/** Checks if a given token can be converted into a number
* The checking mechanism uses states to keep track of 
* the checking state.
* 1 : Handles the normal numbers
* 2 : Handles the decimal portion
* 3 : Handling the exponent of a scientific notation before seeing the sign
* 4 : Handling the exponent of a scientific notation after seeing the sign
* -1 : Error
**/
bool isNumeric(string token) {
	
	int index = 0;
	char c;
	int state = 1;
	bool seenDigit = false;
	
	// Checks if the token is an empty string
	if (token == "") {
		#ifdef ODECOMMON_DEBUG
			cout << "[Error] ode::isNumeric() - Empty token" << endl;
		#endif	
		return false;
	}
	
	// Checks if the first token is a '-' or '+'
	c = token[0];
	if (c == '-' || c == '+') {
		index++;
	}
	
	while (index < token.size()) {
		c = token[index];

		// Checks the state and acts according to the state
		switch (state) {
		case 1:		// Handling normal number
			if (isdigit(c)) {
				seenDigit = true;
				state = 1;
			} else if (c == '.') {
				state = 2;
			} else if (seenDigit == true && (c == 'e' || c == 'E')) {
				state = 3;
			} else {
				state = -1;
			}
			break;
		case 2:		// Handles the decimal portion
			if (isdigit(c)) {
				seenDigit = true;
				state = 2;
			} else if (seenDigit == true && (c == 'e' || c == 'E')) {
				state = 3;
			} else {
				state = -1;
			}
			break;
		case 3:
			if (isdigit(c)) {
				state = 4;
			} else if (c == '+' || c == '-') {
				// Does a peek into the next character to ensure that
				// it is a number
				if (index < token.length() - 1) {
					char c_next = token[index + 1];
					if (isdigit(c_next)) {
						state = 4;
					} else {
						state = -1;
					}
				} else {
					state = -1;
				}				
			}
			break;
		case 4:
			if (isdigit(c)) {
				state = 4;
			} else {
				state = -1;
			}
			break;
		case -1:
			return false;
		default:
			// Do nothing
			break;
		}
		
		index++;
	}
	
	// Checks the final state
	if (state == -1 || seenDigit == false) {
		return false;
	} else {
		return true;
	}
}

/** Converts a string into a double
* Converts a string into a double using the atof function
**/
double toDouble(string token) {
	if (isNumeric(token)) {
		return atof(token.c_str());
	} else {
		return sqrt(-1);
	}
}

/** Checks if a given token is an operator
* The list of operators are: '+', '-', '*', '/', '^', '(', ')'
**/
bool isOperator(string token) {
	
	if (token == "+" ||
		token == "-" ||
		token == "/" ||
		token == "*" ||
		token == "^" ||
		token == "(" ||
		token == ")") {
		return true;
	} else {
		return false;
	}
	
}

/** Checks if a given token is a predefined mathematical function
* The list of mathematical functions are as follows:
* abs, arccos, arccosh, arccot, arccoth, arccsc,
* arccsch, arcsec, arcsech, arcsin, arcsinh
* arctan, arctanh, ceiling, cos, cosh, cot,
* coth, csc, csch, exp, factorial, floor, ln
* log, sec, sech, sin, sinh, tan, tanh
**/
bool isFunction(string token) {

	// Creates a list of keywords
	string functionList[] = {"abs", "arccos", "arccosh", "arccot", "arccoth", "arccsc",
						 	 "arccsch", "arcsec", "arcsech", "arcsin", "arcsinh",
							 "arctan", "arctanh", "ceiling", "cos", "cosh", "cot",
							 "coth", "csc", "csch", "exp", "factorial", "floor", "ln",
							 "log", "sec", "sech", "sin", "sinh", "sqrt", "tan", "tanh"};
	// Performs a loop to check
	for (int i = 0; i < 31; i++) {
		if (token == functionList[i]) {
			return true;
		}
	}
	
	return false;
}

/** Returns the priority of the operator 
* If an invalid operator is encountered, return a -1
**/
int getPriority(string token) {
	if (token == "(" || token == ")") {
		return 1;
	} else if (isFunction(token)) {
		return 2;
	} else if (token == "^") {
		return 3;
	} else if (token == "*" || token == "/") {
		return 4;
	} else if (token == "+" || token == "-") {
		return 5;
	} else {
		return -1;
	}
}

/** Given a token, returns its type
* For variables, it will only return the symbol type as there is
* no context for this token to compare itself against
**/
ODENodeType getType(string token) {
	
	ODENodeType type;

	// An extensive test of the token type to determine what type it is
	if (token == "+") {
		type = nodeType_operator_plus;
	} else if (token == "-") {
		type = nodeType_operator_minus;
	} else if (token == "*") {
		type = nodeType_operator_times;
	} else if (token == "/") {
		type = nodeType_operator_divide;
	} else if (token == "^") {
		type = nodeType_operator_power;
	} else if (token == "abs") {
		type = nodeType_function_abs;
	} else if (token == "arccos") {
		type = nodeType_function_arccos;
	} else if (token == "arccosh") {
		type = nodeType_function_arccosh; 
	} else if (token == "arccot") {
		type = nodeType_function_arccot; 
	} else if (token == "arccoth") {
		type = nodeType_function_arccoth; 
	} else if (token == "arccsc") {
		type = nodeType_function_arccsc;
	} else if (token == "arccsch") {
		type = nodeType_function_arccsch; 
	} else if (token == "arcsec") {
		type = nodeType_function_arcsec; 
	} else if (token == "arcsech") {
		type = nodeType_function_arcsech; 
	} else if (token == "arcsin") {
		type = nodeType_function_arcsin; 
	} else if (token == "arcsinh") {
		type = nodeType_function_arcsinh;
	} else if (token == "arctan") {
		type = nodeType_function_arctan; 
	} else if (token == "arctanh") {
		type = nodeType_function_arctanh; 
	} else if (token == "ceiling") {
		type = nodeType_function_ceiling; 
	} else if (token == "cos") {
		type = nodeType_function_cos; 
	} else if (token == "cosh") {
		type = nodeType_function_cosh; 
	} else if (token == "cot") {
		type = nodeType_function_cot;
	} else if (token == "coth") {
		type = nodeType_function_coth; 
	} else if (token == "csc") {
		type = nodeType_function_csc; 
	} else if (token == "csch") {
		type = nodeType_function_csch; 
	} else if (token == "exp") {
		type = nodeType_function_exp; 
	} else if (token == "factorial") {
		type = nodeType_function_factorial; 
	} else if (token == "floor") {
		type = nodeType_function_floor; 
	} else if (token == "ln") {
		type = nodeType_function_ln;
	} else if (token == "log") {
		type = nodeType_function_log; 
	} else if (token == "sec") {
		type = nodeType_function_sec; 
	} else if (token == "sech") {
		type = nodeType_function_sech; 
	} else if (token == "sin") {
		type = nodeType_function_sin; 
	} else if (token == "sinh") {
		type = nodeType_function_sinh; 
	} else if (token == "sqrt") {
		type = nodeType_function_sqrt; 
	} else if (token == "tan") {
		type = nodeType_function_tan; 
	} else if (token == "tanh") {
		type = nodeType_function_tanh;	
	} else if (isValidSymbol(token)) {
		type = nodeType_symbol;
	} else if (isNumeric(token)) { 
		type = nodeType_value;
	} else {
		type = nodeType_unknown;
	}

	return type;
}

/** Given a token, creates a function
* It goes through an extensive list of keywords to determine
* which function to create. If the keyword is not recognize
* a 0 is returned
**/
ODEFunction* createFunction(string token) {

	// Gets the type of the node and use the function creation method based on node type
	ODENodeType type = getType(token);
	return createFunction(type);
}

/** Given a type, creates the function object
**/
ODEFunction* createFunction(ODENodeType type) {

	// An extensive test of the  type to create the function
	ODEFunction* function;
	
	if (type == nodeType_operator_plus) {
		function = new ODEFunction_plus();
	} else if (type == nodeType_operator_minus) {
		function = new ODEFunction_minus();
	} else if (type == nodeType_operator_times) {
		function = new ODEFunction_times();
	} else if (type == nodeType_operator_divide) {
		function = new ODEFunction_divide();
	} else if (type == nodeType_operator_power) {
		function = new ODEFunction_power();
	} else if (type == nodeType_function_abs) {
		function = new ODEFunction_abs();
	} else if (type == nodeType_function_arccos) {
		function = new ODEFunction_arccos();
	} else if (type == nodeType_function_arccosh) {
		function = new ODEFunction_arccosh();
	} else if (type == nodeType_function_arccot) {
		function = new ODEFunction_arccot();
	} else if (type == nodeType_function_arccoth) {
		function = new ODEFunction_arccoth();
	} else if (type == nodeType_function_arccsc) {
		function = new ODEFunction_arccsc();
	} else if (type == nodeType_function_arccsch) {
		function = new ODEFunction_arccsch();
	} else if (type == nodeType_function_arcsec) {
		function = new ODEFunction_arcsec();
	} else if (type == nodeType_function_arcsech) {
		function = new ODEFunction_arcsech();
	} else if (type == nodeType_function_arcsin) {
		function = new ODEFunction_arcsin();
	} else if (type == nodeType_function_arcsinh) {
		function = new ODEFunction_arcsinh();
	} else if (type == nodeType_function_arctan) {
		function = new ODEFunction_arctan();
	} else if (type == nodeType_function_arctanh) {
		function = new ODEFunction_arctanh();
	} else if (type == nodeType_function_ceiling) {
		function = new ODEFunction_ceiling();
	} else if (type == nodeType_function_cos) {
		function = new ODEFunction_cos();
	} else if (type == nodeType_function_cosh) {
		function = new ODEFunction_cosh();
	} else if (type == nodeType_function_cot) {
		function = new ODEFunction_cot();
	} else if (type == nodeType_function_coth) {
		function = new ODEFunction_coth();
	} else if (type == nodeType_function_csc) {
		function = new ODEFunction_csc();
	} else if (type == nodeType_function_csch) {
		function = new ODEFunction_csch();
	} else if (type == nodeType_function_exp) {
		function = new ODEFunction_exp();
	} else if (type == nodeType_function_factorial) {
		function = new ODEFunction_factorial();
	} else if (type == nodeType_function_floor) {
		function = new ODEFunction_floor();
	} else if (type == nodeType_function_ln) {
		function = new ODEFunction_ln();
	} else if (type == nodeType_function_log) {
		function = new ODEFunction_log();
	} else if (type == nodeType_function_sec) {
		function = new ODEFunction_sec();
	} else if (type == nodeType_function_sech) {
		function = new ODEFunction_sech();
	} else if (type == nodeType_function_sin) {
		function = new ODEFunction_sin();
	} else if (type == nodeType_function_sinh) {
		function = new ODEFunction_sinh();
	} else if (type == nodeType_function_sqrt) {
		function = new ODEFunction_sqrt();
	} else if (type == nodeType_function_tan) {
		function = new ODEFunction_tan();
	} else if (type == nodeType_function_tanh) {
		function = new ODEFunction_tanh();
	} else {
		function = 0;
	}
	
	return function;
}

/** Given a type, returns its string representation
**/
string typeToString(ODENodeType type) {

	// An extensive test of the type to convert into its string representation
	string str;
	
	if (type == nodeType_operator_plus) {
		str = "+";
	} else if (type == nodeType_operator_minus) {
		str = "-";
	} else if (type == nodeType_operator_times) {
		str = "*";
	} else if (type == nodeType_operator_divide) {
		str = "/";
	} else if (type == nodeType_operator_power) {
		str = "^";
	} else if (type == nodeType_function_abs) {
		str = "abs";
	} else if (type == nodeType_function_arccos) {
		str = "arccos";
	} else if (type == nodeType_function_arccosh) {
		str = "arccosh";
	} else if (type == nodeType_function_arccot) {
		str = "arccot";
	} else if (type == nodeType_function_arccoth) {
		str = "arccoth";
	} else if (type == nodeType_function_arccsc) {
		str = "arccsc";
	} else if (type == nodeType_function_arccsch) {
		str = "arccsch";
	} else if (type == nodeType_function_arcsec) {
		str = "arcsec";
	} else if (type == nodeType_function_arcsech) {
		str = "arcsech";
	} else if (type == nodeType_function_arcsin) {
		str = "arcsin";
	} else if (type == nodeType_function_arcsinh) {
		str = "arcsinh";
	} else if (type == nodeType_function_arctan) {
		str = "arctan";
	} else if (type == nodeType_function_arctanh) {
		str = "arctanh";
	} else if (type == nodeType_function_ceiling) {
		str = "ceiling";
	} else if (type == nodeType_function_cos) {
		str = "cos";
	} else if (type == nodeType_function_cosh) {
		str = "cosh";
	} else if (type == nodeType_function_cot) {
		str = "cot";
	} else if (type == nodeType_function_coth) {
		str = "coth";
	} else if (type == nodeType_function_csc) {
		str = "csc";
	} else if (type == nodeType_function_csch) {
		str = "csch";
	} else if (type == nodeType_function_exp) {
		str = "exp";
	} else if (type == nodeType_function_factorial) {
		str = "factorial";
	} else if (type == nodeType_function_floor) {
		str = "floor";
	} else if (type == nodeType_function_ln) {
		str = "ln";
	} else if (type == nodeType_function_log) {
		str = "log";
	} else if (type == nodeType_function_sec) {
		str = "sec";
	} else if (type == nodeType_function_sech) {
		str = "sech";
	} else if (type == nodeType_function_sin) {
		str = "sin";
	} else if (type == nodeType_function_sinh) {
		str = "sinh";
	} else if (type == nodeType_function_sqrt) {
		str = "sqrt";
	} else if (type == nodeType_function_tan) {
		str = "tan";
	} else if (type == nodeType_function_tanh) {
		str = "tanh";
	} else {
		str = "undefined";
	}
	
	return str;
}

/** Does a scalar multiplication of the factor h with the elements of y
* This function will always return true.
**/
bool multiplyVector(vector<double>& result, double h, const vector<double>& y) {
	// Checks the size of the result vector
	if (result.size() != y.size()) {
		result.resize(y.size());
	}
	for (int i = 0; i < y.size(); i++) {
		result[i] = h * y[i];
	}

	return true;
}

/** Does an element wise addition
* If the size of the two vectors do not match, do nothing to the result and return a false
**/
bool addVector(vector<double>& result, const vector<double>& y1, const vector<double>& y2) {
	if (y1.size() != y2.size()) {
		#ifdef ODECOMMON_DEBUG
			cout << "[Error] ode::addVector() - Incompatible vector sizes" << endl;
		#endif
		return false;
	} else {
		// Resize the result if it does not match
		if (result.size() != y1.size()) {
			result.resize(y1.size());
		}
		for (int i = 0; i < y1.size(); i++) {
			result[i] = y1[i] + y2[i];
		}
		return true;
	}
}

/** Does an element wise subtraction
* If the size of the two vectors do not match, do nothing to the result and return a false
**/
bool subtractVector(vector<double>& result, const vector<double>& y1, const vector<double>& y2) {
	if (y1.size() != y2.size()) {
		#ifdef ODECOMMON_DEBUG
			cout << "[Error] ode::subtractVector() - Incompatible vector sizes" << endl;
		#endif
		return false;
	} else {
		// Resize the result if it does not match
		if (result.size() != y1.size()) {
			result.resize(y1.size());
		}
		for (int i = 0; i < y1.size(); i++) {
			result[i] = y1[i] - y2[i];
		}
		return true;
	}
}


/** Computes the magnitude of a vector, i.e. || vector ||
**/
double magnitude(const vector<double>& v) {

	double value = 0.0;
	for (int i = 0; i < v.size(); i++) {
		value += pow(v[i], 2);
	}
	value = sqrt(value);
	return value;
}

}
