/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: odefunction.h
**  
** Description:
** This is the implementation of a function to be used in the ODE system. It can range from
** simple ones like plus, minus, to the more complicated ones like splines etc. They will all
** extend from a base class - ODEFunction
**
** Last modified: 
** 23 February 2007
** 
** Comments:
**
****************************************************************************/

#ifndef ODEFUNCTION_H_
#define ODEFUNCTION_H_

/* For debugging. Comment out to suppress debugging lines */
#define ODEFUNCTION_DEBUG

/* Standard Includes */
#include <iostream>
#include <sstream>
#include <vector>
#include <math.h>

/* Application Includes */

using namespace std;

namespace ode {

class ODEFunction {
public:
	virtual double f(double arg1, double arg2) { return 0.0; }
	virtual double f(double arg1) { return 0.0; }
	virtual double f(const vector<double>& argList) { return 0.0; }
};

class ODEFunction_constant : public ODEFunction {
public:
	ODEFunction_constant(double value = 0.0) { mValue = value; }
	double f(double arg1, double arg2) { return mValue; }
	double f(double arg1) { return mValue; }
protected:
	double mValue;
};

class ODEFunction_plus : public ODEFunction {
public:
	double f(double arg1, double arg2) { return arg1 + arg2; }
	double f(double arg1) { return arg1; }
};

class ODEFunction_minus : public ODEFunction {
public:
	double f(double arg1, double arg2) { return arg1 - arg2; }
	double f(double arg1) { return -1.0 * arg1; }
};

class ODEFunction_times : public ODEFunction {
public:
	double f(double arg1, double arg2) { return arg1 * arg2; }
};

class ODEFunction_divide : public ODEFunction {
public:
	double f(double arg1, double arg2) { return arg1 / arg2; }
};

class ODEFunction_power : public ODEFunction {
public:
	double f(double arg1, double arg2) { return pow(arg1,arg2); }
};

class ODEFunction_abs : public ODEFunction {
public:
	double f(double arg1) { return fabs(arg1); }
};
  
class ODEFunction_arccos : public ODEFunction {
public:
	double f(double arg1) { return acos(arg1); }
};

class ODEFunction_arccosh : public ODEFunction {
public:
	double f(double arg1) { return acosh(arg1); }
};

class ODEFunction_arccot : public ODEFunction {
public:
	double f(double arg1) { return 1.0 / atan(arg1); }
};

class ODEFunction_arccoth : public ODEFunction {
public:
	double f(double arg1) { return 1.0 / atanh(arg1); }
};

class ODEFunction_arccsc : public ODEFunction {
public:
	double f(double arg1) { return 1.0 / asin(arg1); }
};

class ODEFunction_arccsch : public ODEFunction {
public:
	double f(double arg1) { return 1.0 / asinh(arg1); }
};

class ODEFunction_arcsec : public ODEFunction {
public:
	double f(double arg1) { return 1.0 / acos(arg1); }
};

class ODEFunction_arcsech : public ODEFunction {
public:
	double f(double arg1) { return 1.0 / acosh(arg1); }
};

class ODEFunction_arcsin : public ODEFunction {
public:
	double f(double arg1) { return asin(arg1); }
};

class ODEFunction_arcsinh : public ODEFunction {
public:
	double f(double arg1) { return asinh(arg1); }
};

class ODEFunction_arctan : public ODEFunction {
public:
	double f(double arg1) { return atan(arg1); }
};

class ODEFunction_arctanh : public ODEFunction {
public:
	double f(double arg1) { return atanh(arg1); }
};

class ODEFunction_ceiling : public ODEFunction {
public:
	double f(double arg1) { return ceil(arg1); }
};

class ODEFunction_cos : public ODEFunction {
public:
	double f(double arg1) { return cos(arg1); }
};

class ODEFunction_cosh : public ODEFunction {
public:
	double f(double arg1) { return cosh(arg1); }
};

class ODEFunction_cot : public ODEFunction {
public:
	double f(double arg1) { return 1.0 / tan(arg1); }
};

class ODEFunction_coth : public ODEFunction {
public:
	double f(double arg1) { return 1.0 / tanh(arg1); }
};

class ODEFunction_csc : public ODEFunction {
public:
	double f(double arg1) { return 1.0 / sin(arg1); }
};

class ODEFunction_csch : public ODEFunction {
public:
	double f(double arg1) { return 1.0 / sinh(arg1); }
};

class ODEFunction_exp : public ODEFunction {
public:
	double f(double arg1) { return exp(arg1); }
};
	
class ODEFunction_factorial : public ODEFunction {
public:
	double f(double arg1) {
		
		// Returns if less than 0
		if (arg1 < 0.0) { return 0; }
		int i_arg1 = (int)arg1;
		int result = 1;
		for (int i = 0; i < i_arg1; i++) {
			result = result * (i_arg1 - i);
		}	

		return result;
	}
};
	
class ODEFunction_floor : public ODEFunction {
public:
	double f(double arg1) { return floor(arg1); }
};

class ODEFunction_ln : public ODEFunction {
public:
	double f(double arg1) { return log(arg1); }
};
	
class ODEFunction_log : public ODEFunction {
public:
	double f(double arg1) { return log10(arg1); }
};

class ODEFunction_sec : public ODEFunction {
public:
	double f(double arg1) { return 1.0 / cos(arg1); }
};
	
class ODEFunction_sech : public ODEFunction {
public:
	double f(double arg1) { return 1.0 / cosh(arg1); }
};

class ODEFunction_sin : public ODEFunction {
public:
	double f(double arg1) { return sin(arg1); }
};

class ODEFunction_sinh : public ODEFunction {
public:
	double f(double arg1) { return sinh(arg1); }
};

class ODEFunction_sqrt : public ODEFunction {
public:
	double f(double arg1) { return sqrt(arg1); }
};
	
class ODEFunction_tan : public ODEFunction {
public:
	double f(double arg1) { return tan(arg1); }
};

class ODEFunction_tanh : public ODEFunction {
public:
	double f(double arg1) { return tanh(arg1); }
};

}

#endif /* ODEFUNCTION_H_ */
