/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: odenode.cpp
**  
** Description:
** This is the implementation of a node in the ODE system. A node is a single token/element in the
** equation that describes the system.
**
** Last modified: 
** 20 February 2007
** 
** Comments:
**
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <sstream>
#include <vector>
#include <math.h>

/* Application Includes */
#include "odecommon.h"
#include "odesystem.h"
#include "odenode.h"
#include "odeparser.h"

using namespace std;

namespace ode {

/** Default constructor
**/
ODENode::ODENode(ODENodeType type, ODENode* parent, ODESystem* context) :
	mType(type), mParent(parent), mContext(context), mSymbol(""), mIndex(-1), mValue(0.0) {

	// Creates the function based on the type
	mFunction = createFunction(type);

	#ifdef ODENODE_DEBUG
		cout << "[Info] ODENode::ODENode(1) - Creating node type " << type << endl;
	#endif
	
}

/** Constructor
* Tries to create a syntax node based on the token that is given, i.e. it will perform a check to see
* what kind of token it is before deciding what type to assign this node to.
**/
ODENode::ODENode(string token, ODENode* parent, ODESystem* context) :
	mParent(parent), mContext(context), mSymbol(""), mIndex(-1), mValue(0.0) {

	mType = getType(token);
	mFunction = createFunction(token);
	
	#ifdef ODENODE_DEBUG
		cout << "[Info] ODENode::ODENode(2) - Creating node : " << token << " of type " << mType << endl;
	#endif
	
	// Sets additional fields according to the type it is.
	if (mType == nodeType_value) {
		mValue = toDouble(token);
	} else if (mType == nodeType_symbol_parameter) {
		mIndex = context->indexOfParameter(token);
		mSymbol = token;
	} else if (mType == nodeType_symbol_dvariable) {
		mIndex = context->indexOfDVariable(token);
		mSymbol = token;
	} else if (mType == nodeType_symbol_ivariable) {
		mIndex = context->indexOfIVariable(token);
		mSymbol = token;
	} else if (mType == nodeType_symbol) {
		mSymbol = token;
	} else {
		// Do nothing
	}	
	
}

/** Copy constructor
**/
ODENode::ODENode(const ODENode& src) {
	copyFrom(src);
}

/** Assignment constructor
**/
ODENode& ODENode::operator=(const ODENode& src) {

	// Checks for self assignment
	if (this == &src) {
		return (*this);	
	}
	
	// Deletes all the children in the vector
	vector<ODENode*>::iterator iter = mChildList.begin();
	while (iter != mChildList.end()) {
		delete (*iter);
		iter++;
	}
	mChildList.clear();
	
	// Deletes the function object
	if (mFunction != 0) {
		delete mFunction;
	}
	
	// Unsets the parent and context
	mParent = 0;
	mContext = 0;
	
	// Copy contents
	copyFrom(src);
	
	return (*this);
}

/** Destructor
**/
ODENode::~ODENode() {

	#ifdef ODENODE_DEBUG
		cout << "[Info] ODENode::~ODENode() - Destroying SyntaxNode " << toStringSimple() << endl;
	#endif
	
	// Deletes all the children in the vector
	vector<ODENode*>::iterator iter = mChildList.begin();
	while (iter != mChildList.end()) {
		delete (*iter);
		iter++;
	}
	mChildList.clear();
	
	// Deletes the function object
	if (mFunction != 0) {
		delete mFunction;
	}
	
	// Unsets the parent and context
	mParent = 0;
	mContext = 0;
}

/** Gets the type of the symbol
* For a valid symbol, if the context is set, then use the context to find out
* whether or not it is a parameter, an independent variable or a dependent 
* variable. Otherwise just give it a standard symbol type.
**/
ODENodeType ODENode::getType(string token) {

	ODENodeType type;
	
	// Uses the generic getType to obtain the type
	type = ode::getType(token);

	// If the type is a symbol, further process it if a context is available
	if (type == nodeType_symbol) {
		if (mContext == 0) {
			// Do nothing
		} else {
			int indexParameter = mContext->indexOfParameter(token);
			int indexDVariable = mContext->indexOfDVariable(token);
			int indexIVariable = mContext->indexOfIVariable(token);
		
			// If any one of the indices is not -1, set the type to the corresponding type and set the index
			// to that value
			if (indexParameter != -1) {
				type = nodeType_symbol_parameter;
			} else if (indexDVariable != -1) {
				type = nodeType_symbol_dvariable;
			} else if (indexIVariable != -1) {
				type = nodeType_symbol_ivariable;
			} 
		}
	} 
	
	return type;
}

/** Sets the symbol for this node. 
* Only applicable if the node is of the following types, i.e. 
* 	nodeType_symbol
* 	nodeType_symbol_parameter,
*	nodeType_symbol_dvariable,
*	nodeType_symbol_ivariable
* In addition, if the context is non zero, it checks if the symbols are valid
**/
bool ODENode::setSymbol(string symbol) {

	// Checks if the symbol is valid
	if (!isValidSymbol(symbol)) {
		#ifdef ODENODE_DEBUG
			cout << "[Error] ODENode::setSymbol() - Invalid symbol (" << symbol << ")" << endl;
		#endif
		return false;
	}

	// Checks the type of the node
	if (mType != nodeType_symbol &&
		mType != nodeType_symbol_parameter && 
	    mType != nodeType_symbol_dvariable && 
		mType != nodeType_symbol_ivariable) {
		#ifdef ODENODE_DEBUG
			cout << "[Error] ODENode::setSymbol() - Node is of incorrect type (" << mType << ")" << endl;
		#endif
		return false;
	}
	
	// If the context exists, finds the index. Otherwise, set the index to -1
	if (mContext == 0) {
		mSymbol = symbol;
		mIndex = -1;
	} else {
		int index;
		if (mType == nodeType_symbol_parameter) {
			index = mContext->indexOfParameter(symbol);
		} else if (mType == nodeType_symbol_dvariable) {
			index = mContext->indexOfDVariable(symbol);
		} else if (mType == nodeType_symbol_ivariable) {
			index = mContext->indexOfIVariable(symbol);
		} else {
			int indexParameter = mContext->indexOfParameter(symbol);
			int indexDVariable = mContext->indexOfDVariable(symbol);
			int indexIVariable = mContext->indexOfIVariable(symbol);
			
			// If any one of the indices is not -1, set the type to the corresponding type and set the index
			// to that value
			if (indexParameter != -1) {
				index = indexParameter;
				mType = nodeType_symbol_parameter;
			} else if (indexDVariable != -1) {
				index = indexDVariable;
				mType = nodeType_symbol_dvariable;
			} else if (indexIVariable != -1) {
				index = indexIVariable;
				mType = nodeType_symbol_ivariable;
			} else {
				index = -1;
			}
		}
		
		// Checks the index for validity
		if (index == -1) {
			#ifdef ODENODE_DEBUG
				cout << "[Error] ODENode::setSymbol() - Symbol " << symbol << " does not exists in the system" << endl;
			#endif
			return false;
		} else {
			mSymbol = symbol;
			mIndex = index;
		}
	}

	return true;
}

/** Sets the value for this node.
* Only applicable if the node is of type nodeType_value.
**/
bool ODENode::setValue(double value) {

	// Checks the type of the node
	if (mType != nodeType_value) {
		#ifdef ODENODE_DEBUG
			cout << "[Error] ODENode::setValue() - Node is of incorrect type (" << mType << ")" << endl;
		#endif
		return false;
	} else {
		mValue = value;
	}

	return true;
}

/** Adds a syntax node to the childlist
* There are two checks to be done for this method, i.e. checks if the current node is a variable. If it is
* a variable/value etc. then it must be a leaf and no children can be added. Also the child must have
* its context added.
**/
ODENode* ODENode::addChild(ODENode* child) {
	
	if (child == 0) {
		#ifdef ODENODE_DEBUG
			cout << "[Error] ODENode::addChild() - Invalid child" << endl;
		#endif
		return 0;
	}
	
	// Checks if the child has got a parent. If it has, it cannot be added
	if (child->mParent != 0) {
		#ifdef ODENODE_DEBUG
			cout << "[Error] ODENode::addChild() - Child already has a parent" << endl;
		#endif
		return 0;
	}
	
	// Checks if the current node is a parameter/variable/value
	if (mType == nodeType_symbol_parameter ||
	    mType == nodeType_symbol_dvariable || 
		mType == nodeType_symbol_ivariable ||
		mType == nodeType_value) {
		#ifdef ODENODE_DEBUG
			cout << "[Error] ODENode::addChild() - Node must be a leaf. No more child can be added" << endl;
		#endif
		return 0;
	}
	
	// Sets the context if the current context is not 0
	if (mContext != 0 && child->setContext(mContext)) {
		#ifdef ODENODE_DEBUG
			cout << "[Error] ODENode::addChild() - Child cannot set context" << endl;
		#endif
		return 0;
	}
	
	// Adds the child to the node
	child->mParent = this;
	mChildList.push_back(child);
	return child;
}

/** Removes a child node from the childlist
**/
ODENode* ODENode::removeChild(ODENode* child) {

	if (child == 0) {
		#ifdef ODENODE_DEBUG
			cout << "[Error] ODENode::removeChild() - Invalid child" << endl;
		#endif
		return 0;
	}
	
	// Look for the child in the childlist and remove it
	vector<ODENode*>::iterator iter = mChildList.begin();
	while (iter != mChildList.end()) {
		if ((*iter) == child) {
			break;
		}
		iter++;
	}
	
	if (iter != mChildList.end()) {
		mChildList.erase(iter);
		child->mParent = 0;
		return child;
	} else {
		#ifdef ODENODE_DEBUG
			cout << "[Error] ODENode::removeChild() - Child cannot be found" << endl;
		#endif
		return 0;
	}

}

/** Sets the context of this node 
* It first tries to set the context of the children first before
* setting its own context
**/
bool ODENode::setContext(ODESystem* context) {
	
	// Checks the context itself
	if (context == 0) {
		#ifdef ODENODE_DEBUG
			cout << "[Error] ODENode::setContext() - Invalid context" << endl;
		#endif
		return false;
	}
	
	// Checks the context first. One can only set the context of a node if it 
	// does not have a context already
	if (mContext != 0) {
		#ifdef ODENODE_DEBUG
			cout << "[Error] ODENode::setContext() - Node " << toStringSimple() << " already has a context" << endl;
		#endif
		return false;
	}
	
	bool success = true;
	
	// Sets the context of all the children first
	vector<ODENode*>::iterator iter = mChildList.begin();
	while (iter != mChildList.end()) {
		if (!(*iter)->setContext(context)) {
			success = false;
			break;
		}
		iter++;
	}
	
	// If it is successful, checks if the node is a symbol. If it is a symbol, checks if it is
	// compatible with the context
	if (success) {

		// Performs an initial check to see if the node is less than tanh. If it is
		// then its either a function, or operator and no additional checks is needed.
		// Simply sets its context and exits
		if (mType <= nodeType_function_tanh) {
			mContext = context;	
			return true;
		}

		int index;
		
		if (mType == nodeType_symbol_parameter) {
			index = context->indexOfParameter(mSymbol);
		} else if (mType == nodeType_symbol_ivariable) {
			index = context->indexOfIVariable(mSymbol);
		} else if (mType == nodeType_symbol_dvariable) {
			index = context->indexOfDVariable(mSymbol);
		} else if (mType == nodeType_symbol) {

			int indexParameter = context->indexOfParameter(mSymbol);
			int indexDVariable = context->indexOfDVariable(mSymbol);
			int indexIVariable = context->indexOfIVariable(mSymbol);

			// If any one of the indices is not -1, set the type to the corresponding type and set the index
			// to that value
			if (indexParameter != -1) {
				index = indexParameter;
				mType = nodeType_symbol_parameter;
			} else if (indexDVariable != -1) {
				index = indexDVariable;
				mType = nodeType_symbol_dvariable;
			} else if (indexIVariable != -1) {
				index = indexIVariable;
				mType = nodeType_symbol_ivariable;
			} else {
				index = -1;
			}
		} 

		if (index != -1) {
			mContext = context;
			mIndex = index;
		} else {
			success = false;
		}

	}
	
	// If success is false, then for all the children, unset their context
	if (!success) {
		#ifdef ODENODE_DEBUG
			cout << "[Error] ODENode::setContext() - Could not set context (" << toStringSimple() << ")" << endl;
		#endif
		unsetContext();
		return false;
	}
	
	return true;
}

/** Unsets the context
* Note that the user has to remove the node from the context first before unsetting it, else one would have a dangling 
* reference, i.e. context (system) holds onto the syntax node, but the syntax node cannot get the values of any
* variables/parameters.
**/
bool ODENode::unsetContext() {
	
	mContext = 0;
	mIndex = -1;
	
	// If the type is of symbol, set it to a generic symbol type
	if (mType == nodeType_symbol_parameter ||
	    mType == nodeType_symbol_dvariable ||
		mType == nodeType_symbol_ivariable) {
		mType = nodeType_symbol;
	}
	
	vector<ODENode*>::iterator iter = mChildList.begin();
	while (iter != mChildList.end()) {
		(*iter)->unsetContext();
		iter++;
	}
	
	return true;
}

/** Evaluates the node and returns a value
* If an error occurs, then returns NaN
**/	
double ODENode::evaluate() {

	double result;
	double arg1;
	double arg2;
	
	if (mType == nodeType_value) {
		result = mValue;
	} else if (mType >= nodeType_symbol_parameter && mType <= nodeType_symbol_ivariable) {
		
		// Checks if context is valid
		if (mContext == 0) {
			#ifdef ODENODE_DEBUG
				cout << "[Info] ODENode::evaluate(1) - No context" << endl;
			#endif
			
			return sqrt(-1);
		}
	
		if (mType == nodeType_symbol_parameter) {
			result = mContext->getParameter(mIndex);
		} else if (mType == nodeType_symbol_dvariable) {
			result = mContext->getDVariable(mIndex);
		} else if (mType == nodeType_symbol_ivariable) {
			result = mContext->getIVariable(mIndex);
		}
	} else if (mType <= nodeType_function_tanh) {
		// Checks the number of arguments and calls the appropriate functor
		if (mChildList.size() == 1) {
			arg1 = mChildList[0]->evaluate();
			result = mFunction->f(arg1);
		} else if (mChildList.size() > 1) {
			arg1 = mChildList[0]->evaluate();
			arg2 = mChildList[1]->evaluate();
			result = mFunction->f(arg1, arg2);
		} else {
			result = sqrt(-1);
		} 
	} else {
		result = sqrt(-1);
	}
	
	return result;
}

/** Evaluates the node and returns a value
* This method works under the pretext of no context so it gets all its values from the vectors y, u and p that
* is being passed in
* If an error occurs, returns a NaN
**/
double ODENode::evaluate(const vector<double>& y, const vector<double>& u, const vector<double>& p) {

	double result;
	double arg1;
	double arg2;

	if (mType == nodeType_value) {
		result = mValue;
	} else if (mType == nodeType_symbol_parameter) {
		result = (mIndex < p.size()) ? p[mIndex] : sqrt(-1);
	} else if (mType == nodeType_symbol_dvariable) {
		result = (mIndex < y.size()) ? y[mIndex] : sqrt(-1);
	} else if (mType == nodeType_symbol_ivariable) {
		result = (mIndex < u.size()) ? u[mIndex] : sqrt(-1);
	} else if (mType <= nodeType_function_tanh) {
		// Checks the number of arguments and calls the appropriate functor
		if (mChildList.size() == 1) {
			arg1 = mChildList[0]->evaluate(y, u, p);
			result = mFunction->f(arg1);
		} else if (mChildList.size() > 1) {
			arg1 = mChildList[0]->evaluate(y, u, p);
			arg2 = mChildList[1]->evaluate(y, u, p);
			result = mFunction->f(arg1, arg2);
		} else {
			result = sqrt(-1);
		}
	} else {
		result = sqrt(-1);
	}
	
	return result;

}

/** Returns a simple string representation of the node
**/
string ODENode::toStringSimple() const {
	ostringstream str;
	
	switch (mType) {
	case nodeType_symbol_parameter:
		str << mSymbol;
		break;
	case nodeType_symbol_dvariable:
		str << mSymbol;
		break;
	case nodeType_symbol_ivariable:
		str << mSymbol;
		break;
	case nodeType_symbol:
		str << mSymbol;
		break;
	case nodeType_value:
		str << mValue;
		break;
	case nodeType_unknown:
		str << "unknown";
		break;
	default:
		str << typeToString(mType);
	}
	
	return str.str();
}

/** Returns a string representation of the node
* Returns a post fix representation of the tree that the node encodes for
**/
string ODENode::toString() const {

	ostringstream str;
	
	// If there are child nodes, print them out first
	vector<ODENode*>::const_iterator iter = mChildList.begin();
	while (iter != mChildList.end()) {
		str << (*iter)->toString();
		iter++;
	}
	
	switch (mType) {
	case nodeType_symbol_parameter:
		str << mSymbol << " ";
		break;
	case nodeType_symbol_dvariable:
		str << mSymbol << " ";
		break;
	case nodeType_symbol_ivariable:
		str << mSymbol << " ";
		break;
	case nodeType_symbol:
		str << mSymbol << " ";
		break;
	case nodeType_value:
		str << mValue << " ";
		break;
	case nodeType_unknown:
		str << "unknown ";
		break;
	default:
		str << typeToString(mType) << " ";
	}
	
	return str.str();
}

/** Helper function for copying contents
* It copies everything but assumes that allocated memory (if any)
* has already been deleted. Note that the parent is not copied
**/
void ODENode::copyFrom(const ODENode& src) {

	mType = src.mType;
	mFunction = createFunction(mType);
	mSymbol = src.mSymbol;	
	mIndex = src.mIndex;
	mValue = src.mValue;
	mContext = src.mContext;
	
	mParent = 0;
	
	// Copies the child list. Note that the nodes are being cloned into separate copies
	vector<ODENode*>::const_iterator iter = src.mChildList.begin();
	ODENode* node;
	while (iter != src.mChildList.end()) {
		node = new ODENode(**iter);
		this->addChild(node);
		iter++;
	}
	
}

}
