/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: odeparser.cpp
**  
** Description:
** This is the implementation of an equation parser for an ODE.
**
** Last modified: 
** 19 February 2007
** 
** Comments:
**
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <sstream>
#include <vector>
#include <math.h>

/* Application Includes */
#include "odenode.h"
#include "odeparser.h"
#include "odetokenizer.h"

using namespace std;

namespace ode {

/** Default constructor
**/
ODEParser::ODEParser() {

	#ifdef ODE_DEBUG
		cout << "[Info] ODEParser::ODEParser() - Creating parser" << endl;
	#endif

}

/** Default destructor
**/
ODEParser::~ODEParser() {
	// Clears the stacks and deletes all the nodes off
	clear();
}

/** Takes in a string expression for an equation 
* and returns an abstract syntax tree representing the
* equation
**/
ODENode* ODEParser::parse(string expression) {

	// Clears the main stack and the ops stack
	clear();
	
	string token;
	string tmpToken;
	bool checkUnaryMinus = true;
	bool unaryMinus = false;
	bool error = false;
	ODETokenizer tokenizer(expression);
	ODENode* node;
	ODENode* lNode;
	ODENode* rNode;
	double value;

	while (tokenizer.hasMoreTokens()) {

		// Checks for error first
		if (error) {
			break;
		}
		
		token = tokenizer.nextToken();
		if (isNumeric(token)) {
			node = new ODENode(nodeType_value);
			value = toDouble(token);
			if (unaryMinus) {
				value = -1.0 * value;
				mOpStack.pop_back();
				unaryMinus = false;
			} else {
				// Do nothing
			}
			// Sets the value and pushes it into main stack
			node->setValue(value);
			mStack.push_back(node);
			node = 0;
			checkUnaryMinus = false;
		} else if (isValidSymbol(token)) {
			if (unaryMinus) {
				node = new ODENode(nodeType_operator_minus);
				lNode = new ODENode(token);
				lNode->setSymbol(token);
				node->addChild(lNode);
				mOpStack.pop_back();
				unaryMinus = false;
			} else {
				node = new ODENode(token);
				node->setSymbol(token);
			}
			// Pushes the node into the main stack
			mStack.push_back(node);
			node = 0;
			checkUnaryMinus = false;
		} else if (isOperator(token) || isFunction(token)) {
			if (unaryMinus) {
				#ifdef ODE_DEBUG
					cout << "[Error] ODEParser::parse() - State unary minus when the token is an operator/function : " << token << endl;
				#endif
				error = true;
				break;
			}
			if (checkUnaryMinus) {
				if (token == "-") {
					mOpStack.push_back(token);
					checkUnaryMinus = false;
					unaryMinus = true;
				} else {
					mOpStack.push_back(token);
				}
			} else {
				if (token == "(") {
					// Pushes it into the opstack
					mOpStack.push_back(token);
					checkUnaryMinus = true;
				} else if (token == ")") {
					// Makes sure that the opstack is not empty
					if (mOpStack.size() == 0) {
						#ifdef ODE_DEBUG
							cout << "[Error] ODEParser::parse() - No matching closing braces" << endl;
						#endif
						error = true;
						break;
					}

					tmpToken = mOpStack[mOpStack.size() - 1];
					mOpStack.pop_back();
					
					// If the tmpToken is a "(", check if the following token in the opstack is a function. If it is not,
					// then the "(" can be safely ignored. Otherwise, 
					while (tmpToken != "(") {

						// Checks if the tmpToken is an operator or function. If it is an operator
						// then pops 2 values off the node stack and creates a new node. Otherwise
						// simply pop 1 value off the node stack and creates a new node.
						if (isOperator(tmpToken)) {
							// Checks that there are at least 2 values in the main stack
							if (mStack.size() < 2) {
								#ifdef ODE_DEBUG
									cout << "[Error] ODEParser::parse() - Insufficient operands for operator[1] : " << tmpToken << endl;
								#endif
								error = true;
								break;
							}
							rNode = mStack[mStack.size() - 1];
							lNode = mStack[mStack.size() - 2];
							mStack.pop_back();
							mStack.pop_back();
							
							// Does the linkages between the nodes
							node = new ODENode(tmpToken);
							node->addChild(lNode);
							node->addChild(rNode);
							
							// Pushes the value back into the stack
							mStack.push_back(node);
							
							// Resets the 3 variables
							lNode = 0;
							rNode = 0;
							node = 0;
							
						} else if (isFunction(tmpToken)) {

							// Checks that there is at least 1 value in the main stack
							if (mStack.size() < 1) {
								#ifdef ODE_DEBUG
									cout << "[Error] ODEParser::parse() - Insufficient operands for function[1] : " << tmpToken << endl;
								#endif
								error = true;
								break;
							}
							lNode = mStack[mStack.size() - 1];
							mStack.pop_back();
							
							// Does the linkages between the nodes
							node = new ODENode(tmpToken);
							node->addChild(lNode);
							
							// Pushes the value back into the stack
							mStack.push_back(node);
							
							// Resets the 2 variables
							lNode = 0;
							node = 0;
						
						}
						
						// Makes sure that the opstack is not empty
						if (mOpStack.size() == 0) {
							#ifdef ODE_DEBUG
								cout << "[Error] ODEParser::parse() - No matching closing braces" << endl;
							#endif
							error = true;
							break;
						}
						tmpToken = mOpStack[mOpStack.size() - 1];
						mOpStack.pop_back();
					}
					checkUnaryMinus = false;
				} else {
					if (mOpStack.size() != 0) {
						// Check for precedence
						tmpToken = mOpStack[mOpStack.size() - 1];

						if (getPriority(tmpToken) <= getPriority(token) && tmpToken != "(") {
							
							// Pops the operator off the stack
							mOpStack.pop_back();
							
							// Checks if the tmpToken is an operator or function. If it is an operator
							// then pops 2 values off the node stack and creates a new node. Otherwise
							// simply pop 1 value off the node stack and creates a new node.
							if (isOperator(tmpToken)) {
								// Checks that there are at least 2 values in the main stack
								if (mStack.size() < 2) {
									#ifdef ODE_DEBUG
										cout << "[Error] ODEParser::parse() - Insufficient operands for operator[2] : " << tmpToken << endl;
									#endif
									error = true;
									break;
								}
								rNode = mStack[mStack.size() - 1];
								lNode = mStack[mStack.size() - 2];
								mStack.pop_back();
								mStack.pop_back();
								
								// Does the linkages between the nodes
								node = new ODENode(tmpToken);
								node->addChild(lNode);
								node->addChild(rNode);
								
								// Pushes the value back into the stack
								mStack.push_back(node);
								
								// Resets the 3 variables
								lNode = 0;
								rNode = 0;
								node = 0;
								
							} else if (isFunction(tmpToken)) {

								// Checks that there is at least 1 value in the main stack
								if (mStack.size() < 1) {
									#ifdef ODE_DEBUG
										cout << "[Error] ODEParser::parse() - Insufficient operands for function[2] : " << tmpToken << endl;
									#endif
									error = true;
									break;
								}
								lNode = mStack[mStack.size() - 1];
								mStack.pop_back();
								
								// Does the linkages between the nodes
								node = new ODENode(tmpToken);
								node->addChild(lNode);
								
								// Pushes the value back into the stack
								mStack.push_back(node);
								
								// Resets the 2 variables
								lNode = 0;
								node = 0;
						
							}
						}
					}
					mOpStack.push_back(token);
					checkUnaryMinus = true;
				}
				
			}
		} else {
			error = true;
			break;
		}
	}
	
	while (mOpStack.size() > 0) {
		token = mOpStack[mOpStack.size() - 1];
		mOpStack.pop_back();
		
		// Checks the kind of operator this one is
		// If it is a braces, i.e. ( or ) then return error
		if (isOperator(token)) {
			// Checks that there are at least 2 values in the main stack
			if (mStack.size() < 2) {
				#ifdef ODE_DEBUG
					cout << "[Error] ODEParser::parse() - Insufficient operands for operator[3] : " << token << endl;
				#endif
				error = true;
				break;
			}
			rNode = mStack[mStack.size() - 1];
			lNode = mStack[mStack.size() - 2];
			mStack.pop_back();
			mStack.pop_back();
			
			// Does the linkages between the nodes
			node = new ODENode(token);
			node->addChild(lNode);
			node->addChild(rNode);
			
			// Pushes the value back into the stack
			mStack.push_back(node);
			
			// Resets the 3 variables
			lNode = 0;
			rNode = 0;
			node = 0;
			
		} else if (isFunction(token)) {

			// Checks that there is at least 1 value in the main stack
			if (mStack.size() < 1) {
				#ifdef ODE_DEBUG
					cout << "[Error] ODEParser::parse() - Insufficient operands for function[3] : " << token << endl;
				#endif
				error = true;
				break;
			}
			lNode = mStack[mStack.size() - 1];
			mStack.pop_back();
			
			// Does the linkages between the nodes
			node = new ODENode(token);
			node->addChild(lNode);
			
			// Pushes the value back into the stack
			mStack.push_back(node);
			
			// Resets the 2 variables
			lNode = 0;
			node = 0;
		}
	}

	// Checks for error 
	if (error || mStack.size() != 1) {
		#ifdef ODEPARSER_DEBUG
			cout << "[Error] ODEParser::parse() - Error in expression" << endl;
		#endif	
		
		// Clears all the main stack and opstack
		clear();
		
		return 0;
	}
	
	// Gets the remaining element off the stack and returns it
	node = mStack[0];
	mStack.clear();
	
	return node;
}

/** Clears and empties all the contents in the stacks
**/
void ODEParser::clear() {

	// Clears the main stack and the ops stack
	vector<ODENode*>::iterator iter;
	iter = mStack.begin();
	while (iter != mStack.end()) {
		delete (*iter);
		iter++;
	}
	mStack.clear();
	mOpStack.clear();

}

}
