/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: odesolver.cpp
**  
** Description:
** This is the base class for the ode solvers that will be included in this package. It will
** define the main interface that all the solvers should share
**
** Last modified: 
** 12 March 2007
** 
** Comments:
**
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <sstream>
#include <vector>
#include <math.h>

/* Application Includes */
#include "odecommon.h"
#include "odesolver.h"
#include "odesystem.h"
#include "odetrace.h"

using namespace std;

namespace ode {

/** Default constructor
**/
ODESolver::ODESolver() {
	// Do nothing
}

/** Default destructor
**/
ODESolver::~ODESolver() {
	// Do nothing
}

/** Loads the ODE system
**/
void ODESolver::loadODESystem(ODESystem* system) {
	mSystem = system;
}

/** Unloads the ODE system
**/
void ODESolver::unloadODESystem() {
	mSystem = 0;
}

/** Sets the attribute associated with the solver
**/
void ODESolver::setAttribute(string attribName, double value) {
	mAttributeMap[attribName] = value;
}

/** Default constructor
**/
ODESolverExpEuler::ODESolverExpEuler() {
	#ifdef ODESOLVER_DEBUG
		cout << "[Info] ODESolverExpEuler::ODESolverExpEuler() - Creating solver (Explicit Euler)" << endl;
	#endif

	// Presets the attributes
	mAttributeMap["start"] = 0.0;
	mAttributeMap["stop"] = 15.0;
	mAttributeMap["step"] = 0.01;
	mAttributeMap["logstep"] = 0.01;
}

/** Default destructor
**/
ODESolverExpEuler::~ODESolverExpEuler() {
	#ifdef ODESOLVER_DEBUG
		cout << "[Info] ODESolverExpEuler::~ODESolverExpEuler() - Destroying solver (Explicit Euler)" << endl;
	#endif
}

/** Performs a single step of the simulation
**/
void ODESolverExpEuler::doStep(vector<double>& y, const vector<double>& y_init, double step, double currTime) {

	// Checks if the ODE system is in place and that it is consistent
	if (mSystem == 0 || !mSystem->isConsistent()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverExpEuler::doStep() - No ODE System or inconsistent system" << endl;
		#endif
		return;
	}
	
	// Checks that the initial vector size is correct
	if (y_init.size() != mSystem->dVariableSize()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverExpEuler::doStep() - Initial value vectors not of the correct size" << endl;
		#endif
		return;
	}

	// Resize the resulting vector if necessary
	if (y.size() != y_init.size()) {
		y.resize(y_init.size());
	}
	
	// Vector to hold the derivative
	vector<double> deriv;
	vector<double> u;
	vector<double> p;
	deriv.resize(y.size());
	mSystem->getIVariableAt(u, currTime);
	mSystem->getParameter(p);
	
	// Performs the step
	mSystem->f(deriv, y_init, u, p);
	multiplyVector(deriv, step, deriv);			// deriv = deriv * h
	addVector(y, y_init, deriv);				// y = y+ deriv
	
}

/** Solver for the system using the system values as the initial variables
**/
ODETrace* ODESolverExpEuler::solve() {

	// Checks if the ODE system is in place and that it is consistent
	if (mSystem == 0 || !mSystem->isConsistent()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverExpEuler::solve(1) - No ODE System or inconsistent system" << endl;
		#endif
		return 0;
	}
	
	// Gets the initial values of the system and calls the other solve method
	vector<double> y_init;
	mSystem->getDVariable(y_init);
	return solve(y_init);
}

/** Solver for the system
**/
ODETrace* ODESolverExpEuler::solve(const vector<double>& y_init) {

	// Checks if the ODE system is in place and that it is consistent
	if (mSystem == 0 || !mSystem->isConsistent()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverExpEuler::solve(2) - No ODE System or inconsistent system" << endl;
		#endif
		return 0;
	}
	
	// Gets the attributes
	double start = mAttributeMap["start"];
	double stop = mAttributeMap["stop"];
	double step = mAttributeMap["step"];
	double logStep = mAttributeMap["logstep"];
	
	// Checks the start and stop time are value, i.e. stop > start and step > stop - start
	// Also logStep must be a multiple of step
	if (start >= stop || 
	    step > (stop - start)) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverExpEuler::solve(2) - Error in solver attributes" << endl;
		#endif
		return 0;
	}
	
	// Checks that the initial vector size is correct
	if (y_init.size() != mSystem->dVariableSize()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverExpEuler::solve(2) - Initial value vectors not of the correct size" << endl;
		#endif
		return 0;
	}

	ODETrace* trace = new ODETrace();
	vector<double> y;
	vector<double> u;
	vector<double> p;
	vector<double> deriv;
	double currTime = start;
	double nextLogTime = start + logStep;
	
	// Sets up the variable name list for the trace
	vector<string> variableNameList = mSystem->getDVariableNameList();
	trace->addVariableList(variableNameList);
	
	// Gets the parameters, independent variables
	multiplyVector(y, 1.0, y_init);
	mSystem->getParameter(p);
	
	// Writes initial concentration to trace
	trace->addTraceAt(currTime, y);
	
	while (currTime <= stop) {

		// Gets value for independent variables
		mSystem->getIVariableAt(u, currTime);
		
		// Compute derivative and next time step value
		mSystem->f(deriv, y, u, p);
		multiplyVector(deriv, step, deriv);			// deriv = deriv * h
		addVector(y, y, deriv);						// y = y+ deriv
		
		// Increment time step
		currTime = currTime + step;
		
		// Checks if need to write to trace
		if (currTime >= nextLogTime) {
			trace->addTraceAt(nextLogTime, y);
			nextLogTime = nextLogTime + logStep;
		}
	}

	return trace;
}

/** Default constructor for the implicit Euler solver
**/
ODESolverImpEuler::ODESolverImpEuler() {
	#ifdef ODESOLVER_DEBUG
		cout << "[Info] ODESolverImpEuler::ODESolverImpEuler()- Creating solver (Implicit Euler)" << endl;
	#endif

	// Presets the attributes
	mAttributeMap["start"] = 0.0;
	mAttributeMap["stop"] = 15.0;
	mAttributeMap["step"] = 0.01;
	mAttributeMap["logstep"] = 0.01;
	mAttributeMap["maxiteration"] = 10;
	mAttributeMap["tolerance"] = 0.01;
}

/** Default destructor for the implicit Euler solver
**/
ODESolverImpEuler::~ODESolverImpEuler() {
	#ifdef ODESOLVER_DEBUG
		cout << "[Info] ODESolverImpEuler::~ODESolverImpEuler()- Destroying solver (Implicit Euler)" << endl;
	#endif
}

/** Solver for the system using the system values as the initial variables
**/
ODETrace* ODESolverImpEuler::solve() {
	
	// Checks if the ODE system is in place and that it is consistent
	if (mSystem == 0 || !mSystem->isConsistent()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverImpEuler::solve(1) - No ODE System or inconsistent system" << endl;
		#endif
		return 0;
	}
	
	// Gets the initial values of the system and calls the other solve method
	vector<double> y_init;
	mSystem->getDVariable(y_init);
	return solve(y_init);
}

/** Solver for the system
**/
ODETrace* ODESolverImpEuler::solve(const vector<double>& y_init) {

	// Checks if the ODE system is in place and that it is consistent
	if (mSystem == 0 || !mSystem->isConsistent()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverImpEuler::solve(2) - No ODE System or inconsistent system" << endl;
		#endif
		return 0;
	}
	
	// Gets the attributes
	double start = mAttributeMap["start"];
	double stop = mAttributeMap["stop"];
	double step = mAttributeMap["step"];
	double logStep = mAttributeMap["logstep"];
	
	// Checks the start and stop time are value, i.e. stop > start and step > stop - start
	// Also logStep must be a multiple of step
	if (start >= stop || 
	    step > (stop - start)) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverImpEuler::solve(2) - Error in solver attributes" << endl;
		#endif
		return 0;
	}
	
	// Checks that the initial vector size is correct
	if (y_init.size() != mSystem->dVariableSize()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverImpEuler::solve(2) - Initial value vectors not of the correct size" << endl;
		#endif
		return 0;
	}

	ODETrace* trace = new ODETrace();
	vector<double> y;
	vector<double> delta_y;
	vector<double> u;
	vector<double> p;
	vector<double> deriv;
	double currTime = start;
	double nextLogTime = start + logStep - ODESOLVER_TINY;
	
	// Sets up the variable name list for the trace
	vector<string> variableNameList = mSystem->getDVariableNameList();
	trace->addVariableList(variableNameList);
	
	// Gets the parameters, independent variables
	
	multiplyVector(y, 1.0, y_init);
	mSystem->getParameter(p);
	
	// Writes initial concentration to trace
	trace->addTraceAt(currTime, y);
	
	// Since this is implicit euler, the algorithm is as such
	// x2 = x1 + hf(x2) --> x1 + _x = x1 + hf(x1 + _h)
	//			  -->        _x =         hf(x1) + hf'(x1)_x)
	// 			  -->	   _x - hf'(x1)(_x) = hf(x1)
    // 			  -->       _x ( I - hf'(x1)) = hf(x1)	
	// 			  -->       _x ( I/h - f'(x1)) = f(x1)	
	// 			  -->       _x = ( I/h - f'(x1)) ^ -1 . f(x1)	
	
	// Computes I/h first
	Matrix iMatrix(mSystem->dVariableSize(), mSystem->dVariableSize());
	Matrix jacMatrix(mSystem->dVariableSize(), mSystem->dVariableSize());
	Matrix buffMatrix(mSystem->dVariableSize(), mSystem->dVariableSize());
	iMatrix.fillDiagonal(1.0 / step);

	while (currTime <= stop) {

		// Gets value for independent variables
		mSystem->getIVariableAt(u, currTime + step);

		// Compute the jacobian 		
		mSystem->jacobian(jacMatrix, y, u, p);
		
		// Does I/h - Jac
		buffMatrix.subtract(iMatrix, jacMatrix);
		
		// Inverse the matrix
		buffMatrix.inverse();
		
		// Multiply with the vector f(y)
		mSystem->f(deriv, y, u, p);
		
		buffMatrix.multiplyVector(delta_y, deriv);
		
		addVector(y, y, delta_y);						// y = y+ y_delta
		
		// Increment time step
		currTime = currTime + step;
		
		// Checks if need to write to trace
		if (currTime >= nextLogTime) {
			trace->addTraceAt(currTime, y);
			nextLogTime = nextLogTime + logStep;
		}
	}

	return trace;

}

/** Default constructor for the explicit Runge Kutta 45 solver
**/
ODESolverRungeKutta45::ODESolverRungeKutta45() {
	#ifdef ODESOLVER_DEBUG
		cout << "[Info] ODESolverRungeKutta45::ODESolverRungeKutta45()- Creating solver (Runge Kutta 45)" << endl;
	#endif

	// Presets the attributes
	mAttributeMap["start"] = 0.0;
	mAttributeMap["stop"] = 15.0;
	mAttributeMap["step"] = 0.01;
	mAttributeMap["logstep"] = 0.01;
}

/** Default destructor for the explicit Runge Kutta 45 solver
**/
ODESolverRungeKutta45::~ODESolverRungeKutta45() {
	#ifdef ODESOLVER_DEBUG
		cout << "[Info] ODESolverRungeKutta45::~ODESolverRungeKutta45()- Destroying solver (Runge Kutta 45)" << endl;
	#endif
}

/** Performs a single step of the RK45 algorithm
**/
void ODESolverRungeKutta45::doStep(vector<double>& y, const vector<double>& y_init, double step, double currTime) {

	// Checks if the ODE system is in place and that it is consistent
	if (mSystem == 0 || !mSystem->isConsistent()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverExpEuler::doStep() - No ODE System or inconsistent system" << endl;
		#endif
		return;
	}
	
	// Checks that the initial vector size is correct
	if (y_init.size() != mSystem->dVariableSize()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverExpEuler::doStep() - Initial value vectors not of the correct size" << endl;
		#endif
		return;
	}

	// Resize the resulting vector if necessary
	if (y.size() != y_init.size()) {
		y.resize(y_init.size());
	}
	
	// Vector to hold the derivative
	vector<double> u;
	vector<double> p;
	vector<double> delta_y;
	vector<double> k1, k1_buffer;
	vector<double> k2, k2_buffer;
	vector<double> k3, k3_buffer;
	vector<double> k4, k4_buffer;

	// Allocates space for the vectors
	k1.resize(mSystem->dVariableSize());
	k1_buffer.resize(mSystem->dVariableSize());
	k2.resize(mSystem->dVariableSize());
	k2_buffer.resize(mSystem->dVariableSize());
	k3.resize(mSystem->dVariableSize());
	k3_buffer.resize(mSystem->dVariableSize());
	k4.resize(mSystem->dVariableSize());
	k4_buffer.resize(mSystem->dVariableSize());
	y.resize(mSystem->dVariableSize());
	delta_y.resize(mSystem->dVariableSize());
	mSystem->getParameter(p);

	// First step, compute for k1
	mSystem->getIVariableAt(u, currTime);
	mSystem->f(k1, y_init, u, p);
	multiplyVector(k1, step, k1);						// k1 = hf(t, y)
	
	// Gets the independent variables at time = t + step/2
	mSystem->getIVariableAt(u, currTime + step / 2);

	// Compute the buffer for k1 to compute k2
	multiplyVector(k1_buffer, 0.5, k1);
	addVector(k1_buffer, y_init, k1_buffer);
	mSystem->f(k2, k1_buffer, u, p);
	multiplyVector(k2, step, k2);						// k2 = hf(t + h/2, y + k1/2)

	// Compute the buffer for k2 to compute k3
	multiplyVector(k2_buffer, 0.5, k2);
	addVector(k2_buffer, y_init, k2_buffer);
	mSystem->f(k3, k2_buffer, u, p);
	multiplyVector(k3, step, k3);						// k3 = hf(t + h/2, y + k2/2)
	
	// Compute the buffer for k3 to compute k4
	multiplyVector(k3_buffer, 0.5, k3);
	addVector(k3_buffer, y_init, k3_buffer);
	mSystem->f(k4, k3_buffer, u, p);
	multiplyVector(k4, step, k4);						// k4 = hf(t + h/2, y + k3/2)
	
	// Update the current y
	multiplyVector(k1_buffer, 1.0/6.0, k1);
	multiplyVector(k2_buffer, 1.0/3.0, k2);
	multiplyVector(k3_buffer, 1.0/3.0, k3);
	multiplyVector(k4_buffer, 1.0/6.0, k4);
	addVector(delta_y, k1_buffer, k2_buffer);
	addVector(delta_y, delta_y, k3_buffer);
	addVector(delta_y, delta_y, k4_buffer);
	addVector(y, y_init, delta_y);						// y = y + k1/6 + k2/3 + k3/3 + k4/6
	
}

/** Solver for the system using the system values as the initial variables
**/
ODETrace* ODESolverRungeKutta45::solve() {
	
	// Checks if the ODE system is in place and that it is consistent
	if (mSystem == 0 || !mSystem->isConsistent()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverRungeKutta45::solve(1) - No ODE System or inconsistent system" << endl;
		#endif
		return 0;
	}
	
	// Gets the initial values of the system and calls the other solve method
	vector<double> y_init;
	mSystem->getDVariable(y_init);
	return solve(y_init);
}

/** Solver for the system
**/
ODETrace* ODESolverRungeKutta45::solve(const vector<double>& y_init) {

	// Checks if the ODE system is in place and that it is consistent
	if (mSystem == 0 || !mSystem->isConsistent()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverRungeKutta45::solve(2) - No ODE System or inconsistent system" << endl;
		#endif
		return 0;
	}
	
	// Gets the attributes
	double start = mAttributeMap["start"];
	double stop = mAttributeMap["stop"];
	double step = mAttributeMap["step"];
	double logStep = mAttributeMap["logstep"];
	
	// Checks the start and stop time are value, i.e. stop > start and step > stop - start
	// Also logStep must be a multiple of step
	if (start >= stop || 
	    step > (stop - start)) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverRungeKutta45::solve(2) - Error in solver attributes" << endl;
		#endif
		return 0;
	}
	
	// Checks that the initial vector size is correct
	if (y_init.size() != mSystem->dVariableSize()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverRungeKutta45::solve(2) - Initial value vectors not of the correct size" << endl;
		#endif
		return 0;
	}

	ODETrace* trace = new ODETrace();
	
	vector<double> y;
	vector<double> delta_y;
	vector<double> u;
	vector<double> p;

	vector<double> k1, k1_buffer;
	vector<double> k2, k2_buffer;
	vector<double> k3, k3_buffer;
	vector<double> k4, k4_buffer;

	// Allocates space for the vectors
	k1.resize(mSystem->dVariableSize());
	k1_buffer.resize(mSystem->dVariableSize());
	k2.resize(mSystem->dVariableSize());
	k2_buffer.resize(mSystem->dVariableSize());
	k3.resize(mSystem->dVariableSize());
	k3_buffer.resize(mSystem->dVariableSize());
	k4.resize(mSystem->dVariableSize());
	k4_buffer.resize(mSystem->dVariableSize());
	y.resize(mSystem->dVariableSize());
	delta_y.resize(mSystem->dVariableSize());
	
	double currTime = start;
	double nextLogTime = start + logStep - ODESOLVER_TINY;
	
	// Sets up the variable name list for the trace
	vector<string> variableNameList = mSystem->getDVariableNameList();
	trace->addVariableList(variableNameList);

	// Gets the parameters, independent variables
	multiplyVector(y, 1.0, y_init);
	mSystem->getParameter(p);

	// Writes initial concentration to trace
	trace->addTraceAt(currTime, y);

	while (currTime <= stop) {

		// First step, compute for k1
		mSystem->getIVariableAt(u, currTime);
		mSystem->f(k1, y, u, p);
		multiplyVector(k1, step, k1);						// k1 = hf(t, y)
		
		// Gets the independent variables at time = t + step/2
		mSystem->getIVariableAt(u, currTime + step / 2);

		// Compute the buffer for k1 to compute k2
		multiplyVector(k1_buffer, 0.5, k1);
		addVector(k1_buffer, y, k1_buffer);
		mSystem->f(k2, k1_buffer, u, p);
		multiplyVector(k2, step, k2);						// k2 = hf(t + h/2, y + k1/2)

		// Compute the buffer for k2 to compute k3
		multiplyVector(k2_buffer, 0.5, k2);
		addVector(k2_buffer, y, k2_buffer);
		mSystem->f(k3, k2_buffer, u, p);
		multiplyVector(k3, step, k3);						// k3 = hf(t + h/2, y + k2/2)
		
		// Compute the buffer for k3 to compute k4
		multiplyVector(k3_buffer, 0.5, k3);
		addVector(k3_buffer, y, k3_buffer);
		mSystem->f(k4, k3_buffer, u, p);
		multiplyVector(k4, step, k4);						// k4 = hf(t + h/2, y + k3/2)
		
		// Update the current y
		multiplyVector(k1_buffer, 1.0/6.0, k1);
		multiplyVector(k2_buffer, 1.0/3.0, k2);
		multiplyVector(k3_buffer, 1.0/3.0, k3);
		multiplyVector(k4_buffer, 1.0/6.0, k4);
		addVector(delta_y, k1_buffer, k2_buffer);
		addVector(delta_y, delta_y, k3_buffer);
		addVector(delta_y, delta_y, k4_buffer);
		addVector(y, y, delta_y);							// y = y + k1/6 + k2/3 + k3/3 + k4/6
		
		// Increment time step
		currTime = currTime + step;
		
		// Checks if need to write to trace
		if (currTime >= nextLogTime) {
			trace->addTraceAt(currTime, y);
			nextLogTime = nextLogTime + logStep;
		}
	}

	return trace;

}

/** Default constructor for the explicit Adams-Bashforth-Moulten variable step solver
**/
ODESolverAdams::ODESolverAdams() {
	#ifdef ODESOLVER_DEBUG
		cout << "[Info] ODESolverAdams::ODESolverAdams()- Creating solver (Adams-Bashforth-Moulton)" << endl;
	#endif

	// Presets the attributes
	mAttributeMap["start"] = 0.0;
	mAttributeMap["stop"] = 15.0;
	mAttributeMap["step"] = 0.01;
	mAttributeMap["logstep"] = 0.01;
	
	mAttributeMap["minstep"] = 1e-6;	// This is the minimum step before reporting an error
	mAttributeMap["maxstep"] = 100;	// Maximum internal step
	
	mAttributeMap["reltol"] = 5e-6;	// The relative tolerance
	mAttributeMap["small"] = 1e-4;	// A small number
	
}

/** Default destructor for the explicit Adams-Bashforth-Moulten variable step solver
**/
ODESolverAdams::~ODESolverAdams() {
	#ifdef ODESOLVER_DEBUG
		cout << "[Info] ODESolverAdams::~ODESolverAdams()- Destroying solver (Adams-Bashforth-Moulton)" << endl;
	#endif
}

/** Solver for the system using the system values as the initial variables
**/
ODETrace* ODESolverAdams::solve() {
	
	// Checks if the ODE system is in place and that it is consistent
	if (mSystem == 0 || !mSystem->isConsistent()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverAdams::solve(1) - No ODE System or inconsistent system" << endl;
		#endif
		return 0;
	}
	
	// Gets the initial values of the system and calls the other solve method
	vector<double> y_init;
	mSystem->getDVariable(y_init);
	return solve(y_init);
}

/** Solver for the system
**/
ODETrace* ODESolverAdams::solve(const vector<double>& y_init) {

	// Checks if the ODE system is in place and that it is consistent
	if (mSystem == 0 || !mSystem->isConsistent()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverAdams::solve(2) - No ODE System or inconsistent system" << endl;
		#endif
		return 0;
	}

	// Gets the attributes
	double start = mAttributeMap["start"];
	double stop = mAttributeMap["stop"];
	double step = mAttributeMap["step"];
	double logstep = mAttributeMap["logstep"];
	
	double reltol = mAttributeMap["reltol"];
	double small = mAttributeMap["small"];
	double minstep = mAttributeMap["minstep"];
	double maxstep = mAttributeMap["maxstep"];
	
	// Checks the start and stop time are value, i.e. stop > start and step > stop - start
	// Also logStep must be a multiple of step
	if (start >= stop || 
	    step > (stop - start)) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverAdams::solve(2) - Error in solver attributes" << endl;
		#endif
		return 0;
	}
	
	// Checks that the initial vector size is correct
	if (y_init.size() != mSystem->dVariableSize()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverAdams::solve(2) - Initial value vectors not of the correct size" << endl;
		#endif
		return 0;
	}

	ODETrace* trace = new ODETrace();
	
	vector<double> y;
	vector<double> p;
	y.resize(mSystem->dVariableSize());
	
	double currTime = start;
	double nextLogTime = start + logstep - ODESOLVER_TINY;

	// Sets up the variable name list for the trace
	vector<string> variableNameList = mSystem->getDVariableNameList();
	trace->addVariableList(variableNameList);

	// Gets the parameters, independent variables
	multiplyVector(y, 1.0, y_init);
	mSystem->getParameter(p);

	// Writes initial concentration to trace
	trace->addTraceAt(currTime, y);

	while (currTime <= stop) {
	

	}
	
	return trace;
	
}

/** Initializes the first 3 points using runge kutta methods
**/
bool ODESolverAdams::init(vector<double>& y1, vector<double>& y2, vector<double>& y3, const vector<double>& y0, double time, double step) {

	// Checks if the ODE system is in place and that it is consistent
	if (mSystem == 0 || !mSystem->isConsistent()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverAdams::init() - No ODE System or inconsistent system" << endl;
		#endif
		return false;
	}

	// Checks the size for y0 and resize for the rest
	if (y0.size() != mSystem->dVariableSize()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverAdams::init() - Initial value vectors not of the correct size" << endl;
		#endif
		return false;	
	}
	
	if (y1.size() != y0.size()) y1.resize(y0.size());
	if (y2.size() != y0.size()) y2.resize(y0.size());
	if (y3.size() != y0.size()) y3.resize(y0.size());
	
	// Vectors to hold temporary values when computing
	vector<double> k1, k1_buffer;
	vector<double> k2, k2_buffer;
	vector<double> k3, k3_buffer;
	vector<double> k4, k4_buffer;
	vector<double> u;
	vector<double> delta_y;

	// Resize all the vectors
	k1.resize(y0.size());
	k1_buffer.resize(y0.size());
	k2.resize(y0.size());
	k2_buffer.resize(y0.size());
	k3.resize(y0.size());
	k3_buffer.resize(y0.size());
	k4.resize(y0.size());
	k4_buffer.resize(y0.size());
	u.resize(y0.size());
	delta_y.resize(y0.size());
	
	////////////////////
	// Compute for y1
	////////////////////
	{
	// First step, compute for k1
	mSystem->getIVariableAt(u, time);
	mSystem->f(k1, y0, u);
	multiplyVector(k1, step, k1);						
	
	// Gets the independent variables at time = t + step/2
	mSystem->getIVariableAt(u, time + step / 2);

	// Compute the buffer for k1 to compute k2
	multiplyVector(k1_buffer, 0.5, k1);
	addVector(k1_buffer, y0, k1_buffer);
	mSystem->f(k2, k1_buffer, u);
	multiplyVector(k2, step, k2);						

	// Compute the buffer for k2 to compute k3
	multiplyVector(k2_buffer, 0.5, k2);
	addVector(k2_buffer, y0, k2_buffer);
	mSystem->f(k3, k2_buffer, u);
	multiplyVector(k3, step, k3);						
	
	// Compute the buffer for k3 to compute k4
	multiplyVector(k3_buffer, 0.5, k3);
	addVector(k3_buffer, y0, k3_buffer);
	mSystem->f(k4, k3_buffer, u);
	multiplyVector(k4, step, k4);						
	
	// Update the current y
	multiplyVector(k1_buffer, 1.0/6.0, k1);
	multiplyVector(k2_buffer, 1.0/3.0, k2);
	multiplyVector(k3_buffer, 1.0/3.0, k3);
	multiplyVector(k4_buffer, 1.0/6.0, k4);
	addVector(delta_y, k1_buffer, k2_buffer);
	addVector(delta_y, delta_y, k3_buffer);
	addVector(delta_y, delta_y, k4_buffer);
	addVector(y1, y0, delta_y);							// y = y + k1/6 + k2/3 + k3/3 + k4/6
	}
	
	// Update time
	time = time + step;
	
	////////////////////
	// Compute for y2
	////////////////////
	{
	// First step, compute for k1
	mSystem->getIVariableAt(u, time);
	mSystem->f(k1, y1, u);
	multiplyVector(k1, step, k1);						
	
	// Gets the independent variables at time = t + step/2
	mSystem->getIVariableAt(u, time + step / 2);

	// Compute the buffer for k1 to compute k2
	multiplyVector(k1_buffer, 0.5, k1);
	addVector(k1_buffer, y1, k1_buffer);
	mSystem->f(k2, k1_buffer, u);
	multiplyVector(k2, step, k2);						

	// Compute the buffer for k2 to compute k3
	multiplyVector(k2_buffer, 0.5, k2);
	addVector(k2_buffer, y1, k2_buffer);
	mSystem->f(k3, k2_buffer, u);
	multiplyVector(k3, step, k3);						
	
	// Compute the buffer for k3 to compute k4
	multiplyVector(k3_buffer, 0.5, k3);
	addVector(k3_buffer, y1, k3_buffer);
	mSystem->f(k4, k3_buffer, u);
	multiplyVector(k4, step, k4);						
	
	// Update the current y
	multiplyVector(k1_buffer, 1.0/6.0, k1);
	multiplyVector(k2_buffer, 1.0/3.0, k2);
	multiplyVector(k3_buffer, 1.0/3.0, k3);
	multiplyVector(k4_buffer, 1.0/6.0, k4);
	addVector(delta_y, k1_buffer, k2_buffer);
	addVector(delta_y, delta_y, k3_buffer);
	addVector(delta_y, delta_y, k4_buffer);
	addVector(y2, y1, delta_y);							// y = y + k1/6 + k2/3 + k3/3 + k4/6
	}
	
	// Update time
	time = time + step;

	////////////////////
	// Compute for y3
	////////////////////
	{
	// First step, compute for k1
	mSystem->getIVariableAt(u, time);
	mSystem->f(k1, y2, u);
	multiplyVector(k1, step, k1);						
	
	// Gets the independent variables at time = t + step/2
	mSystem->getIVariableAt(u, time + step / 2);

	// Compute the buffer for k1 to compute k2
	multiplyVector(k1_buffer, 0.5, k1);
	addVector(k1_buffer, y2, k1_buffer);
	mSystem->f(k2, k1_buffer, u);
	multiplyVector(k2, step, k2);						

	// Compute the buffer for k2 to compute k3
	multiplyVector(k2_buffer, 0.5, k2);
	addVector(k2_buffer, y2, k2_buffer);
	mSystem->f(k3, k2_buffer, u);
	multiplyVector(k3, step, k3);						
	
	// Compute the buffer for k3 to compute k4
	multiplyVector(k3_buffer, 0.5, k3);
	addVector(k3_buffer, y2, k3_buffer);
	mSystem->f(k4, k3_buffer, u);
	multiplyVector(k4, step, k4);						
	
	// Update the current y
	multiplyVector(k1_buffer, 1.0/6.0, k1);
	multiplyVector(k2_buffer, 1.0/3.0, k2);
	multiplyVector(k3_buffer, 1.0/3.0, k3);
	multiplyVector(k4_buffer, 1.0/6.0, k4);
	addVector(delta_y, k1_buffer, k2_buffer);
	addVector(delta_y, delta_y, k3_buffer);
	addVector(delta_y, delta_y, k4_buffer);
	addVector(y3, y2, delta_y);							// y = y + k1/6 + k2/3 + k3/3 + k4/6
	}
	
	// Check all three vectors to ensure that the error is still kept within limits
	// To be implemented later
	
	return true;
}	

bool ODESolverAdams::predict(vector<double>& y_pred, 
                             const vector<double>& f3, 
							 const vector<double>& f2, 
							 const vector<double>& f1, 
							 const vector<double>& f0, 
							 const vector<double>& y0,
							 double step) {
	
	vector<double> buffer;
	vector<double> f0_buffer, f1_buffer, f2_buffer, f3_buffer;
	buffer.resize(y0.size());
	f0_buffer.resize(y0.size());
	f1_buffer.resize(y0.size());
	f2_buffer.resize(y0.size());
	f3_buffer.resize(y0.size());
	
	// Performs this operation
	// y_pred = y0 + h/24 (-9 x f3 + 37 x f2 -59 x f1 + 55 x f0)
	multiplyVector(f0_buffer, 55.0, f0);
	multiplyVector(f1_buffer, -59.0, f1);
	multiplyVector(f2_buffer, 37.0, f2);
	multiplyVector(f3_buffer, -9.0, f3);
	addVector(buffer, f0_buffer, f1_buffer);
	addVector(buffer, buffer, f2_buffer);
	addVector(buffer, buffer, f3_buffer);
	multiplyVector(buffer, (step / 24.0), buffer);
	addVector(y_pred, y0, buffer);
	
	return true;
}
 
bool ODESolverAdams::correct(vector<double>& y_corr, 
                             const vector<double>& f2, 
							 const vector<double>& f1, 
							 const vector<double>& f0, 
							 const vector<double>& f_pred, 
							 const vector<double>& y0,
							 double step) {

	vector<double> buffer;
	vector<double> f0_buffer, f1_buffer, f_pred_buffer;
	buffer.resize(y0.size());
	f0_buffer.resize(y0.size());
	f1_buffer.resize(y0.size());
	f_pred_buffer.resize(y0.size());
	
	// Performs this operation
	// y_corr = y0 + h/24 (f2 - 5 x f1 + 19 x f0 + 9 x f_pred)
	multiplyVector(f0_buffer, 19.0, f0);
	multiplyVector(f1_buffer, -5.0, f1);
	multiplyVector(f_pred_buffer, 9.0, f_pred);
	addVector(buffer, f_pred_buffer, f0_buffer);
	addVector(buffer, buffer, f1_buffer);
	addVector(buffer, buffer, f2);
	multiplyVector(buffer, (step / 24.0), buffer);
	addVector(y_corr, y0, buffer);
	
	return true;
							 
}
	
/** Default constructor for the adaptive step solver
**/
ODESolverAdaptiveStep::ODESolverAdaptiveStep() {

	#ifdef ODESOLVER_DEBUG
		cout << "[Info] ODESolverAdaptiveStep::ODESolverAdaptiveStep()- Creating solver (Adaptive step)" << endl;
	#endif

	// Presets the attributes
	mAttributeMap["start"] = 0.0;
	mAttributeMap["stop"] = 15.0;
	mAttributeMap["step"] = 0.01;
	mAttributeMap["logstep"] = 0.01;
	mAttributeMap["uppertolerance"] = 0.1;
	mAttributeMap["lowertolerance"] = 1.0e-6;
	mAttributeMap["maxstep"] = 10000;
	mAttributeMap["minstep"] = 0.00001;

}

/** Default destructor for the adaptive step solver
**/
ODESolverAdaptiveStep::~ODESolverAdaptiveStep() {
	#ifdef ODESOLVER_DEBUG
		cout << "[Info] ODESolverAdaptiveStep::~ODESolverAdaptiveStep()- Destroying solver (Adaptive step)" << endl;
	#endif
}

ODETrace* ODESolverAdaptiveStep::solve() {

	// Checks if the ODE system is in place and that it is consistent
	if (mSystem == 0 || !mSystem->isConsistent()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverAdaptiveStep::solve(1) - No ODE System or inconsistent system" << endl;
		#endif
		return 0;
	}
	
	// Gets the initial values of the system and calls the other solve method
	vector<double> y_init;
	mSystem->getDVariable(y_init);
	return solve(y_init);

}

ODETrace* ODESolverAdaptiveStep::solve(const vector<double>& y_init) {

	// Checks if the ODE system is in place and that it is consistent
	if (mSystem == 0 || !mSystem->isConsistent()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverAdaptiveStep::solve(2) - No ODE System or inconsistent system" << endl;
		#endif
		return 0;
	}

	// Gets the attributes
	double start = mAttributeMap["start"];
	double stop = mAttributeMap["stop"];
	double step = mAttributeMap["step"];
	double logstep = mAttributeMap["logstep"];
	double uppertolerance = mAttributeMap["uppertolerance"];
	double lowertolerance = mAttributeMap["lowertolerance"];
	double minstep = mAttributeMap["minstep"];
	double maxstep = mAttributeMap["maxstep"];
	
	// Checks the start and stop time are value, i.e. stop > start and step > stop - start
	// Also logStep must be a multiple of step
	if (start >= stop || 
	    step > (stop - start)) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverAdaptiveStep::solve(2) - Error in solver attributes" << endl;
		#endif
		return 0;
	}
	
	// Checks that the initial vector size is correct
	if (y_init.size() != mSystem->dVariableSize()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverAdaptiveStep::solve(2) - Initial value vectors not of the correct size" << endl;
		#endif
		return 0;
	}

	ODETrace* trace = new ODETrace();
	
	vector<double> y;
	vector<double> p;
	y.resize(mSystem->dVariableSize());

	// Creates an addition 2 solvers for estimation of Local Truncation Error
	// In this solver we choose the explicit euler and rk4
	ODESolverExpEuler solver1;
	ODESolverRungeKutta45 solver2;
	
	// Loads the system into the solvers
	solver1.loadODESystem(mSystem);
	solver2.loadODESystem(mSystem);
	
	double currTime = start;
	double currStep = step;
	double nextLogTime = start + logstep;

	// Sets up the variable name list for the trace
	vector<string> variableNameList = mSystem->getDVariableNameList();
	trace->addVariableList(variableNameList);

	// Gets the parameters, independent variables
	multiplyVector(y, 1.0, y_init);
	mSystem->getParameter(p);
	
	// Writes initial concentration to trace
	trace->addTraceAt(currTime, y);

	// Prepares the two vectors for outputs of the 2 solvers
	vector<double> y1;
	vector<double> y2;
	vector<double> y_log;
	double error;
	
	// Resize the vectors
	y1.resize(y.size());
	y2.resize(y.size());
	y_log.resize(y.size());
	
	while (currTime <= stop) {

		// If the current step is smaller than the minimum step, then have to 
		// exit
		if (currStep < minstep) {
			#ifdef ODESOLVER_DEBUG
				cout << "[Error] ODESolverAdaptiveStep::solve(2) - Step smaller than minimal step" << endl;
			#endif
			break;
		}
		
		// If current step is smaller than maximum step, reduce it by half
		// else carry on
		if (currStep > maxstep) {
			currStep = currStep / 2;
		} else {
			
			// Solve using the current step
			solver1.setAttribute("step", currStep);
			solver2.setAttribute("step", currStep);
			
			// Solve for the 2 solvers
			solver1.doStep(y1, y, currStep, currTime);
			solver2.doStep(y2, y, currStep, currTime);
			
			// Estimate error
			error = estimateError(y1, y2);
		
			if (error > uppertolerance) {
				currStep = currStep / 2.0;
			} else {

				// Increment time step
				currTime = currTime + currStep;

				// Checks if there is a need to record the log (use y1)
				while ((nextLogTime < (currTime + currStep)) && (nextLogTime <= stop)) {
					interpolate(y_log, y, y2, (nextLogTime - currTime) / (currStep));
					trace->addTraceAt(nextLogTime, y_log);
					nextLogTime = nextLogTime + logstep;
				}
				
				// Updates y with y1
				multiplyVector(y, 1.0, y2);
				
				if (error < lowertolerance && currStep <= maxstep) {
					currStep = currStep * 2.0;
				}
			}
		}
	}

	return trace;
	
}

/** Estimates the error between y1 and y2
* The error is the maximal difference between the common entries of y1
* and y2, i.e. they need not be of the same size but must have at least 1 entry each
**/
double ODESolverAdaptiveStep::estimateError(const vector<double>& y1, const vector<double>& y2) {

	if (y1.size() < 1 || y2.size() < 1) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverAdaptiveStep::estimateError() - Empty vector" << endl;
		#endif
		return -1.0;
	}
	
	int indexSize;
	double error = -1.0;
	double tmpError;
	
	if (y1.size() < y2.size()) {
		indexSize = y1.size();
	} else {
		indexSize = y2.size();
	}
	
	for (int i = 0; i < indexSize; i++) {
		tmpError = fabs(y1[i] - y2[i]);
		if (tmpError > error) {
			error = tmpError;
		}
	}
	
	return error;
}

/** Interpolates y1 and y2 using the given factor
**/
void ODESolverAdaptiveStep::interpolate(vector<double>& result, const vector<double>& y1, const vector<double>& y2, double factor) {
	
	// Checks the size of the two input vectors
	if (y1.size() != y2.size()) {
		#ifdef ODESOLVER_DEBUG
			cout << "[Error] ODESolverAdaptiveStep::interpolate() - Input vector size not compatible" << endl;
		#endif
		return;
	}
	
	// Resize the result if necessary
	if (result.size() != y1.size()) {
		result.resize(y1.size());
	}

	// Performs the interpolation
	for (int i = 0; i < y1.size(); i++) {
		result[i] = y1[i] + ((y2[i] - y1[i]) * factor);
	}
	
}

}

