/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: odesolver.h
**  
** Description:
** This is the base class for the ode solvers that will be included in this package. It will
** define the main interface that all the solvers should share. This is an abstract base class
** so it should not be instantiated.
** Included here is also the most commonly used solver, i.e. explicit euler
**
** Last modified: 
** 03 April 2007
** 
** Comments:
** (1) The parameters for the solvers are kept in a map. The purpose of the constructor
**      for each concrete class is then to set those parameters to default values
** (2) Subsequent solvers we will only need to write the constructor and the solve methods
****************************************************************************/

#ifndef ODESOLVER_H_
#define ODESOLVER_H_

/* For debugging. Comment out to suppress debugging lines */
// #define ODESOLVER_DEBUG

/* Definition of constants to be used */
#define ODESOLVER_TINY 1.0e-10
#define ODESOLVER_MINTHRESHOLD 1.0e-20

/* Standard Includes */
#include <iostream>
#include <sstream>
#include <vector>
#include <map>

/* Application Includes */
#include "odecommon.h"
#include "odesystem.h"
#include "odetrace.h"

using namespace std;

namespace ode {

class ODESolver {
public:

	/* Constructors and destructors */
	ODESolver();
	~ODESolver();

	/* Loads the ODE system etc */
	void loadODESystem(ODESystem* system);
	void unloadODESystem();
	void setAttribute(string attribName, double value);
	
	/* Solve the system */
	// virtual void doStep(vector<double>& y, const vector<double>& y_init, double step, double currTime);
	virtual ODETrace* solve() = 0;
	virtual ODETrace* solve(const vector<double>& y_init) = 0;

protected:
	ODESystem* mSystem;
	map<string, double> mAttributeMap;
	
};

class ODESolverExpEuler : public ODESolver {
public:
	/* Constructors and destructors */
	ODESolverExpEuler();
	virtual ~ODESolverExpEuler();
	
	/* Solve the system */
	void doStep(vector<double>& y, const vector<double>& y_init, double step, double currTime);
	ODETrace* solve();
	ODETrace* solve(const vector<double>& y_init);
};

class ODESolverImpEuler : public ODESolver {
public:
	/* Constructors and destructors */
	ODESolverImpEuler();
	virtual ~ODESolverImpEuler();
	
	/* Solve the system */
	ODETrace* solve();
	ODETrace* solve(const vector<double>& y_init);
	
};

class ODESolverRungeKutta45 : public ODESolver {
public:
	/* Constructors and destructors */
	ODESolverRungeKutta45();
	virtual ~ODESolverRungeKutta45();
	
	/* Solve the system */
	void doStep(vector<double>& y, const vector<double>& y_init, double step, double currTime);
	ODETrace* solve();
	ODETrace* solve(const vector<double>& y_init);
};

class ODESolverAdams : public ODESolver {
public:
	/* Constructors and destructors */
	ODESolverAdams();
	virtual ~ODESolverAdams();
	
	/* Solve the system */
	ODETrace* solve();
	ODETrace* solve(const vector<double>& y_init);
	
protected:
	/* Initialize the vectors */
	bool init(vector<double>& y1, vector<double>& y2, vector<double>& y3, const vector<double>& y0, double time, double step);
	bool predict(vector<double>& y_pred, const vector<double>& f3, const vector<double>& f2, const vector<double>& f1, const vector<double>& f0, const vector<double>& y0, double step);
	bool correct(vector<double>& y_corr, const vector<double>& f2, const vector<double>& f1, const vector<double>& f0, const vector<double>& f_pred, const vector<double>& y0, double step);
	double error();
};

class ODESolverAdaptiveStep : public ODESolver  {
public:
	/* Constructors and destructors */
	ODESolverAdaptiveStep();
	virtual ~ODESolverAdaptiveStep();
	
	/* Solve the system */
	ODETrace* solve();
	ODETrace* solve(const vector<double>& y_init);
	
protected:
	double estimateError(const vector<double>& y1, const vector<double>& y2);
	void interpolate(vector<double>& result, const vector<double>& y1, const vector<double>& y2, double factor);
};

}

#endif /* ODESOLVER_H_ */
