/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: odesystem.cpp
**  
** Description:
** This is the implementation of a system of first order ordinary differential equations expressed
** in the explicit form. The implementation of the system of ODE is such that it is a stateful system,
** i.e. it keeps a record of the current values of all the variables and parameters. However, its state
** can also be kept external by maintaining 3 vectors (for parameters, dependent variables and independent
** variables) and also a t variable to represent time.
** The evaluation of the system can either use the internal state of the system, or passing in the necessary
** state vectors to be handled on the spot.

** Last modified: 
** 26 February 2007
** 
** Comments:
**
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <sstream>
#include <vector>
#include <math.h>

/* Application Includes */
#include "odecommon.h"
#include "odesystem.h"
#include "odenode.h"
#include "odeparser.h"

using namespace std;
using namespace matrix;

namespace ode {

/** Default constructor
**/
ODESystem::ODESystem(string name) : mName(name), mTime(0.0) {

	#ifdef ODESYSTEM_DEBUG
		cout << "[Info] ODESystem::ODESystem() - Creating ODESystem " << name << endl;
	#endif

}

/** Copy constructor
**/
ODESystem::ODESystem(const ODESystem& src) {
	copyFrom(src);
}

/** Assignment constructor
**/
ODESystem& ODESystem::operator=(const ODESystem& src) {

	// Checks for self assignment
	if (this == &src) {
		return (*this);	
	}
	
	// Deletes all the parameter and variable vectors
	mParameterNameList.clear();
	mParameterList.clear();
	mIVariableNameList.clear();
	mIVariableList.clear();
	mDVariableNameList.clear();
	mDVariableList.clear();
	
	// Deletes all the nodes and equations
	mEquationList.clear();
	vector<ODENode*>::iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		if ((*iter) != 0) {
			delete (*iter);
		}
		iter++;
	}
	vector<ODEFunction*>::iterator func_iter = mFunctionList.begin();
	while (func_iter != mFunctionList.end()) {
		if ((*func_iter) != 0) {
			delete (*func_iter);
		}
		func_iter++;
	}
	mNodeList.clear();
	mEquationList.clear();
	mFunctionList.clear();
	
	// Unsets the parent and context
	mName = "";
	mTime = 0.0;
	
	// Copy contents
	copyFrom(src);
	
	return (*this);
}

/** Default destructor
**/
ODESystem::~ODESystem() {

	#ifdef ODESYSTEM_DEBUG
		cout << "[Info] ODESystem::~ODESystem() - Destroying ODESystem " << mName << endl;
	#endif
	
	// Deletes all the parameter and variable vectors
	mParameterNameList.clear();
	mParameterList.clear();
	mIVariableNameList.clear();
	mIVariableList.clear();
	mDVariableNameList.clear();
	mDVariableList.clear();
	
	// Deletes all the nodes and equations
	vector<ODENode*>::iterator iter = mNodeList.begin();
	while (iter != mNodeList.end()) {
		if ((*iter) != 0) {
			delete (*iter);
		}
		iter++;
	}
	vector<ODEFunction*>::iterator func_iter = mFunctionList.begin();
	while (func_iter != mFunctionList.end()) {
		if ((*func_iter) != 0) {
			delete (*func_iter);
		}
		func_iter++;
	}
	mNodeList.clear();
	mEquationList.clear();
	mFunctionList.clear();
}

/** Sets the name of the system
* There are no restrictions on the name of the system
**/
void ODESystem::setName(string name) {
	mName = name;
}

/** Returns the name of the system
**/
string ODESystem::getName() {
	return mName;
}

/** Adds a parameter to the system
* Checks if the parameter is valid and if it is not already present in the
* system before adding it in
**/
bool ODESystem::addParameter(const string parameter, double value) {

	if (!isValidSymbol(parameter) || hasSymbol(parameter)) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::addParameter() - Invalid parameter name " << parameter << endl;
		#endif	
		return false;
	}

	mParameterNameList.push_back(parameter);
	mParameterList.push_back(value);
	return true;
}

/** Removes a parameter from the system
* The nodes are not checked so there is a possibility of inconsistency after this
* operation.
**/
bool ODESystem::removeParameter(const string parameter) {

	// Looks for the index of the parameter
	int index = indexOfParameter(parameter);
	
	if (index == -1) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::removeParameter(1) - Parameter not found: " << parameter << endl;
		#endif	
		return false;
	} else {
		return removeParameter(index);
	}

}

/** Removes a parameter from the system given the index
* The nodes are not checked so there is a possibility of inconsistency after this
* operation.
**/
bool ODESystem::removeParameter(const int index) {

	// Checks for valid index
	if (index < 0 || index >= mParameterNameList.size()) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::removeParameter(2) - Invalid index " << index << endl;
		#endif	
		return false;
	}
	
	mParameterNameList.erase(mParameterNameList.begin() + index);
	mParameterList.erase(mParameterList.begin() + index);
	return true;
	
}

/** Adds an independent variable into the system
**/
bool ODESystem::addIVariable(const string variable, double value) {

	if (!isValidSymbol(variable) || hasSymbol(variable)) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::addIVariable() - Invalid variable name " << variable << endl;
		#endif	
		return false;
	}

	mIVariableNameList.push_back(variable);
	mIVariableList.push_back(value);
	
	// Adds an entry in the equation and the node list
	mFunctionList.push_back(0);
	
	return true;

}

/** Removes an independent variable from the system
* The nodes are not checked so there is a possibility of inconsistency after this
* operation.
**/
bool ODESystem::removeIVariable(const string variable) {

	// Looks for the index of the parameter
	int index = indexOfIVariable(variable);
	
	if (index == -1) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::removeIVariable(1) - Variable not found: " << variable << endl;
		#endif	
		return false;
	} else {
		return removeIVariable(index);
	}

}

/** Removes an independent variable from the system given the index
* The nodes are not checked so there is a possibility of inconsistency after this
* operation.
**/
bool ODESystem::removeIVariable(const int index) {

	// Checks for valid index
	if (index < 0 || index >= mIVariableNameList.size()) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::removeIVariable(2) - Invalid index " << index << endl;
		#endif	
		return false;
	}
	
	mIVariableNameList.erase(mIVariableNameList.begin() + index);
	mIVariableList.erase(mIVariableList.begin() + index);
	
	// Deletes the function
	if (mFunctionList[index] != 0) {
		delete mFunctionList[index];
	}
	mFunctionList.erase(mFunctionList.begin() + index);
	
	return true;

}

/** Adds a dependent variable into the system
**/
bool ODESystem::addDVariable(const string variable, double value) {

	if (!isValidSymbol(variable) || hasSymbol(variable)) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::addDVariable() - Invalid variable name " << variable << endl;
		#endif	
		return false;
	}

	mDVariableNameList.push_back(variable);
	mDVariableList.push_back(value);
	
	// Adds an entry in the equation and the node list
	mNodeList.push_back(0);
	mEquationList.push_back("0");
	
	return true;

}

/** Removes a dependent variable from the system
* The nodes are not checked so there is a possibility of inconsistency after this
* operation.
**/
bool ODESystem::removeDVariable(const string variable) {

	// Looks for the index of the parameter
	int index = indexOfDVariable(variable);
	
	if (index == -1) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::removeDVariable(1) - Variable not found: " << variable << endl;
		#endif	
		return false;
	} else {
		return removeDVariable(index);
	}

}

/** Removes a dependent variable from the system given the index
* The nodes are not checked so there is a possibility of inconsistency after this
* operation.
**/
bool ODESystem::removeDVariable(const int index) {

	// Checks for valid index
	if (index < 0 || index >= mDVariableNameList.size()) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::removeDVariable(2) - Invalid index " << index << endl;
		#endif	
		return false;
	}
	
	mDVariableNameList.erase(mDVariableNameList.begin() + index);
	mDVariableList.erase(mDVariableList.begin() + index);
	
	// Deletes off the value in the node list
	if (mNodeList[index] != 0) {
		delete mNodeList[index];
	}
	mNodeList.erase(mNodeList.begin() + index);
	mEquationList.erase(mEquationList.begin() + index);
	
	return true;

}

/** Sets the values of the parameters in the system
**/
void ODESystem::setParameter(const vector<double>& parameterList) {

	// Checks the size of the parameter list
	if (parameterList.size() != mParameterList.size()) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::setParameter(1) - Incompatible parameter list size" << endl;
		#endif	
		return;
	}
	
	// Replace all the values of the mParameterList with the given list
	for (int i = 0; i < mParameterList.size(); i++) {
		mParameterList[i] = parameterList[i];
	}
}

/** Sets the value of the given parameter to a given value
**/
void ODESystem::setParameter(const string parameter, double value) {

	// Gets the index of the parameter
	int index = indexOfParameter(parameter);
	
	if (index == -1) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::setParameter(2) - Parameter " << parameter << " not found" << endl;
		#endif	
		return;
	} else {
		setParameter(index, value);
	}
}

/** Sets the value of the parameter given the index and its new value
**/
void ODESystem::setParameter(const int index, double value) {

	// Checks the index
	if (index < 0 || index >= mParameterList.size()) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::setParameter(3) - Index out of bounds (" << index << ")" << endl;
		#endif	
		return;
	}
	
	mParameterList[index] = value;
	return;
}

/** Returns the parameter given the parameter name.
* If the parameter is not found, a NaN is returned
**/
double ODESystem::getParameter(const string parameter) const {

	// Gets the index of the parameter
	int index = indexOfParameter(parameter);
	
	if (index == -1) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::getParameter(1) - Parameter " << parameter << " not found" << endl;
		#endif	
		return sqrt(-1);
	} else {
		double value = getParameter(index);
		return value;
	}

}

/** Returns the parameter given its index
* If the index is out of bounds, then a NaN is returned
**/
double ODESystem::getParameter(const int index) const {

	// Checks the index if it is out of bounds
	if (index < 0 || index > mParameterList.size()) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::getParameter(2) - Index out of bounds (" << index << ")" << endl;
		#endif	
		return sqrt(-1);
	}
	
	return mParameterList[index];
}

/** Returns all the parameter values into the given vector reference
* If the vector reference is not of the correct size, resize it
**/
void ODESystem::getParameter(vector<double>& result) const {

	if (result.size() != mParameterList.size()) {
		result.clear();
		for (int i = 0; i < mParameterList.size(); i++) {
			result.push_back(mParameterList[i]);
		}
		return;
	} else {
		for (int i = 0; i < mParameterList.size(); i++) {
			result[i] = mParameterList[i];
		}
	}
}

/** Given the parameter, locates the index in the parameter name list
**/
int ODESystem::indexOfParameter(const string parameter) const {

	int index = -1;

	// Loops through the name list and returns the index
	for (int i = 0; i < mParameterNameList.size(); i++) {
		if (mParameterNameList[i] == parameter) {
			index = i;
			break;
		}
	}
	
	return index;
}

/** Gets the parameter name at the given index
**/
string ODESystem::getParameterName(const int index) const {
	if (index < 0 || index >= mParameterNameList.size()) {
		return "";
	} else {
		return mParameterNameList[index];
	}
}

/** Returns the list of parameter name list
**/
vector<string> ODESystem::getParameterNameList() const {
	return mParameterNameList;
}

/** Sets the values of the independent variables in the system
**/
void ODESystem::setIVariable(const vector<double>& variableList) {

	// Checks the size of the variable list
	if (variableList.size() != mIVariableList.size()) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::setIVariable(1) - Incompatible variable list size" << endl;
		#endif	
		return;
	}
	
	// Replace all the values of the mIVariableList with the given list
	for (int i = 0; i < mIVariableList.size(); i++) {
		mIVariableList[i] = variableList[i];
	}

}

/** Sets the value of the given independent variable to a given value
**/
void ODESystem::setIVariable(const string variable, double value) {

	// Gets the index of the variable
	int index = indexOfIVariable(variable);
	
	if (index == -1) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::setIVariable(2) - Variable " << variable << " not found" << endl;
		#endif	
		return;
	} else {
		setIVariable(index, value);
	}

}

/** Sets the value of the independent variable given the index and its new value
**/
void ODESystem::setIVariable(const int index, double value) {
	// Checks the index
	if (index < 0 || index >= mIVariableList.size()) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::setIVariable(3) - Index out of bounds (" << index << ")" << endl;
		#endif	
		return;
	}
	
	mIVariableList[index] = value;
	return;
}

/** Returns the independent variable given the variable name.
* If the variable is not found, a NaN is returned
**/
double ODESystem::getIVariable(const string variable) const {

	// Gets the index of the variable
	int index = indexOfIVariable(variable);
	
	if (index == -1) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::getIVariable(1) - Variable " << variable << " not found" << endl;
		#endif	
		return sqrt(-1);
	} else {
		return getIVariable(index);
	}

}

/** Returns the independent variable given its index
* If the index is out of bounds, then a NaN is returned
**/
double ODESystem::getIVariable(const int index) const {

	// Checks the index if it is out of bounds
	if (index < 0 || index > mIVariableList.size()) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::getIVariable(2) - Index out of bounds (" << index << ")" << endl;
		#endif	
		return sqrt(-1);
	}
	
	return mIVariableList[index];

}

/** Returns all the independent values into the given vector reference
* If the vector reference is not of the correct size, resize it
**/
void ODESystem::getIVariable(vector<double>& result) const {

	if (result.size() != mIVariableList.size()) {
		result.clear();
		for (int i = 0; i < mIVariableList.size(); i++) {
			result.push_back(mIVariableList[i]);
		}
		return;
	} else {
		for (int i = 0; i < mIVariableList.size(); i++) {
			result[i] = mIVariableList[i];
		}
	}

}

/** Get the values of the independent variable at the specific time
* This is only possible with the independent variables
**/
void ODESystem::getIVariableAt(vector<double>& result, double time) const {

	double value;
	
	if (result.size() != mIVariableList.size()) {
		result.clear();
		for (int i = 0; i < mFunctionList.size(); i++) {
			if (mFunctionList[i] != 0) {
				value = mFunctionList[i]->f(time);
			} else {
				value = 0.0;
			}
			result.push_back(value);
		}
		return;
	} else {
		for (int i = 0; i < mFunctionList.size(); i++) {
			if (mFunctionList[i] != 0) {
				value = mFunctionList[i]->f(time);
			} else {
				value = 0.0;
			}
			result[i] = value;
		}
	}

}

/** Given the independent variable, locates the index in the variable name list
**/
int ODESystem::indexOfIVariable(const string variable) const {

	int index = -1;
	
	// Loops through the name list and returns the index
	for (int i = 0; i < mIVariableNameList.size(); i++) {
		if (mIVariableNameList[i] == variable) {
			index = i;
			break;
		}
	}
	
	return index;

}

/** Gets the name of the independent variable given the index
**/
string ODESystem::getIVariableName(const int index) const {
	if (index < 0 || index >= mIVariableNameList.size()) {
		return "";
	} else {
		return mIVariableNameList[index];
	}
}

/** Returns the list of independent variable names
**/
vector<string> ODESystem::getIVariableNameList() const {
	return mIVariableNameList;
}

/** Sets the values of the dependent variables in the system
**/
void ODESystem::setDVariable(const vector<double>& variableList) {

	// Checks the size of the variable list
	if (variableList.size() != mDVariableList.size()) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::setDVariable(1) - Incompatible variable list size" << endl;
		#endif	
		return;
	}
	
	// Replace all the values of the mDVariableList with the given list
	for (int i = 0; i < mDVariableList.size(); i++) {
		mDVariableList[i] = variableList[i];
	}

}

/** Sets the value of the given dependent variable to a given value
**/
void ODESystem::setDVariable(const string variable, double value) {

	// Gets the index of the variable
	int index = indexOfDVariable(variable);
	
	if (index == -1) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::setDVariable(2) - Variable " << variable << " not found" << endl;
		#endif	
		return;
	} else {
		setDVariable(index, value);
	}

}

/** Sets the value of the dependent variable given the index and its new value
**/
void ODESystem::setDVariable(const int index, double value) {
	// Checks the index
	if (index < 0 || index >= mDVariableList.size()) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::setDVariable(3) - Index out of bounds (" << index << ")" << endl;
		#endif	
		return;
	}
	
	mDVariableList[index] = value;
	return;
}

/** Returns the dependent variable given the variable name.
* If the variable is not found, a NaN is returned
**/
double ODESystem::getDVariable(const string variable) const {

	// Gets the index of the variable
	int index = indexOfDVariable(variable);
	
	if (index == -1) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::getDVariable(1) - Variable " << variable << " not found" << endl;
		#endif	
		return sqrt(-1);
	} else {
		return getDVariable(index);
	}

}

/** Returns the dependent variable given its index
* If the index is out of bounds, then a NaN is returned
**/
double ODESystem::getDVariable(const int index) const {

	// Checks the index if it is out of bounds
	if (index < 0 || index > mDVariableList.size()) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::getDVariable(2) - Index out of bounds (" << index << ")" << endl;
		#endif	
		return sqrt(-1);
	}
	
	return mDVariableList[index];

}

/** Returns all the dependent values into the given vector reference
* If the vector reference is not of the correct size, resize it
**/
void ODESystem::getDVariable(vector<double>& result) const {

	if (result.size() != mDVariableList.size()) {
		result.clear();
		for (int i = 0; i < mDVariableList.size(); i++) {
			result.push_back(mDVariableList[i]);
		}
		return;
	} else {
		for (int i = 0; i < mDVariableList.size(); i++) {
			result[i] = mDVariableList[i];
		}
	}

}

/** Given the dependent variable, locates the index in the variable name list
**/
int ODESystem::indexOfDVariable(const string variable) const {

	int index = -1;
	
	// Loops through the name list and returns the index
	for (int i = 0; i < mDVariableNameList.size(); i++) {
		if (mDVariableNameList[i] == variable) {
			index = i;
			break;
		}
	}
	
	return index;

}

/** Gets the name of the dependent variable given the index
**/
string ODESystem::getDVariableName(const int index) const {
	if (index < 0 || index >= mDVariableNameList.size()) {
		return "";
	} else {
		return mDVariableNameList[index];
	}
}

/** Returns the list of dependent variable names
**/
vector<string> ODESystem::getDVariableNameList() const {
	return mDVariableNameList;
}

/** Sets the time of the system to the given time
* It will also update all its independent variables
**/
void ODESystem::setTime(double time) {
	
	if (time < 0.0) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::setTime() - Negative time (" << time << ")" << endl;
		#endif	
		return;
	}
	
	mTime = time;
	
	// Update the independent variables
	
	return;
}

void ODESystem::incTime(double delta) {

	// Checks if the increment will lead to a negative time
	if (mTime - delta < 0.0) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::incTime() - Invalid increment" << endl;
		#endif	
		return;
	}
	
	mTime += delta;
	
	// Update the independent variables
	
	return;
}

/** Returns the current time of the system
**/
double ODESystem::getTime() {
	return mTime;
}

/** Adds an node to the system and associates it with a dependent variable
**/
bool ODESystem::addNode	(string variable, string expression) {
	
	// Creates a parser to parse the equation
	ODEParser parser;
	ODENode* node = parser.parse(expression);
	
	// Checks the status. If the node is 0, return false
	// otherwise if the index is -1, then delete the node and
	// returns a false. Otherwise delete the current entry in the node list
	// and adds the equation in
	if (node == 0) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::addNode(1) - Invalid expression " << expression << endl;
		#endif	
		return false;
	} else {
		if (addNode(variable, node)) {
			int index = indexOfDVariable(variable);
			mEquationList[index] = expression;
			return true;
		} else {
			delete node;
			return false;
		}
	}
}

/** Adds an node to the system and associates it with a dependent variable
**/
bool ODESystem::addNode(string variable, ODENode* node) {
	
	int index = indexOfDVariable(variable);
	
	if (index == -1) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::addNode(2) - Invalid variable " << variable << endl;
		#endif	
		return false;
	} 

	// Sets the context for the new node
	if (!node->setContext(this)) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::addNode(2) - Unable to set context" << endl;
		#endif	
		return false;
	}

	// Checks if a node already exists in the vector. If it does, deletes it, otherwise
	// just set the new node in place
	if (mNodeList[index] != 0) {
		delete mNodeList[index];
	}
	mNodeList[index] = node;
	return true;
	
}

/** Associates a function with an independent variable in the system
**/
bool ODESystem::addFunction(string variable, ODEFunction* function) {

	int index = indexOfIVariable(variable);
	
	if (index == -1) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::addFunction() - Invalid variable " << variable << endl;
		#endif	
		return false;
	}

	// Checks if a function already exists in the vector. If it does, deletes it, otherwise
	// just set the new function in place
	if (mFunctionList[index] != 0) {
		delete mFunctionList[index];
	}
	mFunctionList[index] = function;
	return true;
	
}

/** Deletes a node from the node list and sets the location to 0
**/
bool ODESystem::removeNode(string variable) {
	
	int index = indexOfDVariable(variable);
	if (index != -1) {
		// Deletes the node off the node list
		if (mNodeList[index] != 0) {
			delete mNodeList[index];
			mNodeList[index] = 0;
		}
		// Sets the equation list to empty string
		mEquationList[index] = "0";
		return true;
	} else {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::removeNode() - Invalid variable " << variable << endl;
		#endif	
		return false;
	}
}

/** Removes a function from the function list and sets the location to 0
**/
bool ODESystem::removeFunction(string variable) {
	
	int index = indexOfIVariable(variable);
	if (index != -1) {
		// Deletes the function off the function list
		if (mFunctionList[index] != 0) {
			delete mFunctionList[index];
			mFunctionList[index] = 0;
		}
		return true;
	} else {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODESystem::removeFunction() - Invalid variable " << variable << endl;
		#endif	
		return false;
	}
}

bool ODESystem::isConsistent() {
	return true;
}

/** Checks if the system has a given symbol
* It checks all three name list for the occurrence of such a symbol.
**/
bool ODESystem::hasSymbol(string symbol) {

	if (indexOfParameter(symbol) == -1 &&
	    indexOfIVariable(symbol) == -1 &&
		indexOfDVariable(symbol) == -1) {
		return false;
	} else {
		return true;
	}
}

/** Solve for the set of equation using the current state of the system
**/
bool ODESystem::f(vector<double>& result) {
	f(result, mDVariableList, mIVariableList, mParameterList);
	return true;
}

/** Solve for the set of equations with y as the input
**/
bool ODESystem::f(vector<double>& result, const vector<double>& y) {
	f(result, y, mIVariableList, mParameterList);
	return true;
}

/** Solve for the set of equations with y and u as the input
**/
bool ODESystem::f(vector<double>& result, const vector<double>& y, const vector<double>& u) {
	f(result, y, u, mParameterList);
	return true;
}
	
/** Solve for the set of equations using the vectors given
* There is no error checking done here
**/
bool ODESystem::f(vector<double>& result, const vector<double>& y, const vector<double>& u, const vector<double>& p) {

	// Checks if the size of the result vector is correct
	if (result.size() != mDVariableList.size()) {
		result.clear();
		for (int i = 0; i < mNodeList.size(); i++) {
			if (mNodeList[i] != 0) {
				result.push_back(mNodeList[i]->evaluate(y, u, p));
			} else {
				result.push_back(0.0);
			}
		}
	} else {
		for (int i = 0; i < mNodeList.size(); i++) {
			if (mNodeList[i] != 0) {
				result[i] = mNodeList[i]->evaluate(y, u, p);
			} else {
				result[i] = 0.0;
			}
		}
	}
	
	return true;
}

/** Computes the jacobian of the system using the current state of the 
* system.
**/
bool ODESystem::jacobian(Matrix& result) {
	jacobian(result, mDVariableList, mIVariableList, mParameterList);
	return true;
}

/** Computes the jacobian of the system using the given y, u, p vectors
**/
bool ODESystem::jacobian(Matrix& result, const vector<double>& y, const vector<double>& u, const vector<double>& p) {

	// Resize the result matrix if it is not of the correct size. Otherwise, clears it
	// to 0
	if (result.rows() != mDVariableList.size() || 
	    result.cols() != mDVariableList.size()) {
		result.resize(mDVariableList.size(), mDVariableList.size());
	} else {
		result.clear();
	}

	vector<double> f;
	vector<double> delta_f;
	vector<double> delta_y;
	double tmp;
	double h = ODESYSTEM_DELTA;
	
	// Gets the value for the original f
	this->f(f, y, u, p);
	
	// Puts the values of y into the delta_y vector
	for (int i = 0; i < y.size(); i++) {
		delta_y.push_back(y[i]);
	}
	
	// For each value of delta y, changes it by a little and re-evaluates
	for (int i = 0; i < delta_y.size(); i++) {
		tmp = delta_y[i];
		delta_y[i] = delta_y[i] + h;
		this->f(delta_f, delta_y, u, p);
		for (int j = 0; j < delta_f.size(); j++) {
			result(j, i) = (delta_f[j] - f[j]) / h;
		}
		delta_y[i] = tmp;
	}

	return true;
} 
	
/** Returns the number of parameters in the system
**/
int ODESystem::parameterSize() const {
	return mParameterNameList.size();
}

/** Returns the number of independent variables in the system
**/
int ODESystem::iVariableSize() const {
	return mIVariableNameList.size();
}

/** Returns the number of dependent variables in the system
**/
int ODESystem::dVariableSize() const {
	return mDVariableNameList.size();
}

/** Resets the system, setting its independent variables and dependent variables to 0
* and setting the time to 0. The parameters are not affected by this action
**/
void ODESystem::reset() {
	// Sets the time to 0
	mTime = 0;
	// Resets the independent variables
	getIVariableAt(mIVariableList, mTime);
	// Resets the dependent variables
	for (int i = 0;  i < mDVariableList.size(); i++) {
		mDVariableList[i] = 0.0;
	}
}

/** Returns a string representation of the system
**/
string ODESystem::toString() const {

	ostringstream str;
	
	str << "---ODESystem---" << endl;
	str << "mName : " << mName << endl;
	str << "mTime : " << mTime << endl;
	str << "List of parameters : " ;
	for (int i = 0; i < mParameterList.size(); i++) {
		str << "[" << mParameterNameList[i] << ", " << mParameterList[i] << "] ";
	}
	str << endl;
	str << "List of independent variables : ";
	for (int i = 0; i < mIVariableList.size(); i++) {
		str << "[" << mIVariableNameList[i] << ", " << mIVariableList[i] << "] ";
	}
	str << endl;
	str << "List of dependent variables : ";
	for (int i = 0; i < mDVariableList.size(); i++) {
		str << "[" << mDVariableNameList[i] << ", " << mDVariableList[i] << "] ";
	}
	str << endl;
	str << "List of ODEs : " << endl;
	for (int i = 0; i < mDVariableList.size(); i++) {
		str << mDVariableNameList[i] << "' = " << mEquationList[i] << endl;
	}

	return str.str();
}

/** Helper function for copying contents
* It copies everything but assumes that allocated memory (if any)
* has already been deleted. Note that the parent is not copied
**/
void ODESystem::copyFrom(const ODESystem& src) {

	mName = src.mName;
	mTime = src.mTime;
	
	// Copies all the parameter/variableList
	for (int i = 0; i < src.mParameterList.size(); i++) {
		mParameterList.push_back(src.mParameterList[i]);
		mParameterNameList.push_back(src.mParameterNameList[i]);
	}
	for (int i = 0; i < src.mIVariableList.size(); i++) {
		mIVariableList.push_back(src.mIVariableList[i]);
		mIVariableNameList.push_back(src.mIVariableNameList[i]);
		mFunctionList.push_back(0);
	}
	for (int i = 0; i < src.mDVariableList.size(); i++) {
		mDVariableList.push_back(src.mDVariableList[i]);
		mDVariableNameList.push_back(src.mDVariableNameList[i]);
		mNodeList.push_back(0);
	}
	
	// Copies the equation list and re-parses the expressions.
	string expression;
	for (int i = 0; i < src.mEquationList.size(); i++) {
		expression = src.mEquationList[i];
		mEquationList.push_back(expression);
		addNode(mDVariableNameList[i], expression);
	}
}

}
