/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: odesystem.h
**  
** Description:
** This is the implementation of a system of first order ordinary differential equations expressed
** in the explicit form. The implementation of the system of ODE is such that it is a stateful system,
** i.e. it keeps a record of the current values of all the variables and parameters. However, its state
** can also be kept external by maintaining 3 vectors (for parameters, dependent variables and independent
** variables) and also a t variable to represent time.
** The evaluation of the system can either use the internal state of the system, or passing in the necessary
** state vectors to be handled on the spot.
**
** Last modified: 
** 26 February 2007
** 
** Comments:
**1) There for the ode, the functions associated with dependent variables are known as nodes and the
**    ones that are associated with the independent variables are known as functions. This is because
**    the independent variables can take on a larger variety of values including from trace, splines, fix functions
**    and even constants.
****************************************************************************/

#ifndef ODESYSTEM_H_
#define ODESYSTEM_H_

/* For debugging. Comment out to suppress debugging lines */
#define ODESYSTEM_DEBUG

/* Definition of constants to be used */
#define ODESYSTEM_DELTA 1e-10

/* Standard Includes */
#include <iostream>
#include <vector>

/* Application Includes */
#include "odecommon.h"
#include "odenode.h"
#include "odeparser.h"
#include "matrix.h"

using namespace std;
using namespace matrix;

namespace ode {

/* Forward Declarations */
class ODENode;

class ODESystem {

public:

	/* Constructors and destructors */
	ODESystem(string name = "");
	ODESystem(const ODESystem& src);	
	ODESystem& operator=(const ODESystem& src);
	virtual ~ODESystem(); 

	/* Sets the name of the system */
	void setName(string name);
	string getName();
	
	/* Adding and removing parameters/variables */
	bool addParameter(const string parameter, double value = 0.0);
	bool removeParameter(const string parameter);
	bool removeParameter(const int index);
	bool addIVariable(const string variable, double value = 0.0);
	bool removeIVariable(const string variable);
	bool removeIVariable(const int index);
	bool addDVariable(const string variable, double value = 0.0);
	bool removeDVariable(const string variable);
	bool removeDVariable(const int index);
	
	/* Setting and obtaining the parameters and variables */
	void setParameter(const vector<double>& parameterList);
	void setParameter(const string parameter, double value);
	void setParameter(const int index, double value);
	double getParameter(const string parameter) const;
	double getParameter(const int index) const;
	void getParameter(vector<double>& result) const;
	int indexOfParameter(const string parameter) const;
	string getParameterName(const int index) const;
	vector<string> getParameterNameList() const;
	
	void setIVariable(const vector<double>& variableList);
	void setIVariable(const string variable, double value);
	void setIVariable(const int index, double value);
	double getIVariable(const string variable) const;
	double getIVariable(const int index) const;
	void getIVariable(vector<double>& result) const;
	void getIVariableAt(vector<double>& result, double time) const;
	int indexOfIVariable(const string variable) const;
	string getIVariableName(const int index) const;
	vector<string> getIVariableNameList() const;
	
	void setDVariable(const vector<double>& variableList);
	void setDVariable(const string variable, double value);
	void setDVariable(const int index, double value);
	double getDVariable(const string variable) const;
	double getDVariable(const int index) const;
	void getDVariable(vector<double>& result) const;
	int indexOfDVariable(const string variable) const;
	string getDVariableName(const int index) const;
	vector<string> getDVariableNameList() const;
	
	void setTime(double time);
	void incTime(double delta);
	double getTime();

	/* Set and remove the equations that defines the system */
	bool addNode(string variable, string expression);
	bool addNode(string variable, ODENode* node);
	bool addFunction(string variable, ODEFunction* function);
	bool removeNode(string variable);
	bool removeFunction(string variable);
	
	/* Checks the system for consistency */
	bool isConsistent();
	bool hasSymbol(string symbol);
	
	/* Calculate the derivative and jacobian of the dependent variables */
	bool f(vector<double>& result);
	bool f(vector<double>& result, const vector<double>& y);
	bool f(vector<double>& result, const vector<double>& y, const vector<double>& u);
	bool f(vector<double>& result, const vector<double>& y, const vector<double>& u, const vector<double>& p); 
	bool jacobian(Matrix& result);
	bool jacobian(Matrix& result, const vector<double>& y, const vector<double>& u, const vector<double>& p); 
	
	/* Printing and debugging */
	int parameterSize() const;
	int iVariableSize() const;
	int dVariableSize() const;
	void reset();
	string toString() const;
	
protected:

	string mName;									// Name of the system
	double mTime;									// The current time of the system
	vector<string> mParameterNameList;
	vector<double> mParameterList;
	vector<string> mIVariableNameList;
	vector<double> mIVariableList;
	vector<string> mDVariableNameList;
	vector<double> mDVariableList;
	vector<ODENode*> mNodeList;
	vector<string> mEquationList;
	vector<ODEFunction*> mFunctionList;
	
private:
	void copyFrom(const ODESystem& src);

};

}

#endif /* ODESYSTEM_H_ */
