/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: odetokenizer.cpp
**  
** Description:
** This is the implementation of an equation tokenizer for an ODE.
**
** Last modified: 
** 23 February 2007
** 
** Comments:
**
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <sstream>
#include <vector>

/* Application Includes */
#include "odetokenizer.h"

using namespace std;

namespace ode {

/** Default constructor
**/
ODETokenizer::ODETokenizer() {
	// Do nothing
}

/** Constructor with an initial expression to tokenize
**/
ODETokenizer::ODETokenizer(string expression) {
	// Tokenize the given expression
	tokenize(expression);
}

/** Copy constructor 
**/
ODETokenizer::ODETokenizer(const ODETokenizer& src) {
	copyFrom(src);
}

/** Assignment constructor
**/
ODETokenizer& ODETokenizer::operator=(const ODETokenizer& src) {

	// Checks for self assignment
	if (this == &src) {
		return (*this);	
	}

	// Deletes the buffer
	mBuffer.clear();
	
	// Copy contents
	copyFrom(src);
	
	return (*this);
}

/** Default Destructor
**/
ODETokenizer::~ODETokenizer() {
	mBuffer.clear();
}

/** Pops the next token off the vector
* If there are no more tokens, return an empty string
**/
string ODETokenizer::nextToken() {
	string token;
	if (mBuffer.size() > 0) {
		token = mBuffer[0];
		mBuffer.erase(mBuffer.begin());
	} else {
		#ifdef ODETOKENIZER_DEBUG
			cout << "[Error] ODETokenizer::nextToken() - No more tokens" << endl;
		#endif	
		token = "";
	}
	
	return token;
}

/** Sees what is the next token without removing it from the vector
**/
string ODETokenizer::peek() {
	string token;
	if (mBuffer.size() > 0) {
		token = mBuffer[0];
	} else {
		#ifdef ODETOKENIZER_DEBUG
			cout << "[Error] ODETokenizer::peek() - No more tokens" << endl;
		#endif	
		token = "";
	}
	
	return token;
}

/** Checks if the tokenizer has more tokens
**/
bool ODETokenizer::hasMoreTokens() {
	return (mBuffer.size() > 0);
}

/** Tokenize a mathematical expression and places the tokens into a buffer
* For all delimiters, they will be included in the expression except for white spaces.
* This tokenizer goes by a state system, i.e. it maintains its current state
* and acts according to the next character that it sees
* List of states:
* 0 : Start of new token
* 1 : Handling a identifier
* 2 : Handling a number
* 3 : Handling a decimal point
* 4 : Handling the exponent of a scientific notation before seeing the sign
* 5 : Handling the exponent of a scientific notation after seeing the sign
* -1 : Fail state
**/	
bool ODETokenizer::tokenize(string expression) {
	// Clear the current buffer
	mBuffer.clear();
	
	string token;
	char c;
	int index;
	int state;
	bool seeExp;	// For testing to see if an exponential has been hit when parsing a number
	
	// Sets the index to 0
	index = 0;
		
	// Sets the state to 0
	state = 0;
	
	while (index < expression.length()) {
		
		// Gets the next character
		c = expression[index];
		
		// Checks the state and acts according to the state
		switch (state) {
		case 0:		// Start state
			if (isalpha(c) || c == '_') {
				state = 1;
				token += c;
			} else if (isdigit(c)) {
				state = 2;
				token += c;
			} else if (c == '.') {
				state = -1;
			} else if (ispunct(c)) {
				token += c;
				mBuffer.push_back(token);
				token = "";
			} else {
				// Useless character. Drop it
			}
			break;
		case 1:		// Handling an identifier
			if (isalnum(c) || c == '_' || c == '.') {
				state = 1;
				token += c;
			} else if (ispunct(c)) {
				state = 0;
				mBuffer.push_back(token);
				token = "";
				index--;
			} else {
				state = 0;
				mBuffer.push_back(token);
				token = "";
			}
			break;
		case 2:		// Handling a number
			if (isdigit(c)) {
				state = 2;
				token += c;
			} else if (c == '.') {
				state = 3;
				token += c;
			} else if (c == 'e' || c == 'E') {
				state = 4;
				token += c;
			} else if (ispunct(c)) {
				state = 0;
				mBuffer.push_back(token);
				token = "";
				index--;
			} else if (isalpha(c)) {
				state = -1;
				mBuffer.push_back(token);
				token = "";
			} else {
				state = 0;
				mBuffer.push_back(token);
				token = "";
			}
			break;
		case 3:		// Handling the decimal portion
			if (isdigit(c)) {
				state = 3;
				token += c;
			} else if (c == '.') {
				state = -1;
				mBuffer.push_back(token);
				token = "";
			} else if (c == 'e' || c == 'E') {
				state = 4;
				token += c;
			} else if (ispunct(c)) {
				state = 0;
				mBuffer.push_back(token);
				token = "";
				index--;
			} else if (isalpha(c)) {
				state = -1;
				mBuffer.push_back(token);
				token = "";
			} else {
				state = 0;
				mBuffer.push_back(token);
				token = "";
			}
			break;
		case 4:		// Handling the exponent
			if (isdigit(c)) {
				state = 5;
				token += c;
			} else if (c == '+' || c == '-') {
				// Does a peek into the next character to ensure that
				// it is a number
				if (index < expression.length() - 1) {
					char c_next = expression[index + 1];
					if (isdigit(c_next)) {
						state = 5;
						token += c;
					} else {
						state = -1;
						mBuffer.push_back(token);
						token = "";
					}
				} else {
					state = -1;
					mBuffer.push_back(token);
					token = "";
				}				
			} else if (c == '.' || ispunct(c) || isalpha(c)) {
				state = -1;
				mBuffer.push_back(token);
				token = "";
			} else {
				state = 0;
				mBuffer.push_back(token);
				token = "";
			}
			break;
		case 5:
			if (isdigit(c)) {
				state = 5;
				token += c;
			} else if (c == '.') {
				state = -1;
				mBuffer.push_back(token);
				token = "";
			} else if (ispunct(c)) {
				state = 0;
				mBuffer.push_back(token);
				token = "";
				index--;
			} else if (isalpha(c)) {
				state = -1;
				mBuffer.push_back(token);
				token = "";
			} else {
				state = 0;
				mBuffer.push_back(token);
				token = "";
			}
			break;
		case -1:
			#ifdef ODETOKENIZER_DEBUG
				cout << "[Error] ODETokenizer::tokenize() - Error in expression" << endl;
			#endif	
			return false;
		default:
			// Do nothing
			break;
		}

		index++;
	}
	
	// Checks the token. If it is not nothing then adds it to the buffer
	if (token != "") mBuffer.push_back(token);
	
	return true;
}

/** Returns a string representation of the contents of the buffer 
**/	
string ODETokenizer::toString() const {
	ostringstream str;
	
	str << "---ODETokenizer---" << endl;
	str << "mBuffer : ";
	str << "[";
	if (mBuffer.size() != 0) {
		str << mBuffer[0];
	}
	for (int i = 1; i < mBuffer.size(); i++) {
		str << ", " << mBuffer[i];
	}
	str << "]";
	str << endl;
	
	return str.str();
}
	
/** Helper function for copying contents
* It copies everything but assumes that allocated memory (if any)
* has already been deleted. Note that the parent is not copied
**/
void ODETokenizer::copyFrom(const ODETokenizer& src) {
	for (int i = 0; i < src.mBuffer.size(); i++) {
		mBuffer.push_back(src.mBuffer[i]);
	}
}

}	
