/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: odetrace.cpp
**  
** Description:
** This is a wrapper for a 2D array representing the trace results from the simulations
**
** Last modified: 
** 03 April 2007
** 
** Comments:
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>
#include <math.h>

/* Application Includes */
#include "odetrace.h"
#include "odetokenizer.h"

using namespace std;

namespace ode {

/** Default constructor
**/
ODETrace::ODETrace() : mInitialized(false) {

	#ifdef ODESYSTEM_DEBUG
		cout << "[Info] ODETrace::ODETrace() - Creating ODETrace" << endl;
	#endif
	
}

/** Copy constructor
**/
ODETrace::ODETrace(const ODETrace& src) {
	copyFrom(src);
}

/** Assignment constructor
**/
ODETrace& ODETrace::operator=(const ODETrace& src) {

	// Checks for self assignment
	if (this == &src) {
		return (*this);	
	}
	
	// Re-initializes the member variables
	mInitialized = false;
	
	// Clears all the vectors
	mTimeList.clear();
	mTraceList.clear();
	mVariableNameList.clear();
		
	// Copy contents
	copyFrom(src);
	
	return (*this);
}

/** Default destructor
**/
ODETrace::~ODETrace() {

	#ifdef ODESYSTEM_DEBUG
		cout << "[Info] ODETrace::~ODETrace() - Destroying ODETrace " << endl;
	#endif

	// Clears all the vectors
	mTimeList.clear();
	mTraceList.clear();
	mVariableNameList.clear();
}

/** Adds a variable to the variable list
* No checking is done other than that no value has been added in yet, i.e. mInitialized = false
**/
bool ODETrace::addVariable(string variable) {
	if (!mInitialized) {
		vector<double> newTraceList;
		mTraceList.push_back(newTraceList);
		mVariableNameList.push_back(variable);
		return true;
	} else {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODETrace::addVariable(1) - Trace object already has already been initialized. Cannot add any more variables" << endl;
		#endif
		return false;
	}
}

/** Adds a list of variables to the trace object
* No checking is done other than that no value has been added in yet, i.e. mInitialized = false
**/
bool ODETrace::addVariableList(const vector<string>& variableNameList) {
	if (!mInitialized) {
		for (int i = 0; i < variableNameList.size(); i++) {
			vector<double> newTraceList;
			mTraceList.push_back(newTraceList);
			mVariableNameList.push_back(variableNameList[i]);
		}
		return true;
	} else {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODETrace::addVariable(2) - Trace object already has already been initialized. Cannot add any more variables" << endl;
		#endif
		return false;
	}
}

/** Gets the list of time points
**/
const vector<double>& ODETrace::getTimeList() const {
	return mTimeList;
}

/** Gets the trace of the particular variable
* If the variable is not found, returns an empty trace
**/
const vector<double>& ODETrace::getTrace(string variable) const {
	int index = indexOf(variable);
	if (index != -1) {
		return mTraceList[index];
	} else {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODETrace::getTrace(1) - Variable " << variable << " not found" << endl;
		#endif
		return mEmptyList;
	}
}

/** Gets the trace given the index
* If the index is out of bounds, returns an empty trace
**/
const vector<double>& ODETrace::getTrace(int index) const {
	if (index >= 0 && index < mTraceList.size()) {
		return mTraceList[index];
	} else {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODETrace::getTrace(1) - Variable " << variable << " not found" << endl;
		#endif
		return mEmptyList;
	}
}

/** Adds a trace to the trace.
* It only checks for the size of the input trace
**/
bool ODETrace::addTraceAt(double time, const vector<double>& trace) {
	
	if (trace.size() == mTraceList.size()) {
		// Sets the initialized flag to true
		mInitialized = true;
		
		// Pushes the time and values to the last slot in their respective traces
		mTimeList.push_back(time);
		for (int i = 0; i < trace.size(); i++) {
			mTraceList[i].push_back(trace[i]);
		}
		
		return true;
	} else {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODETrace::addTraceAt() - Trace list has incompatible sizes (" << mTraceList.size() << ":" << trace.size() << ")" << endl;
		#endif
		return false;
	}
}

/** Gets the value of a variable given the index to the time
* Returns NaN if any of the indices are out of bounds
**/
double ODETrace::getValueAt(string variable, int timeIndex) const {
	int varIndex = indexOf(variable);
	
	if (varIndex == -1) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODETrace::getValueAt(1) - Variable " << variable << " not found" << endl;
		#endif
		return sqrt(-1);
	} else {
		return getValueAt(varIndex, timeIndex);
	}
}

/** Gets the value of a variable given the index to the time and the variable itself
* Returns NaN if any of the indices are out of bounds
**/
double ODETrace::getValueAt(int varIndex, int timeIndex) const {
	if (varIndex < 0 || varIndex >= mVariableNameList.size() || 
	    timeIndex < 0 || timeIndex >= mTimeList.size()) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODETrace::getValueAt(2) - Index out of bounds" << endl;
		#endif
		return sqrt(-1);
	} else {
		return mTraceList[varIndex][timeIndex];
	}
}

/** Sets the value of a variable given the index to the time
* If out of bounds then nothing is done
**/	
void ODETrace::setValueAt(string variable, int timeIndex, double value) {
	int varIndex = indexOf(variable);
	
	if (varIndex == -1) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODETrace::setValueAt(1) - Variable " << variable << " not found" << endl;
		#endif
		// Do nothing
	} else {
		setValueAt(varIndex, timeIndex, value);
	}

}

/** Sets the value of a variable given the index to the time and the variable itself
* If out of bounds then nothing is done
**/
void ODETrace::setValueAt(int varIndex, int timeIndex, double value) {
	if (varIndex < 0 || varIndex >= mVariableNameList.size() || 
	    timeIndex < 0 || timeIndex >= mTimeList.size()) {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODETrace::setValueAt(2) - Index out of bounds" << endl;
		#endif
	} else {
		mTraceList[varIndex][timeIndex] = value;
	}

}

/** Writes the contents of the trace object onto a file
**/
bool ODETrace::save(const string filename, string delimiter) const {

	// Opens the file to write
	fstream outputStream;
	outputStream.open(filename.c_str(), ios::out);
	
	if (outputStream) {
		
		// Prints out the names
		outputStream << "Time";
		for (int i = 0; i < mVariableNameList.size(); i++) {
			outputStream << delimiter << mVariableNameList[i];
		}
		
		outputStream << endl;
		
		for (int i = 0; i < mTimeList.size(); i++) {
			outputStream << mTimeList[i];
			for (int j = 0; j < mTraceList.size(); j++) {	
				outputStream << delimiter << mTraceList[j][i];
			}
			outputStream << endl;
		}

		outputStream.close();
		return true;
	} else {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODETrace::save() - Error opening file " << filename << endl;
		#endif
		return false;
	}
}

/** Loads the contents of the file into the trace
**/
bool ODETrace::load(const string filename, string delimiter) {

	// Opens the file to write
	fstream inputStream;
	inputStream.open(filename.c_str(), ios::in);
	
	if (inputStream) {
	
		// Clears all the vectors
		mTimeList.clear();
		mTraceList.clear();
		mVariableNameList.clear();
		mInitialized = false;
		
		string line;
		string token;
		double time;
		double value;
		vector<double> trace;
		ODETokenizer tokenizer;
		
		// Reads the first line (header)
		if (!inputStream.eof()) {
			getline(inputStream, line, '\n');
		} else {
			#ifdef ODESYSTEM_DEBUG
				cout << "[Error] ODETrace::load() - Empty file" << endl;
			#endif
			inputStream.close();
			return false;
		}
		
		// Tokenize the line
		tokenizer.tokenize(line);
		
		// Discards the first one (should be time)
		token = tokenizer.nextToken();
		
		while (tokenizer.hasMoreTokens()) {
			addVariable(tokenizer.nextToken());
		}
		
		// Loops and reads the trace
		while (!inputStream.eof()) {
			getline(inputStream, line, '\n');
			tokenizer.tokenize(line);
			
			if (tokenizer.hasMoreTokens()) {
				token = tokenizer.nextToken();
				time = atof(token.c_str());
			}
			
			trace.clear();
			while (tokenizer.hasMoreTokens()) {
				token = tokenizer.nextToken();
				value = atof(token.c_str());
				trace.push_back(value);
			}
			
			// Adds the trace
			addTraceAt(time, trace);
		}

		// Closes the stream
		inputStream.close();
		
	} else {
		#ifdef ODESYSTEM_DEBUG
			cout << "[Error] ODETrace::load() - Error opening file " << filename << endl;
		#endif
		return false;
	}
	
	return true;

}

/** Gives the index of the given variable
* If the variable cannot be found, returns a -1
**/
int ODETrace::indexOf(string variable) const {
	int index = -1;

	// Loops through the variable list and returns the index
	for (int i = 0; i < mVariableNameList.size(); i++) {
		if (mVariableNameList[i] == variable) {
			index = i;
			break;
		}
	}
	
	return index;
}

/** Gives the index of the time 
**/
int ODETrace::timeIndexOf(double time) const {
	int index = -1;
	
	// Loops through the time list and returns the index
	for (int i = 0; i < mTimeList.size(); i++) {
		if (mTimeList[i] == time) {
			index = i;
			break;
		}
	}
	
	return index;
}

/** Returns the number of time points
**/
int ODETrace::stepSize() const {
	return mTimeList.size();
}

/** Returns the number of variables
**/
int ODETrace::variableSize() const {
	return mVariableNameList.size();
}

/** Helper function for copying contents
* It copies everything but assumes that allocated memory (if any)
* has already been deleted.
**/
void ODETrace::copyFrom(const ODETrace& src) {

	mInitialized = src.mInitialized;

	// Copies the time trace
	for (int i = 0; i < src.mTimeList.size(); i++) {
		mTimeList.push_back(src.mTimeList[i]);
	}
	
	// Copies the variables and trace
	for (int i = 0; i < src.mVariableNameList.size(); i++) {
		mVariableNameList.push_back(src.mVariableNameList[i]);
		vector<double> newTraceList;
		for (int j = 0; j < src.mTraceList[i].size(); j++) {
			newTraceList.push_back(src.mTraceList[i][j]);
		}
		mTraceList.push_back(newTraceList);
	}
	
}

}

