/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: random.cpp
**  
** Description:
** The header file for a 0uniform random number generator. The different
** generators are implemented as singletons.
**  
** Last modified: 
** 26 March 2007
** 
** Comments:
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <math.h>

/* APplication Includes */
#include "random.h"

namespace random {

/** Initialize the random number generator with a seed
**/
void MersenneTwister::init(unsigned long seed) {

	getInstance().stateArray[0] = seed & 0xffffffffUL;
	for (int i = 1; i < N; ++i) {
		getInstance().stateArray[i] = (1812433253UL * (getInstance().stateArray[i - 1] ^ (getInstance().stateArray[i - 1] >> 30)) + i)  & 0xffffffffUL;           /* for >32 bit machines */
	}
	
	getInstance().p0 = getInstance().stateArray;
	getInstance().p1 = getInstance().stateArray + 1;
	getInstance().pm = getInstance().stateArray + M;
}

/** Initialize the random number generator with a seed array
**/
void MersenneTwister::initArray(unsigned long seedArray[], int length) {

	int i, j;

	init(19650218UL);
	i = 1;
	j = 0;
  
	for (int k = (N > length ? N : length); k; --k) {
		// Non linear
		getInstance().stateArray[i] = ((getInstance().stateArray[i] ^ ((getInstance().stateArray[i - 1] ^ (getInstance().stateArray[i - 1] >> 30)) * 1664525UL)) + seedArray[j] + j) & 0xffffffffUL;
		if (++i >= N) {
			getInstance().stateArray[0] = getInstance().stateArray[N - 1];
			i = 1;
		}
		if (++j >= length) {
			j = 0;
		}
	}
	
	for (int k = N - 1; k; --k) {
		// Non linear
		getInstance().stateArray[i] = ((getInstance().stateArray[i] ^ ((getInstance().stateArray[i - 1] ^ (getInstance().stateArray[i - 1] >> 30)) * 1566083941UL)) - i) & 0xffffffffUL;
		if (++i >= N) {
			getInstance().stateArray[0] = getInstance().stateArray[N - 1];
			i = 1;
		}
	}
	
	// MSB is 1; assuring non-zero initial array
	getInstance().stateArray[0] = 0x80000000UL;          
}

/** Generates a random number on the interval [0, 0xffffffff]
**/
unsigned long MersenneTwister::randInt32() {
 
	unsigned long y;

	if (!getInstance().p0) {
		// Default seed 
		getInstance().init(5489UL);
	}
	
	// Twisted feedback
	y = *getInstance().p0 = *getInstance().pm++ ^ (((*getInstance().p0 & UPPER_MASK) | (*getInstance().p1 & LOWER_MASK)) >> 1) ^ (-(*getInstance().p1 & 1) & MATRIX_A);
	getInstance().p0 = getInstance().p1++;
  
	if (getInstance().pm == getInstance().stateArray + N) {
		getInstance().pm = getInstance().stateArray;
	}
	
	if (getInstance().p1 == getInstance().stateArray + N) {
		getInstance().p1 = getInstance().stateArray;
	}
	
	// Temper
	y ^= y >> 11;
	y ^= y << 7 & 0x9d2c5680UL;
	y ^= y << 15 & 0xefc60000UL;
	y ^= y >> 18;

	return y;
}

/** Generates a random number on the interval [0,0x7fffffff] 
**/
long MersenneTwister::randInt31() {
	return (long) (randInt32() >> 1);
}

/** Generates a random number on the read close interval [0, 1]
**/
double MersenneTwister::randRealClose() {
	// Divide by 2^32-1
	return randInt32() * (1.0 / 4294967295.0);
}

/** Generates a random number on the real semi-open interval [0, 1)
**/
double MersenneTwister::randRealHalf() {
	// Divide by 2^32
	return randInt32() * (1.0 / 4294967296.0);
}

/** Generates a random number on the open interval (0, 1)
**/
double MersenneTwister::randRealOpen() {
	// Divide by 2^32
	return (((double) randInt32()) + 0.5) * (1.0 / 4294967296.0);
}

/** Generates a 53-bit random number on the real interval [0,1) 
**/
double MersenneTwister::randRes53() {
	unsigned long a = randInt32() >> 5, b = randInt32() >> 6;
	return (a * 67108864.0 + b) * (1.0 / 9007199254740992.0);
}

/** Default constructor (protected)
**/
MersenneTwister::MersenneTwister() {
	// Do nothing
}

/** Copy constructor (protected)
**/
MersenneTwister::MersenneTwister(const MersenneTwister&) {
	// Do nothing
}

/** Assignment constructor (protected)
**/
MersenneTwister& MersenneTwister::operator=(const MersenneTwister&) {
	return (*this);
}

/** Default destructor (protected)
**/
MersenneTwister::~MersenneTwister() {
	// Do nothing
}

/** Initialize the random number generator with a seed
* This method actually initializes the Mersenne Twister as the generation is based
* on a uniform random number generator. Hence if the Mersenne Twister is already
* initialized, then it is no longer necessary to call this method.
**/
void Gaussian::init(unsigned long seed) {
	// Initializes the Mersenne twister
	MersenneTwister::init(seed);
}

/** Returns a random number from the distibution N(0, 1)
**/
double Gaussian::rand() {

	double x1, x2, w;

	if (getInstance().gaussianRegen) {
		do {
		    x1 = 2.0 * MersenneTwister::randInt32() / (4294967295.0 + 1.0) - 1.0;
			x2 = 2.0 * MersenneTwister::randInt32() / (4294967295.0 + 1.0) - 1.0;
		    w = x1 * x1 + x2 * x2;		
		} while ( w >= 1.0 );
		
	 	w = sqrt((-2.0 * log(w)) / w);
	    getInstance().gauss1 = x1 * w;
	    getInstance().gauss2 = x2 * w;

		getInstance().gaussianRegen = false;
		return getInstance().gauss1;	 	
	} else {
		getInstance().gaussianRegen = true;
		return getInstance().gauss2;
	}
	
}

/** Default constructor (protected)
**/
Gaussian::Gaussian() : gaussianRegen(true) {
	// Do nothing
}

/** Copy constructor (protected)
**/
Gaussian::Gaussian(const Gaussian&) {
	// Do nothing
}

/** Assignment constructor (protected)
**/
Gaussian& Gaussian::operator=(const Gaussian&) {
	return (*this);
}

/** Default destructor (protected)
**/
Gaussian::~Gaussian() {
	// Do nothing
}


}
