/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: random.h
**  
** Description:
** The header file for a uniform random number generator. The different
** generators are implemented as singletons.
**  
** Last modified: 
** 26 March 2007
** 
** Comments:
** 1) Reference for Mersenne Twister 
**     M. Matsumoto and T. Nishimura, "Mersenne Twister:
**     A 623-Dimensionally Equidistributed Uniform Pseudo-Random Number
**     Generator", ACM Transactions on Modeling and Computer Simulation,
**     Vol. 8, No. 1, January 1998, pp 3--30.
** 2) Other random number will be implemented later, i.e gaussian etc.
****************************************************************************/

#ifndef RANDOM_H_
#define RANDOM_H_

/* Period parameters */
#define N 624
#define M 397
#define MATRIX_A 0x9908b0dfUL   
#define UPPER_MASK 0x80000000UL 	// most significant w-r bits
#define LOWER_MASK 0x7fffffffUL 	// least significant r bits

/* For debugging. Comment out to suppress debugging lines */
#define RANDOM_DEBUG

/* Standard Includes */
#include <iostream>

using namespace std;

namespace random {

class MersenneTwister {

public:
	
	/* Singleton accessor */
	static MersenneTwister& getInstance() {
		static MersenneTwister twister;
		return twister;
	}
	
	/* Initializer */
	static void init(unsigned long seed);
	static void initArray(unsigned long seedArray[], int length);
	
	/* Number generator */
	static unsigned long randInt32();
	static long randInt31();
	static double randRealClose();
	static double randRealOpen();
	static double randRealHalf();
	static double randRes53();
	
protected:

	unsigned long stateArray[N];      // Array for the states
	unsigned long *p0, *p1, *pm;	  // Pointers to the state arrays

	MersenneTwister();
	MersenneTwister(const MersenneTwister&);
	MersenneTwister& operator=(const MersenneTwister&);
	virtual ~MersenneTwister();

private:

	static MersenneTwister* mInstance;

};

class Gaussian {

public:
	
	/* Singleton accessor */
	static Gaussian& getInstance() {
		static Gaussian gaussian;
		return gaussian;
	}
	
	/* Initializer */
	static void init(unsigned long seed);
	
	/* Number generator */
	static double rand();
	
protected:

	bool gaussianRegen;
	double gauss1, gauss2;

	Gaussian();
	Gaussian(const Gaussian&);
	Gaussian& operator=(const Gaussian&);
	virtual ~Gaussian();

private:

	static Gaussian* mInstance;

};

}

#endif /* RANDOM_H_ */
