/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: common.cpp
**  
** Description:
** Provides the commonly used functionalities
**
** Last modified: 
** 12 March 2009
** 
** Comments:
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <vector>

/* Application Includes */
#include "common.h"
#include "ode.h"

using namespace ode;

vector<datapoint> getDataPoints(const vector<double>& timeList, 
							    const vector<string>& variableList, 
								const vector<double>& weightList,
								const ODETrace& trace) {

	vector<datapoint> pointList;
	
	int timeIndex;
	int variableIndex;
	double weight;
	
	for (int i = 0; i < timeList.size(); i++) {
	
		timeIndex = trace.timeIndexOf(timeList[i]);
		
		if (timeIndex != -1) {
			for (int j = 0; j < variableList.size(); j++) {
			
				variableIndex = trace.indexOf(variableList[j]);
				
				if (j >= 0 && j < weightList.size()) {
					weight = weightList[j];
				} else {
					weight = 1.0;
				}
				
				if (variableIndex != -1) {
				
					// Creates a data point
					datapoint point;
					point.timeIndex = timeIndex;
					point.variableIndex = variableIndex;
					point.weight = weight;
					point.value = trace.getValueAt(variableIndex, timeIndex);
					
					pointList.push_back(point);
					
				} else {
					cout << "Variable " << variableList[j] << " not found in the trace" << endl;
				}
				
			}
		} else {
			cout << "Time point " << timeList[i] << " not found in the trace" << endl;
		}
		
	}
	
	return pointList;
}

// Scores based on the data points
double score(const vector<datapoint>& pointList, const ODETrace& trace) {

	double score = 0.0;
	double value;
	
	// For each data point, if it is within range, i.e. corresponding value found, then the difference is
	// added onto the score variable. Otherwise it will be ignored
	for (int i = 0; i < pointList.size(); i++) {
		value = trace.getValueAt(pointList[i].variableIndex, pointList[i].timeIndex);
		if (!isnan(value)) {
			score += pow(value - pointList[i].value, 2) / pow(pointList[i].weight, 2);
		}
	}

	// Square root the score
	score = sqrt(score);
	
	return score;
}

// Converts int to string
string itos(int i) {
	stringstream s;
	s << i;
	return s.str();
}

// Sets the timeTaken to be the current time
void timeStart() {
	duration = time(NULL);
}

// Stops the timer and use time as a duration
void timeStop() {
	duration = time(NULL) - duration;
}

// Returns a string representation of the time taken
string timeTaken() {

	long tmpTime = duration;
	int hours = tmpTime / 3600;
	tmpTime = tmpTime % 3600;
	int min = tmpTime / 60;
	tmpTime = tmpTime % 60;
	int sec = tmpTime;
	
	// Formulate the string representation
	string durationStr = "";
	
	if (hours != 0) {
		durationStr = itos(hours) + " hrs ";
		durationStr = durationStr + itos(min) + " min ";
	} else if (min != 0) {
		durationStr = itos(min) + " min ";
	}
	
	durationStr = durationStr + itos(sec) + " sec";
		
	return durationStr;
	
	
}

