/****************************************************************************
** Copyright (C) 2006 Koh Yeow Nam, Geoffrey
** All rights reserved
**
** This file may be used under the terms of the GNU General Public
** License version 2.0 as published by the Free Software Foundation
** and appearing in the file LICENSE.GPL included in the packaging of
** this file.  
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** Filename: model1.cpp
**  
** Description:
**
** Last modified: 
** 24 April 2007
** 
** Comments:
****************************************************************************/

/* Standard Includes */
#include <iostream>
#include <vector>

/* Application Includes */
#include "common.h"
#include "ode.h"
#include "beliefpropagation.h"
#include "random.h"
#include "model1.h"

using namespace std;
using namespace ode;
using namespace factorgraph;
using namespace propagator;
using namespace random;
using namespace domain;

namespace model1 {

// Creates the ODE system of the split pathway
ODESystem* createModel() {

	// ================
	// Creates the system
	// ================
	ODESystem* system = new ODESystem("model1");
	
	// Adds in the parameters
	system->addParameter("k1");
	system->addParameter("k2");
	system->addParameter("k3");
	system->addParameter("k4");
	system->addParameter("k5");
	system->addParameter("k6");
	system->addParameter("k7");
	system->addParameter("k8");
	system->addParameter("k9");
	system->addParameter("k10");
	system->addParameter("k11");
	system->addParameter("k12");
		
	// Adds in the dependent variables
	system->addDVariable("x1");
	system->addDVariable("x2");
	system->addDVariable("x3");
	system->addDVariable("x4");
	system->addDVariable("x5");
	system->addDVariable("x6");
	system->addDVariable("x7");
	system->addDVariable("x8");
	system->addDVariable("x9");
	system->addDVariable("x10");
	system->addDVariable("x11");
				
	// Adds in the independent variables
	
	// Adds in the equations
	system->addNode("x1", "k1 - (k2 * x1)");
	system->addNode("x2", "(k4 * x3) - (k3 * x2 * x1)");
	system->addNode("x3", "(k3 * x2 * x1) - (k4 * x3)");
	system->addNode("x4", "(k6 * x5) - (k5 * x4 * x3)");
	system->addNode("x5", "(k5 * x4 * x3) - (k6 * x5)");
	system->addNode("x6", "(k8 * x7) - (k7 * x6 * x5)");
	system->addNode("x7", "(k7 * x6 * x5) - (k8 * x7)");
	system->addNode("x8", "(k10 * x9) - (k9 * x8 * x3)");
	system->addNode("x9", "(k9 * x8 * x3) - (k10 * x9)");
	system->addNode("x10", "(k12 * x11) - (k11 * x10 * x9)");
	system->addNode("x11", "(k11 * x10 * x9) - (k12 * x11)");
		
	return system;

}

// Creates the simulation trace using the parameters and the initial concentration
ODETrace* createTrace(const vector<double>& p, const vector<double>& y, ODESystem& model) {

	// Checks the size of the parameters and initial y concentration
	if (p.size() != model.parameterSize() ||
	    y.size() != model.dVariableSize()) {
		cout << "split::createTrace() : Uncompatible parameter or initial concentration vector" << endl;
		return 0;
	}
	
	ODETrace* trace;
	
	// Creates the solver
	// ODESolverImpEuler solver;
	ODESolverExpEuler solver;

	// Loads the model and its parameters
	model.setParameter(p);
	solver.loadODESystem(&model);

	// Sets solver settings
	solver.setAttribute("start", 0.0);
	solver.setAttribute("stop", 10.0);
	solver.setAttribute("step", 0.1);
	solver.setAttribute("logstep", 0.5);
	
	// Simulates using the initial values
	trace = solver.solve(y);
	
	return trace;
}

// Creates the nominal trace using the parameters and variables and adding noise to it
ODETrace* createNominalTrace() {

	vector<double> p = getParameters();
	vector<double> y = getVariables();
	ODESystem* model = createModel();

	ODETrace* trace;
	
	// Creates the solver
	ODESolverExpEuler solver;
	// ODESolverImpEuler solver;

	// Loads the model and its parameters
	model->setParameter(p);
	solver.loadODESystem(model);

	// Sets solver settings
	solver.setAttribute("start", 0.0);
	solver.setAttribute("stop", 10.0);
	solver.setAttribute("step", 0.1);
	solver.setAttribute("logstep", 0.5);
	
	// Simulates using the initial values
	trace = solver.solve(y);
	
	// Apply Gaussian noise
	double value, tmpValue;
	for (int i = 0; i < trace->variableSize(); i++) {
		for (int j = 0; j < trace->stepSize(); j++) {
			value = trace->getValueAt(i, j);
			do {
				tmpValue = value + (Gaussian::rand() * 0.00);
			} while (tmpValue < 0.0);
			trace->setValueAt(i, j, tmpValue);
		}
	}
	
	// Save the trace
	trace->save("trace.model1.txt");
	
	delete trace;
	delete model;

}

// Creates a factor graph equivalent of the split pathway
FactorGraph* createFactorGraph() {

	FactorGraph* graph = new FactorGraph();
	
	// Adds in the variable nodes (which in this case are the parameters
	graph->addNode("k1", "k1", nodeType_variable);
	graph->addNode("k2", "k2", nodeType_variable);
	graph->addNode("k3", "k3", nodeType_variable);
	graph->addNode("k4", "k4", nodeType_variable);
	graph->addNode("k5", "k5", nodeType_variable);
	graph->addNode("k6", "k6", nodeType_variable);
	graph->addNode("k7", "k7", nodeType_variable);
	graph->addNode("k8", "k8", nodeType_variable);
	graph->addNode("k9", "k9", nodeType_variable);
	graph->addNode("k10", "k10", nodeType_variable);
	graph->addNode("k11", "k11", nodeType_variable);
	graph->addNode("k12", "k12", nodeType_variable);
		
	// Adds in the factor nodes
	graph->addNode("f1_1", "f1_1", nodeType_factor);
	graph->addNode("f1_2", "f1_2", nodeType_factor);
	graph->addNode("f2_1", "f2_1", nodeType_factor);
	graph->addNode("f2_2", "f2_2", nodeType_factor);
	graph->addNode("f3_1", "f3_1", nodeType_factor);
	graph->addNode("f3_2", "f3_2", nodeType_factor);
	graph->addNode("f4", "f4", nodeType_factor);
	graph->addNode("f5", "f5", nodeType_factor);
	graph->addNode("f6", "f6", nodeType_factor);
	graph->addNode("f7", "f7", nodeType_factor);
	graph->addNode("f8", "f8", nodeType_factor);
	graph->addNode("f9", "f9", nodeType_factor);
	graph->addNode("f10", "f10", nodeType_factor);
	graph->addNode("f11", "f11", nodeType_factor);
	
	// Connects the nodes
	graph->connect("f1_1", "k1");
	graph->connect("f1_1", "k2");
	graph->connect("f1_1", "k3");
	
	graph->connect("f1_2", "k1");
	graph->connect("f1_2", "k2");
	graph->connect("f1_2", "k3");
	
	graph->connect("f2_1", "k3");
	graph->connect("f2_1", "k4");
	
	graph->connect("f2_2", "k3");
	graph->connect("f2_2", "k4");
	
	graph->connect("f3_1", "k3");
	graph->connect("f3_1", "k4");
	graph->connect("f3_1", "k5");
	
	graph->connect("f3_2", "k3");
	graph->connect("f3_2", "k4");
	graph->connect("f3_2", "k9");
	
	graph->connect("f4", "k5");
	graph->connect("f4", "k6");
	
	graph->connect("f5", "k5");
	graph->connect("f5", "k6");
	graph->connect("f5", "k7");
	
	graph->connect("f6", "k7");
	graph->connect("f6", "k8");
	
	graph->connect("f7", "k7");
	graph->connect("f7", "k8");
	
	graph->connect("f8", "k9");
	graph->connect("f8", "k10");
	
	graph->connect("f9", "k9");
	graph->connect("f9", "k10");
	graph->connect("f9", "k11");
	
	graph->connect("f10", "k11");
	graph->connect("f10", "k12");
	
	graph->connect("f11", "k11");
	graph->connect("f11", "k12");
	
	return graph;
}

// Returns a vector for the nominal parameters
vector<double> getParameters() {
	vector<double> p;
	p.resize(12);
	p[0] = 0.625;	// k1 
	p[1] = 0.228; 	// k2
	p[2] = 0.112;  	// k3
	p[3] = 0.96;   	// k4
	p[4] = 0.579; 	// k5
	p[5] = 0.312;	// k6
	p[6] = 0.628;	// k7
	p[7] = 0.104;	// k8
	p[8] = 0.04;	// k9
	p[9] = 0.286;	// k10
	p[10] = 0.624;	// k11
	p[11] = 0.88;	// k12
	
	return p;
}

// Returns a vector of initial concentration
vector<double> getVariables() {
	vector<double> y;
	y.resize(11);
	y[0] = 0.0;
	y[1] = 5.0;
	y[2] = 0.0;
	y[3] = 5.0;
	y[4] = 0.0;
	y[5] = 5.0;
	y[6] = 0.0;
	y[7] = 5.0;
	y[8] = 0.0;
	y[9] = 5.0;
	y[10] = 0.0;
	
	return y;
}

// Returns the set of weights
vector<double> getWeights() {
	vector<double> w;
	w.resize(11);
	w[0] = 2.0;
	w[1] = 5.0;
	w[2] = 5.0;
	w[3] = 5.0;
	w[4] = 5.0;
	w[5] = 5.0;
	w[6] = 5.0;
	w[7] = 5.0;
	w[8] = 5.0;
	w[9] = 5.0;
	w[10] = 5.0;
	return w;
}

// Creates and runs the model
void runModel() {

	// Create the system
	ODESystem* model = createModel();
	
	// Gets the parameters and initial concentration
	vector<double> p = getParameters();
	vector<double> y = getVariables();
	
	// Creates a trace and saves it.
	ODETrace* trace = createTrace(p, y, *model);

	// Checks that the trace is valid
	if (trace == 0) {
		cout << "[Error] model1::runModel - Error simulating the model..." << endl;
		delete model;
		return;
	}
	
	// Saves the trace
	trace->save("trace.model1.txt");
	delete trace;
	delete model;
	
	return;
	
}

// The actual belief propagation code
void runPropagate() {

	// Create the system
	ODESystem* model = createModel();
	
	// Gets the parameters and initial concentration
	vector<double> p = getParameters();
	vector<double> y = getVariables();
	
	// Loads the nominal trace and gets the data points.
	ODETrace* trace = new ODETrace();
	trace->load("trace.model1.txt");
	
	// Checks that the trace is valid
	if (trace == 0) {
		cout << "[Error] model1::runPropagate - Error getting the datapoints..." << endl;
		delete model;
		return;
	}

	// Gets the data points
	vector<double> weightList = getWeights();
	
	vector<double> timeList;
	timeList.push_back(0.5);
	timeList.push_back(1.0);
	timeList.push_back(1.5);
	timeList.push_back(2.0);
	timeList.push_back(2.5);
	timeList.push_back(3.0);
	timeList.push_back(3.5);
	timeList.push_back(4.0);
	timeList.push_back(4.5);
	timeList.push_back(5.0);
	timeList.push_back(5.5);
	timeList.push_back(6.0);
	timeList.push_back(6.5);
	timeList.push_back(7.0);
	timeList.push_back(7.5);
	timeList.push_back(8.0);
	timeList.push_back(8.5);
	timeList.push_back(9.0);
	timeList.push_back(9.5);
	timeList.push_back(10.0);
	
	// Gets the variable list. The ones for x1, x4, x5, x6, x7 are for
	// the individual list. C1 and C2 are component-wise variable list
	
	vector<string> varList_x1;
	varList_x1.push_back("x1");
	
	vector<string> varList_x5;
	varList_x5.push_back("x5");
	
	vector<string> varList_x7;
	varList_x7.push_back("x7");
	
	vector<string> varList_x9;
	varList_x9.push_back("x9");
	
	vector<string> varList_x11;
	varList_x11.push_back("x11");

	vector<string> varList_C1;
	varList_C1.push_back("x1");
	varList_C1.push_back("x5");
	varList_C1.push_back("x7");
	
	vector<string> varList_C2;
	varList_C2.push_back("x1");
	varList_C2.push_back("x9");
	varList_C2.push_back("x11");
	
	vector<datapoint> pointList_x1 = getDataPoints(timeList, varList_x1, weightList, *trace);
	vector<datapoint> pointList_x5 = getDataPoints(timeList, varList_x5, weightList, *trace);
	vector<datapoint> pointList_x7 = getDataPoints(timeList, varList_x7, weightList, *trace);
	vector<datapoint> pointList_x9 = getDataPoints(timeList, varList_x9, weightList, *trace);
	vector<datapoint> pointList_x11 = getDataPoints(timeList, varList_x11, weightList, *trace);
	
	vector<datapoint> pointList_C1 = getDataPoints(timeList, varList_C1, weightList, *trace);
	vector<datapoint> pointList_C2 = getDataPoints(timeList, varList_C2, weightList, *trace);

	cout << "Datapoint list size" << endl;
	cout << "pointList_x1 : " << pointList_x1.size() << endl;
	cout << "pointList_x5 : " << pointList_x5.size() << endl;
	cout << "pointList_x7 : " << pointList_x7.size() << endl;
	cout << "pointList_x9 : " << pointList_x9.size() << endl;
	cout << "pointList_x11 : " << pointList_x11.size() << endl;
	cout << "pointList_C1 : " << pointList_C1.size() << endl;
	cout << "pointList_C2 : " << pointList_C2.size() << endl;
		
	// ==========================================================================
	// Setting up the factor graph and distributions
	// ==========================================================================
	
	// Creates the propagator
	MaxProductPropagator propagator;
	propagator.setMaxIterations(50);
	propagator.setTolerance(0.000001);
		
	// Creates the factor graph
	FactorGraph* graph = createFactorGraph();
	
	// Loads the factor graph
	propagator.loadFactorGraph(graph);
	
	// Creates the parameter domains
	MersenneTwister::init(time(NULL));
	DiscDomain domain(0, 1.0);
	domain.repartition(10);
	
	// Creates the distributions
	DiscreteDistribution* dist_k1 = new DiscreteDistribution();
	DiscreteDistribution* dist_k2 = new DiscreteDistribution();
	DiscreteDistribution* dist_k3 = new DiscreteDistribution();
	DiscreteDistribution* dist_k4 = new DiscreteDistribution();	
	DiscreteDistribution* dist_k5 = new DiscreteDistribution();
	DiscreteDistribution* dist_k6 = new DiscreteDistribution();
	DiscreteDistribution* dist_k7 = new DiscreteDistribution();
	DiscreteDistribution* dist_k8 = new DiscreteDistribution();
	DiscreteDistribution* dist_k9 = new DiscreteDistribution();
	DiscreteDistribution* dist_k10 = new DiscreteDistribution();
	DiscreteDistribution* dist_k11 = new DiscreteDistribution();
	DiscreteDistribution* dist_k12 = new DiscreteDistribution();
	
	DiscreteDistribution* dist_f1_1 = new DiscreteDistribution();
	DiscreteDistribution* dist_f1_2 = new DiscreteDistribution();
	DiscreteDistribution* dist_f2_1 = new DiscreteDistribution();
	DiscreteDistribution* dist_f2_2 = new DiscreteDistribution();
	DiscreteDistribution* dist_f3_1 = new DiscreteDistribution();
	DiscreteDistribution* dist_f3_2 = new DiscreteDistribution();
	DiscreteDistribution* dist_f4 = new DiscreteDistribution();
	DiscreteDistribution* dist_f5 = new DiscreteDistribution();
	DiscreteDistribution* dist_f6 = new DiscreteDistribution();
	DiscreteDistribution* dist_f7 = new DiscreteDistribution();
	DiscreteDistribution* dist_f8 = new DiscreteDistribution();
	DiscreteDistribution* dist_f9 = new DiscreteDistribution();
	DiscreteDistribution* dist_f10 = new DiscreteDistribution();
	DiscreteDistribution* dist_f11 = new DiscreteDistribution();
	
	// Initializes the distributions for the parameters
	dist_k1->addVariable("k1", domain.partitionSize()); 
	dist_k1->initializeTable(); 
	dist_k1->makeUniform();
	
	dist_k2->addVariable("k2", domain.partitionSize()); 
	dist_k2->initializeTable(); 
	dist_k2->makeUniform();
	
	dist_k3->addVariable("k3", domain.partitionSize()); 
	dist_k3->initializeTable(); 
	dist_k3->makeUniform();
	
	dist_k4->addVariable("k4", domain.partitionSize()); 
	dist_k4->initializeTable(); 
	dist_k4->makeUniform();
	
	dist_k5->addVariable("k5", domain.partitionSize()); 
	dist_k5->initializeTable(); 
	dist_k5->makeUniform();
	
	dist_k6->addVariable("k6", domain.partitionSize()); 
	dist_k6->initializeTable(); 
	dist_k6->makeUniform();
	
	dist_k7->addVariable("k7", domain.partitionSize()); 
	dist_k7->initializeTable(); 
	dist_k7->makeUniform();
	
	dist_k8->addVariable("k8", domain.partitionSize()); 
	dist_k8->initializeTable(); 
	dist_k8->makeUniform();
	
	dist_k9->addVariable("k9", domain.partitionSize()); 
	dist_k9->initializeTable(); 
	dist_k9->makeUniform();
	
	dist_k10->addVariable("k10", domain.partitionSize()); 
	dist_k10->initializeTable(); 
	dist_k10->makeUniform();
	
	dist_k11->addVariable("k11", domain.partitionSize()); 
	dist_k11->initializeTable(); 
	dist_k11->makeUniform();
	
	dist_k12->addVariable("k12", domain.partitionSize()); 
	dist_k12->initializeTable(); 
	dist_k12->makeUniform();
	
	// Initialize the factor distributions.
	dist_f1_1->addVariable("k1", domain.partitionSize()); 
	dist_f1_1->addVariable("k2", domain.partitionSize()); 
	dist_f1_1->addVariable("k3", domain.partitionSize()); 
	dist_f1_1->initializeTable();
	dist_f1_1->setValue(-1.0);

	dist_f1_2->addVariable("k1", domain.partitionSize()); 
	dist_f1_2->addVariable("k2", domain.partitionSize()); 
	dist_f1_2->addVariable("k3", domain.partitionSize()); 
	dist_f1_2->initializeTable();
	dist_f1_2->setValue(-1.0);
	
	dist_f2_1->addVariable("k3", domain.partitionSize()); 
	dist_f2_1->addVariable("k4", domain.partitionSize()); 
	dist_f2_1->initializeTable();
	dist_f2_1->setValue(-1.0);
	
	dist_f2_2->addVariable("k3", domain.partitionSize()); 
	dist_f2_2->addVariable("k4", domain.partitionSize()); 
	dist_f2_2->initializeTable();
	dist_f2_2->setValue(-1.0);
	
	dist_f3_1->addVariable("k3", domain.partitionSize()); 
	dist_f3_1->addVariable("k4", domain.partitionSize()); 
	dist_f3_1->addVariable("k5", domain.partitionSize()); 
	dist_f3_1->initializeTable();
	dist_f3_1->setValue(-1.0);
	
	dist_f3_2->addVariable("k3", domain.partitionSize()); 
	dist_f3_2->addVariable("k4", domain.partitionSize()); 
	dist_f3_2->addVariable("k9", domain.partitionSize()); 
	dist_f3_2->initializeTable();
	dist_f3_2->setValue(-1.0);
	
	dist_f4->addVariable("k5", domain.partitionSize()); 
	dist_f4->addVariable("k6", domain.partitionSize()); 
	dist_f4->initializeTable();
	dist_f4->setValue(-1.0);
	
	dist_f5->addVariable("k5", domain.partitionSize()); 
	dist_f5->addVariable("k6", domain.partitionSize()); 
	dist_f5->addVariable("k7", domain.partitionSize()); 
	dist_f5->initializeTable();
	dist_f5->setValue(-1.0);
	
	dist_f6->addVariable("k7", domain.partitionSize()); 
	dist_f6->addVariable("k8", domain.partitionSize()); 
	dist_f6->initializeTable();
	dist_f6->setValue(-1.0);
	
	dist_f7->addVariable("k7", domain.partitionSize()); 
	dist_f7->addVariable("k8", domain.partitionSize()); 
	dist_f7->initializeTable();
	dist_f7->setValue(-1.0);
	
	dist_f8->addVariable("k9", domain.partitionSize()); 
	dist_f8->addVariable("k10", domain.partitionSize()); 
	dist_f8->initializeTable();
	dist_f8->setValue(-1.0);
	
	dist_f9->addVariable("k9", domain.partitionSize()); 
	dist_f9->addVariable("k10", domain.partitionSize()); 
	dist_f9->addVariable("k11", domain.partitionSize()); 
	dist_f9->initializeTable();
	dist_f9->setValue(-1.0);
	
	dist_f10->addVariable("k11", domain.partitionSize()); 
	dist_f10->addVariable("k12", domain.partitionSize()); 
	dist_f10->initializeTable();
	dist_f10->setValue(-1.0);
	
	dist_f11->addVariable("k11", domain.partitionSize()); 
	dist_f11->addVariable("k12", domain.partitionSize()); 
	dist_f11->initializeTable();
	dist_f11->setValue(-1.0);
	
	// Strings to locate the positions in the distribution tables
	string f1_1_expr;
	string f1_2_expr;
	string f2_1_expr;
	string f2_2_expr;
	string f3_1_expr;
	string f3_2_expr;
	string f4_expr;
	string f5_expr;
	string f6_expr;
	string f7_expr;
	string f8_expr;
	string f9_expr;
	string f10_expr;
	string f11_expr;
	
	// Sampling constants for the variables
	int k1 = 0;
	int k2 = 1;
	int k3 = 2;
	int k4 = 3;
	int k5 = 4;
	int k6 = 5;
	int k7 = 6;
	int k8 = 7;
	int k9 = 8;
	int k10 = 9;
	int k11 = 10;
	int k12 = 11;
		
	// Scoring variables
	double scoref1;
	double scoref2;
	double scoref3;
	double scoref4;
	double scoref5;
	double scoref6;
	double scoref7;
	double scoref8;
	double scoref9;
	double scoref10;
	double scoref11;
	double scoreAll;
	
	int maxIterations = 2000;
	
	// Sampling loop for the first component
	for (int i = 0; i < maxIterations; i++) {
		
		// Sample for the parameters
		p[k1] = domain.random();
		p[k2] = domain.random();
		p[k3] = domain.random();
		p[k4] = domain.random();
		p[k5] = domain.random();
		p[k6] = domain.random();
		p[k7] = domain.random();
		p[k8] = domain.random();
		p[k9] = domain.random();
		p[k10] = domain.random();
		p[k11] = domain.random();
		p[k12] = domain.random();
		
		// Simulate and get the trace
		ODETrace* sampleTrace = createTrace(p, y, *model);
		
		if (sampleTrace != 0) {
			scoref1 = score(pointList_x1, *sampleTrace);
			scoref2 = score(pointList_C1, *sampleTrace);
			scoref3 = score(pointList_C1, *sampleTrace);
			scoref4 = score(pointList_C1, *sampleTrace);
			scoref5 = score(pointList_x5, *sampleTrace);
			scoref6 = score(pointList_C1, *sampleTrace);
			scoref7 = score(pointList_x7, *sampleTrace);
			
			cout << p[k1] << ", " << p[k2] << ", " << p[k3] << ", " << p[k4] << ", " << p[k5] << ", " << p[k6] << ", " << p[k7] << ", " << p[k8] << ", "
				 << scoref1 << ", " << scoref5 << ", " << scoref7 <<  ", " << scoref2 << endl;
				  
			delete sampleTrace;
			
			// Form the strings and puts it in the distributions
			f1_1_expr = "k1 " + itos(domain.indexOf(p[k1])) + " k2 " + itos(domain.indexOf(p[k2])) + " k3 " + itos(domain.indexOf(p[k3]));
			f2_1_expr = "k3 " + itos(domain.indexOf(p[k3])) + " k4 " + itos(domain.indexOf(p[k4]));
			f3_1_expr = "k3 " + itos(domain.indexOf(p[k3])) + " k4 " + itos(domain.indexOf(p[k4])) + " k5 " + itos(domain.indexOf(p[k5]));
			f4_expr = "k5 " + itos(domain.indexOf(p[k5])) + " k6 " + itos(domain.indexOf(p[k6]));
			f5_expr = "k5 " + itos(domain.indexOf(p[k5])) + " k6 " + itos(domain.indexOf(p[k6])) + " k7 " + itos(domain.indexOf(p[k7]));
			f6_expr = "k7 " + itos(domain.indexOf(p[k7])) + " k8 " + itos(domain.indexOf(p[k8]));
			f7_expr = "k7 " + itos(domain.indexOf(p[k7])) + " k8 " + itos(domain.indexOf(p[k8]));
			
			// Imputs the score into the 
			dist_f1_1->replaceIfSmallerProbability(f1_1_expr, scoref1);
			dist_f2_1->replaceIfSmallerProbability(f2_1_expr, scoref2);
			dist_f3_1->replaceIfSmallerProbability(f3_1_expr, scoref3);
			dist_f4->replaceIfSmallerProbability(f4_expr, scoref4);
			dist_f5->replaceIfSmallerProbability(f5_expr, scoref5);
			dist_f6->replaceIfSmallerProbability(f6_expr, scoref6);
			dist_f7->replaceIfSmallerProbability(f7_expr, scoref7);
			
		} else {
			// Do nothing
		}
		
	}

	cout << endl;
	
	// Sampling loop for the second component
	for (int i = 0; i < maxIterations; i++) {
		
		// Sample for the parameters
		p[k1] = domain.random();
		p[k2] = domain.random();
		p[k3] = domain.random();
		p[k4] = domain.random();
		p[k5] = domain.random();
		p[k6] = domain.random();
		p[k7] = domain.random();
		p[k8] = domain.random();
		p[k9] = domain.random();
		p[k10] = domain.random();
		p[k11] = domain.random();
		p[k12] = domain.random();
		
		// Simulate and get the trace
		ODETrace* sampleTrace = createTrace(p, y, *model);
		
		if (sampleTrace != 0) {
			scoref1 = score(pointList_x1, *sampleTrace);
			scoref2 = score(pointList_C2, *sampleTrace);
			scoref3 = score(pointList_C2, *sampleTrace);
			scoref8 = score(pointList_C2, *sampleTrace);
			scoref9 = score(pointList_x9, *sampleTrace);
			scoref10 = score(pointList_C2, *sampleTrace);
			scoref11 = score(pointList_x11, *sampleTrace);
			
			cout << p[k1] << ", " << p[k2] << ", " << p[k3] << ", " << p[k4] << ", " << p[k9] << ", " << p[k10] << ", " << p[k11] << ", " << p[k12] << ", "
				 << scoref1 << ", " << scoref9 << ", " << scoref11 <<  ", " << scoref2 << endl;
			
			
			delete sampleTrace;
			
			// Form the strings and puts it in the distributions
			f1_2_expr = "k1 " + itos(domain.indexOf(p[k1])) + " k2 " + itos(domain.indexOf(p[k2])) + " k3 " + itos(domain.indexOf(p[k3]));
			f2_2_expr = "k3 " + itos(domain.indexOf(p[k3])) + " k4 " + itos(domain.indexOf(p[k4]));
			f3_2_expr = "k3 " + itos(domain.indexOf(p[k3])) + " k4 " + itos(domain.indexOf(p[k4])) + " k9 " + itos(domain.indexOf(p[k9]));
			f8_expr = "k9 " + itos(domain.indexOf(p[k9])) + " k10 " + itos(domain.indexOf(p[k10]));
			f9_expr = "k9 " + itos(domain.indexOf(p[k9])) + " k10 " + itos(domain.indexOf(p[k10])) + " k11 " + itos(domain.indexOf(p[k11]));
			f10_expr = "k11 " + itos(domain.indexOf(p[k11])) + " k12 " + itos(domain.indexOf(p[k12]));
			f11_expr = "k11 " + itos(domain.indexOf(p[k11])) + " k12 " + itos(domain.indexOf(p[k12]));
			
			// Imputs the score into the 
			dist_f1_2->replaceIfSmallerProbability(f1_2_expr, scoref1);
			dist_f2_2->replaceIfSmallerProbability(f2_2_expr, scoref2);
			dist_f3_2->replaceIfSmallerProbability(f3_2_expr, scoref3);
			dist_f8->replaceIfSmallerProbability(f8_expr, scoref8);
			dist_f9->replaceIfSmallerProbability(f9_expr, scoref9);
			dist_f10->replaceIfSmallerProbability(f10_expr, scoref10);
			dist_f11->replaceIfSmallerProbability(f11_expr, scoref11);
			
		} else {
			// Do nothing
		}
		
	}

	
	// Get the maximal value for each factor distribution
	double max_f1_1 = dist_f1_1->maximum();
	double max_f1_2 = dist_f1_2->maximum();
	double max_f2_1 = dist_f2_1->maximum();
	double max_f2_2 = dist_f2_2->maximum();
	double max_f3_1 = dist_f3_1->maximum();
	double max_f3_2 = dist_f3_2->maximum();
	double max_f4 = dist_f4->maximum();
	double max_f5 = dist_f5->maximum();
	double max_f6 = dist_f6->maximum();
	double max_f7 = dist_f7->maximum();
	double max_f8 = dist_f8->maximum();
	double max_f9 = dist_f9->maximum();
	double max_f10 = dist_f10->maximum();
	double max_f11 = dist_f11->maximum();
	
	// Replaces empty bins with the maximum value
	dist_f1_1->replaceValue(-1.0, max_f1_1);
	dist_f1_2->replaceValue(-1.0, max_f1_2);
	dist_f2_1->replaceValue(-1.0, max_f2_1);
	dist_f2_2->replaceValue(-1.0, max_f2_2);
	dist_f3_1->replaceValue(-1.0, max_f3_1);
	dist_f3_2->replaceValue(-1.0, max_f3_2);
	dist_f4->replaceValue(-1.0, max_f4);
	dist_f5->replaceValue(-1.0, max_f5);
	dist_f6->replaceValue(-1.0, max_f6);
	dist_f7->replaceValue(-1.0, max_f7);
	dist_f8->replaceValue(-1.0, max_f8);
	dist_f9->replaceValue(-1.0, max_f9);
	dist_f10->replaceValue(-1.0, max_f10);
	dist_f11->replaceValue(-1.0, max_f11);
	
	// Divide each factor by the maximum value and multiply by 10
	dist_f1_1->multiply(20.0 / max_f1_1);
	dist_f1_2->multiply(20.0 / max_f1_2);
	dist_f2_1->multiply(20.0 / max_f2_1);
	dist_f2_2->multiply(20.0 / max_f2_2);
	dist_f3_1->multiply(20.0 / max_f3_1);
	dist_f3_2->multiply(20.0 / max_f3_2);
	dist_f4->multiply(20.0 / max_f4);
	dist_f5->multiply(20.0 / max_f5);
	dist_f6->multiply(20.0 / max_f6);
	dist_f7->multiply(20.0 / max_f7);
	dist_f8->multiply(20.0 / max_f8);
	dist_f9->multiply(20.0 / max_f9);
	dist_f10->multiply(20.0 / max_f10);
	dist_f11->multiply(20.0 / max_f11);

	// Exponent them and inverse the distributions
	dist_f1_1->exponent();
	dist_f1_2->exponent();
	dist_f2_1->exponent();
	dist_f2_2->exponent();
	dist_f3_1->exponent();
	dist_f3_2->exponent();
	dist_f4->exponent();
	dist_f5->exponent();
	dist_f6->exponent();
	dist_f7->exponent();
	dist_f8->exponent();
	dist_f9->exponent();
	dist_f10->exponent();
	dist_f11->exponent();
	
	dist_f1_1->inverse();
	dist_f1_2->inverse();
	dist_f2_1->inverse();
	dist_f2_2->inverse();
	dist_f3_1->inverse();
	dist_f3_2->inverse();
	dist_f4->inverse();
	dist_f5->inverse();
	dist_f6->inverse();
	dist_f7->inverse();
	dist_f8->inverse();
	dist_f9->inverse();
	dist_f10->inverse();
	dist_f11->inverse();

	// Normalize the factor distribution
	dist_f1_1->normalize();
	dist_f1_2->normalize();
	dist_f2_1->normalize();
	dist_f2_2->normalize();
	dist_f3_1->normalize();
	dist_f3_2->normalize();
	dist_f4->normalize();
	dist_f5->normalize();	
	dist_f6->normalize();	
	dist_f7->normalize();	
	dist_f8->normalize();	
	dist_f9->normalize();	
	dist_f10->normalize();	
	dist_f11->normalize();	

	cout << dist_f1_1->toStringDetail() << endl;
	cout << dist_f1_2->toStringDetail() << endl;
	cout << dist_f2_1->toStringDetail() << endl;
	cout << dist_f2_2->toStringDetail() << endl;
	cout << dist_f3_1->toStringDetail() << endl;
	cout << dist_f3_2->toStringDetail() << endl;
	cout << dist_f4->toStringDetail() << endl;
	cout << dist_f5->toStringDetail() << endl;
	cout << dist_f6->toStringDetail() << endl;
	cout << dist_f7->toStringDetail() << endl;
	cout << dist_f8->toStringDetail() << endl;
	cout << dist_f9->toStringDetail() << endl;
	cout << dist_f10->toStringDetail() << endl;
	cout << dist_f11->toStringDetail() << endl;
	
	// Associates distributions
	propagator.setDistribution("k1", dist_k1);
	propagator.setDistribution("k2", dist_k2);
	propagator.setDistribution("k3", dist_k3);
	propagator.setDistribution("k4", dist_k4);
	propagator.setDistribution("k5", dist_k5);
	propagator.setDistribution("k6", dist_k6);
	propagator.setDistribution("k7", dist_k7);
	propagator.setDistribution("k8", dist_k8);
	propagator.setDistribution("k9", dist_k9);
	propagator.setDistribution("k10", dist_k10);
	propagator.setDistribution("k11", dist_k11);
	propagator.setDistribution("k12", dist_k12);
		
	propagator.setDistribution("f1_1", dist_f1_1);
	propagator.setDistribution("f1_2", dist_f1_2);
	propagator.setDistribution("f2_1", dist_f2_1);
	propagator.setDistribution("f2_2", dist_f2_2);
	propagator.setDistribution("f3_1", dist_f3_1);
	propagator.setDistribution("f3_2", dist_f3_2);
	propagator.setDistribution("f4", dist_f4);
	propagator.setDistribution("f5", dist_f5);
	propagator.setDistribution("f6", dist_f6);
	propagator.setDistribution("f7", dist_f7);
	propagator.setDistribution("f8", dist_f8);
	propagator.setDistribution("f9", dist_f9);
	propagator.setDistribution("f10", dist_f10);
	propagator.setDistribution("f11", dist_f11);
	
	// Propagates
	propagator.propagate();	

	// Using the best bins, use the mid point values to simulate the system again
	int k1_bestIndex = propagator.getDistribution("k1")->maximumIndex();
	int k2_bestIndex = propagator.getDistribution("k2")->maximumIndex();
	int k3_bestIndex = propagator.getDistribution("k3")->maximumIndex();
	int k4_bestIndex = propagator.getDistribution("k4")->maximumIndex();
	int k5_bestIndex = propagator.getDistribution("k5")->maximumIndex();
	int k6_bestIndex = propagator.getDistribution("k6")->maximumIndex();
	int k7_bestIndex = propagator.getDistribution("k7")->maximumIndex();
	int k8_bestIndex = propagator.getDistribution("k8")->maximumIndex();
	int k9_bestIndex = propagator.getDistribution("k9")->maximumIndex();
	int k10_bestIndex = propagator.getDistribution("k10")->maximumIndex();
	int k11_bestIndex = propagator.getDistribution("k11")->maximumIndex();
	int k12_bestIndex = propagator.getDistribution("k12")->maximumIndex();

	p[k1] = domain.midValue(k1_bestIndex);
	p[k2] = domain.midValue(k2_bestIndex);
	p[k3] = domain.midValue(k3_bestIndex);
	p[k4] = domain.midValue(k4_bestIndex);
	p[k5] = domain.midValue(k5_bestIndex);
	p[k6] = domain.midValue(k6_bestIndex);
	p[k7] = domain.midValue(k7_bestIndex);
	p[k8] = domain.midValue(k8_bestIndex);
	p[k9] = domain.midValue(k9_bestIndex);
	p[k10] = domain.midValue(k10_bestIndex);
	p[k11] = domain.midValue(k11_bestIndex);
	p[k12] = domain.midValue(k12_bestIndex);
	
	ODETrace* est_trace = createTrace(p, y, *model);
	
	// Writes the best values
	cout << "Best estimated value" << endl;
	cout << "k1 : " << p[k1] << endl;
	cout << "k2 : " << p[k2] << endl;
	cout << "k3 : " << p[k3] << endl;
	cout << "k4 : " << p[k4] << endl;
	cout << "k5 : " << p[k5] << endl;
	cout << "k6 : " << p[k6] << endl;
	cout << "k7 : " << p[k7] << endl;
	cout << "k8 : " << p[k8] << endl;
	cout << "k9 : " << p[k9] << endl;
	cout << "k10 : " << p[k10] << endl;
	cout << "k11 : " << p[k11] << endl;
	cout << "k12 : " << p[k12] << endl;
		
	if (est_trace != 0) {
		est_trace->save("trace.model1estimate.txt");
		delete est_trace;
	}
	
	// Unloads factor graph
	propagator.unloadFactorGraph();
	
	// Frees up memory
	delete trace;
	delete graph;
	delete model;
}

}


